// export-module for VModel files

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "stdafx.h"
#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "TextWin.h"

extern float RFTrans2(matrix *M, vector *S, angles *R, vector *T);

void OutputTransformations(Surface *S, FILE *outfile){
  vector scl; angles rot; vector trn;
  if(RFTrans2(&(S->TM), &scl, &rot, &trn)<0.01f){
    if((fabs(scl.x-1.0f)>0.0001f) || (fabs(scl.y-1.0f)>0.0001f) || (fabs(scl.z-1.0f)>0.0001f)){
      fprintf(outfile,"  scale     < %f, %f, %f >\n", scl.x, scl.y, scl.z);
	}
	if(fabs(rot.z)>0.001f) fprintf(outfile,"  rotate    < 0.0, 0.0, %f >\n",-Rad2Deg*(rot.z));
	if(fabs(rot.x)>0.001f) fprintf(outfile,"  rotate    < %f, 0.0, 0.0 >\n",-Rad2Deg*(rot.x));
	if(fabs(rot.y)>0.001f) fprintf(outfile,"  rotate    < 0.0, %f, 0.0 >\n",-Rad2Deg*(rot.y));
	if((fabs(trn.x-1.0f)>0.0001f) && (fabs(trn.y-1.0f)>0.0001f) && (fabs(trn.z-1.0f)>0.0001f)){
      fprintf(outfile,"  translate < %f, %f, %f >\n", trn.x, trn.y, trn.z);
	}
	return;
  }
}


int chrbtw(char test, char min, char max){ return (test>=min) && (test<=max); }

void povnam(char *dst, char *src)
{
  int i;
  for(i=0; src[i]; i++){
	dst[i]=((src[i]=='_') || chrbtw(src[i],'a','z') || chrbtw(src[i],'A','Z') || chrbtw(src[i],'0','9'))? src[i] : '_';
  }
  dst[i]=0;
}


void CVModelDoc::ExportPOV(char *filename, ViewPtr V)
{
  FILE   *outfile;
  char    newnam[64];
  PrimPtr Prim;
  int     lights=0;

  outfile = fopen(filename,"wt");
  if (outfile){
    TextWin.Output(1,"- Writing POV-File-Header ...");
    fprintf(outfile,"// Persistence Of Vision raytracer version 3.X scene file.\n");
    fprintf(outfile,"// \n");
    fprintf(outfile,"// converted by VModel V0.6 (C)2002 Michael Granz\n\n");

    fprintf(outfile,"camera {\n");
    fprintf(outfile,"  location  < %f, %f, %f >\n", V->Location.x, V->Location.y, V->Location.z);
    fprintf(outfile,"  direction < 0.0, 0.0, %f >\n",1.0/(V->h));
    fprintf(outfile,"  up        < 0.0, 1.0, 0.0 >\n");
    fprintf(outfile,"  right     < -1.33333, 0.0, 0.0 >\n");
    fprintf(outfile,"  sky       < %f, %f, %f >\n", V->Up.x, V->Up.y, V->Up.z);
    fprintf(outfile,"  look_at   < %f, %f, %f >\n", V->Target.x, V->Target.y, V->Target.z);
    fprintf(outfile,"}\n\n");

    fprintf(outfile,"background { color red %f green %f blue %f }\n\n",VMF.Background.R,VMF.Background.G,VMF.Background.B);
    
    Prim = VMF.Base;
    while(Prim){
      if(Prim->idnt==POLYMESH){
        PolyMesh *p = (PolyMesh *)Prim;
		if(p->Mat->Emission.R>0.01f){
		  vector base=p->pts[p->pfv[0]];
		  vector p1=p->pts[p->pfv[1]]-base;
		  vector p2=p->pts[p->pfv[(p->pnv)? p->pnv[0]-1 : 2]]-base;
          vector L = base+0.5f*p1+0.5f*p2-0.01f*unit(p1*p2);
		  fprintf(outfile,"light_source {\n  < %f, %f, %f > color red 1.0 green 1.0 blue 1.0\n",L.x,L.y,L.z);
          fprintf(outfile,"  //area_light <%f, %f, %f>, <%f, %f ,%f>, 5, 5\n",p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);
          fprintf(outfile,"  //adaptive 1 jitter\n}\n\n");
		  lights++;
        }
      }
      Prim=Prim->succ;
    }

	if(!lights){
	  float  d = Betrag(V->Target - V->Location);
      vector N = unit  (V->Target - V->Location);
      vector U = N*V->Up;
      vector L = V->Location - d*N + d*U + d*V->Up;
      fprintf(outfile,"light_source { < %f, %f, %f > color red 1.0 green 1.0 blue 1.0 }\n\n",L.x,L.y,L.z);
	}

    TextWin.Output(1,"\n          Materials ");
    MatPtr mp = VMF.MatList;
    while(mp){
      TextWin.Output(1,".");
	  povnam(newnam,mp->name);
      fprintf(outfile,"#declare MAT_%s = texture {\n",newnam);
      if(mp->Reflection.G<0.0f){
        fprintf(outfile,"  pigment { color red %f green %f blue %f }\n", mp->Reflection.R, mp->Reflection.R, mp->Reflection.R);
	  } else {
        if(mp->Reflection.R<0.0f){
          fprintf(outfile,"  pigment { color red 1.0 green 1.0 blue 1.0 filter %f }\n",mp->Trns(555.0f, mp->IOR(555.0f), VMF.Environment.IOR(555.0f), 1.0f));
		} else {
		  if(mp->Texture){
            fprintf(outfile,"  pigment { image_map { bmp \"%s\" } }\n", mp->Texture->tname);
		  } else {
            fprintf(outfile,"  pigment { color red %f green %f blue %f }\n", mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
		  }
		}
	  }
	  fprintf(outfile,"  finish  {\n");
      fprintf(outfile,"    ambient           %f\n",(VMF.Ambient.R + VMF.Ambient.G + VMF.Ambient.B)/3.0F);
      fprintf(outfile,"    diffuse           %f\n",1.0f-mp->Specularity);
      fprintf(outfile,"    phong             %f phong_size %d\n",mp->Specularity,(int)(mp->PBExp? mp->PBExp : 50.0f));
      if(mp->Specularity>=0.3f) fprintf(outfile,"    reflection        %f\n",mp->Spec(555.0f, mp->IOR(555.0f), VMF.Environment.IOR(555.0f), 1.0f));
      if((mp->Reflection.R < 0.0f) || (mp->Transmission.R > 0.001)) fprintf(outfile,"    refraction on ior %f\n",mp->IOR(555.0f));
      fprintf(outfile,"  }\n");
      fprintf(outfile,"}\n\n");
      mp = mp->next;
    }
    TextWin.Output(1,"\n          Objects ");
    PrimPtr Prim = VMF.Base;
    while(Prim){
      TextWin.Output(1,".");
      switch (Prim->idnt) {
        case ROOT     : break;
        case CUBE     : {
          fprintf(outfile,"box { <-0.5, -0.5, -0.5>, <0.5, 0.5, 0.5>\n");
          OutputTransformations((Surface *)Prim, outfile);
          fprintf(outfile,"  texture { MAT_%s }\n",((Surface *)Prim)->Mat->name);
          fprintf(outfile,"}\n\n");
          break;
        }
        case SPHERE   : {
          fprintf(outfile,"sphere {\n  <0.0, 0.0, 0.0>, 1.0\n");
          OutputTransformations((Surface *)Prim, outfile);
          fprintf(outfile,"  texture { MAT_%s }\n",((Surface *)Prim)->Mat->name);
          fprintf(outfile,"}\n\n");
          break;
        }
        case CYLINDER : {
          fprintf(outfile,"cylinder {\n  <0.0, 0.0, 0.0>, <0.0, 0.0, 1.0>, 1.0 open\n");
          OutputTransformations((Surface *)Prim, outfile);
          fprintf(outfile,"  texture { MAT_%s }\n",((Surface *)Prim)->Mat->name);
          fprintf(outfile,"}\n\n");
          break;
        }
        case TORUS    : {
          fprintf(outfile,"torus {\n  %f, %f\n", ((Torus *)Prim)->R1, ((Torus *)Prim)->R2);
          OutputTransformations((Surface *)Prim, outfile);
          fprintf(outfile,"  texture { MAT_%s }\n",((Surface *)Prim)->Mat->name);
          fprintf(outfile,"}\n\n");
          break;
        }
        case POLYMESH : {
          int      fi,i,j,k;
          vector   p0,n0;
          PolyMesh *p = (PolyMesh *)Prim;
		  if(p->nm > 1){
		    /*for(k=0; k<((PolyMesh *)p)->nm; k++){
			  int nt=0; for(i=0; i<(p->mesh.nf); i++) if(((PolyMesh *)p)->mat[i] == k) nt += (((PolyMesh *)p)->pnv? ((PolyMesh *)p)->pnv[i]-2 : 1);
		      if(nt > 1){
				fprintf(outfile,"mesh {\n");
                for(ix=0,i=0; i<(p->mesh.nf); i++) {
                  ne=p->mesh.face[ix];
				  if(((PolyMesh *)p)->mat[i] == k){
                                   p0=p->mesh.vert[p->mesh.face[ix+2   ]];
                    //if(p->mesh.nn) n0=p->mesh.vnrm[p->mesh.face[ix+2+ne]];
                    for(j=2; j<ne; j++){
                      if(nt > 1) fprintf(outfile,"  ");
                      //if(p->mesh.nn){
                      //  fprintf(outfile,"smooth_triangle { ");
                      //  fprintf(outfile,"< %f, %f, %f >, ",p0.x, p0.y, p0.z);
                      //  fprintf(outfile,"< %f, %f, %f >, ",n0.x, n0.y, n0.z);
                      //  fprintf(outfile,"< %f, %f, %f >, ",p->mesh.vert[p->mesh.face[ix+j+1   ]].x, p->mesh.vert[p->mesh.face[ix+j+1   ]].y, p->mesh.vert[p->mesh.face[ix+j+1   ]].z);
                      //  fprintf(outfile,"< %f, %f, %f >, ",p->mesh.vnrm[p->mesh.face[ix+j+1+ne]].x, p->mesh.vnrm[p->mesh.face[ix+j+1+ne]].y, p->mesh.vnrm[p->mesh.face[ix+j+1+ne]].z);
                      //  fprintf(outfile,"< %f, %f, %f >, ",p->mesh.vert[p->mesh.face[ix+j+2   ]].x, p->mesh.vert[p->mesh.face[ix+j+2   ]].y, p->mesh.vert[p->mesh.face[ix+j+2   ]].z);
                      //  fprintf(outfile,"< %f, %f, %f >"  ,p->mesh.vnrm[p->mesh.face[ix+j+2+ne]].x, p->mesh.vnrm[p->mesh.face[ix+j+2+ne]].y, p->mesh.vnrm[p->mesh.face[ix+j+2+ne]].z);
                      //} else {
                        fprintf(outfile,"triangle { ");
                        fprintf(outfile,"< %f, %f, %f >, ",p0.x, p0.y, p0.z);
                        fprintf(outfile,"< %f, %f, %f >, ",p->mesh.vert[p->mesh.face[ix+j+1]].x, p->mesh.vert[p->mesh.face[ix+j+1]].y, p->mesh.vert[p->mesh.face[ix+j+1]].z);
                        fprintf(outfile,"< %f, %f, %f >"  ,p->mesh.vert[p->mesh.face[ix+j+2]].x, p->mesh.vert[p->mesh.face[ix+j+2]].y, p->mesh.vert[p->mesh.face[ix+j+2]].z);
                      //}
                      if(nt > 1) fprintf(outfile," }\n"); else fprintf(outfile,"\n");
                    }
				  }
                  ix+=(2*ne+2);
                }
				for(i=0; ((PolyMesh *)p)->mtl[k]->name[i]; i++) newnam[i]=(((PolyMesh *)p)->mtl[k]->name[i]==' ')? '_' : ((PolyMesh *)p)->mtl[k]->name[i]; newnam[i]=0;
                povnam(newnam,((PolyMesh *)p)->mtl[k]->name);
				fprintf(outfile,"  texture { MAT_%s }\n",newnam);
                fprintf(outfile,"}\n\n");
			  }
			}*/
          } else {
		    if((p->nf > 1) || (p->pnv && (p->pnv[0]>3))) fprintf(outfile,"mesh {\n");
            for(fi=0,i=0; i<(p->nf); i++){
			  k=(p->pnv)? p->pnv[i] : 3;
              p0=p->pts[p->pfv[fi]];
			  if(p->nrm) n0=p->nrm[p->pfn? p->pfn[fi] : p->pfv[fi]];
			  for(j=1; j<(k-1); j++){
                if((p->nf > 1) || (p->pnv && (p->pnv[0]>3))) fprintf(outfile,"  ");
                if(p->nrm){
                  fprintf(outfile,"smooth_triangle { ");
                  fprintf(outfile,"< %f, %f, %f >, ",p0.x, p0.y, p0.z);
                  fprintf(outfile,"< %f, %f, %f >, ",n0.x, n0.y, n0.z);
                  fprintf(outfile,"< %f, %f, %f >, ",p->pts[p->pfv[fi+j  ]].x, p->pts[p->pfv[fi+j  ]].y, p->pts[p->pfv[fi+j  ]].z);
                  if(p->pfn){
				    fprintf(outfile,"< %f, %f, %f >, ",p->nrm[p->pfn[fi+j  ]].x, p->nrm[p->pfn[fi+j  ]].y, p->nrm[p->pfn[fi+j  ]].z);
                    fprintf(outfile,"< %f, %f, %f >"  ,p->pts[p->pfv[fi+j+1]].x, p->pts[p->pfv[fi+j+1]].y, p->pts[p->pfv[fi+j+1]].z);
                    fprintf(outfile,"< %f, %f, %f >"  ,p->nrm[p->pfn[fi+j+1]].x, p->nrm[p->pfn[fi+j+1]].y, p->nrm[p->pfn[fi+j+1]].z);
				  } else {
				    fprintf(outfile,"< %f, %f, %f >, ",p->nrm[p->pfv[fi+j  ]].x, p->nrm[p->pfv[fi+j  ]].y, p->nrm[p->pfv[fi+j  ]].z);
                    fprintf(outfile,"< %f, %f, %f >"  ,p->pts[p->pfv[fi+j+1]].x, p->pts[p->pfv[fi+j+1]].y, p->pts[p->pfv[fi+j+1]].z);
                    fprintf(outfile,"< %f, %f, %f >"  ,p->nrm[p->pfv[fi+j+1]].x, p->nrm[p->pfv[fi+j+1]].y, p->nrm[p->pfv[fi+j+1]].z);
				  }
				} else {
				  fprintf(outfile,"triangle { ");
                  fprintf(outfile,"< %f, %f, %f >, ",p0.x, p0.y, p0.z);
                  fprintf(outfile,"< %f, %f, %f >, ",p->pts[p->pfv[fi+j  ]].x, p->pts[p->pfv[fi+j  ]].y, p->pts[p->pfv[fi+j  ]].z);
                  fprintf(outfile,"< %f, %f, %f >"  ,p->pts[p->pfv[fi+j+1]].x, p->pts[p->pfv[fi+j+1]].y, p->pts[p->pfv[fi+j+1]].z);
                }
				if((p->nf > 1) || (p->pnv && (p->pnv[0]>3))) fprintf(outfile," }\n"); else fprintf(outfile,"\n");
			  }
			  fi+=k;
			}
	        povnam(newnam, p->Mat->name);
			//if(p->Mat->Texture){
            //  fprintf(outfile,"  texture  { MAT_%s scale < %f, %f, 1.0> translate < %f, %f, %f> }\n",newnam);
			//} else {
              fprintf(outfile,"  texture  { MAT_%s }\n",newnam);
			//}
            fprintf(outfile,"}\n\n");
		  }
          break;
        }
      }
      Prim=Prim->succ;
    }
    fclose(outfile);
    TextWin.Output(1,"\n  done.\n");
  } else TextWin.Output(3,"- Datei konnte nicht geoeffnet werden!\n");
}


void CVModelDoc::ExportDXF(char *filename, ViewPtr V)
{
  FILE   *outfile;

  outfile = fopen(filename,"wt");
  if (outfile){
    TextWin.Output(1,"- Writing DXF-File-Header ...");
    fprintf(outfile,"  0\nSECTION\n  2\nENTITIES\n");
    TextWin.Output(1,"\n  Objects ");
    PrimPtr Prim = VMF.Base;
    while(Prim){
      TextWin.Output(1,".");
      switch (Prim->idnt) {
        case ROOT     : break;
        case POLYMESH : {
		  int k,ii,i;
          PolyMesh *p = (PolyMesh *)Prim;
          fprintf(outfile,"  0\nPOLYLINE\n  8\n0\n 66\n     1\n 10\n0.0\n 20\n0,0\n 30\n0.0\n");
          fprintf(outfile," 70\n    64\n 71\n%6d\n 72\n%6d\n", p->nv, p->nf);
          for(i=0; i<(p->nv); i++){
            fprintf(outfile,"  0\nVERTEX\n  8\n0\n 10\n%f\n 20\n%f\n 30\n%f\n 70\n   192\n", p->pts[i].x, p->pts[i].y, p->pts[i].z);
		  }
          for(ii=0,i=0; i<(p->nf); i++){
            fprintf(outfile,"  0\nVERTEX\n  8\n0\n 10\n0.0\n 20\n0.0\n 30\n0.0\n 70\n   128\n");
		    k=(p->pnv)? p->pnv[i] : 3;
			if(k==4){
              fprintf(outfile," 71\n%6d\n 72\n%6d\n 73\n%6d\n 74\n%6d\n", p->pfv[ii]+1, p->pfv[ii+1]+1, p->pfv[ii+2]+1, p->pfv[ii+3]+1);
			} else {
              fprintf(outfile," 71\n%6d\n 72\n%6d\n 73\n%6d\n", p->pfv[ii]+1, p->pfv[ii+1]+1, p->pfv[ii+2]+1);
			}
			ii+=k;
		  }
          fprintf(outfile,"  0\nSEQEND\n");
          break;
        }
      }
      Prim=Prim->succ;
    }
    fprintf(outfile,"  0\nENDSEC\n  0\nEOF\n");
    fclose(outfile);
    TextWin.Output(1,"\n  done.\n");
  } else TextWin.Output(3,"- Datei konnte nicht geffnet werden!\n");
}

void OutputRIBTranss(Surface *S, FILE *outfile){
  vector scl; angles rot; vector trn;
  if(RFTrans2(&(S->TM), &scl, &rot, &trn)<0.01f){
	if((fabs(trn.x)>0.0001f) || (fabs(trn.y)>0.0001f) || (fabs(trn.z)>0.0001f)){
      fprintf(outfile,"  Translate %f %f %f\n", trn.x, trn.z, trn.y);
	}
	if(fabs(rot.z)>0.001f) fprintf(outfile,"  Rotate    %f 0 1 0\n",Rad2Deg*(rot.z));
	if(fabs(rot.x)>0.001f) fprintf(outfile,"  Rotate    %f 1 0 0\n",Rad2Deg*(rot.x));
	if(fabs(rot.y)>0.001f) fprintf(outfile,"  Rotate    %f 0 0 1\n",Rad2Deg*(rot.y));
    if((fabs(scl.x-1.0f)>0.0001f) || (fabs(scl.y-1.0f)>0.0001f) || (fabs(scl.z-1.0f)>0.0001f)){
      fprintf(outfile,"  Scale     %f %f %f\n", scl.x, scl.z, scl.y);
	}
  }
}

void OutputRIBMat(MatPtr mp, FILE *outfile){
  if(mp->Specularity<0.01){
    fprintf(outfile,"  Surface   \"matte\"\n");
  } else {
    if((mp->Reflection.R<0.0f) || (mp->Transmission.R>0.01f)){
      fprintf(outfile,"  Declare   \"eta\" \"float\"\n");
      fprintf(outfile,"  Surface   \"glass\" \"eta\" %f\n", mp->IOR(555.0f));
	} else {
      fprintf(outfile,"  Surface   \"shiny\" \"Kd\" %f \"Ks\" %f \"Kr\" %f\n", 1.0f-mp->Specularity, mp->Specularity, mp->Specularity);
	}
  }
  if((mp->Reflection.R>=0.0f) && (mp->Transmission.R<0.01f)){
    if(mp->Reflection.G<0.0f){
      fprintf(outfile,"  Color     [%f %f %f]\n",mp->Reflection.R, mp->Reflection.R, mp->Reflection.R);
	} else {
      fprintf(outfile,"  Color     [%f %f %f]\n",mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
	}
  }
}

void CVModelDoc::ExportRIB(char *filename, ViewPtr V)
{
  FILE   *outfile;
  int     lights=0;

  outfile = fopen(filename,"wt");
  if (outfile){
    TextWin.Output(1,"- writing RIB-file header ...");
    fprintf(outfile,"##RenderMan RIB-Structure 1.0\nversion 3.03\n");
    fprintf(outfile,"\n");
    fprintf(outfile,"# converted by VModel V0.6 (C)2002 Michael Granz\n\n");

    fprintf(outfile,"Display \"vmd.tif\" \"file\" \"rgba\"\nFormat 480 360 -1\nPixelSamples 2 2\n");
    fprintf(outfile,"Projection \"perspective\" \"fov\" [%f]\n",(float)(Rad2Deg*2.0*atan(V->h/2.0)));
    vector D = V->Target - V->Location;
	fprintf(outfile,"Translate %f %f %f\n", D.x, D.z, D.y);
	fprintf(outfile,"Rotate %f 0 1 0\n\n", Rad2Deg*atan(D.x/-D.y));

    fprintf(outfile,"Option \"radiosity\" \"steps\" [100]\n\n");
    fprintf(outfile,"WorldBegin\n\n");
    fprintf(outfile,"Declare \"patchsize\" \"float\"\nDeclare  \"elemsize\" \"float\"\nDeclare   \"minsize\" \"float\"\nAttribute \"radiosity\" \"patchsize\" [1] \"elemsize\" [0.5] \"minsize\" [0.1]\nAttribute \"radiosity\" \"has_caustics\" 1\n\n");

    PrimPtr Prim = VMF.Base;
    while(Prim){
      if(Prim->idnt==POLYMESH){
        int fi,i,j,k; PolyMesh *p = (PolyMesh *)Prim;
		if(p->Mat->Emission.R>0.01f){
          fprintf(outfile,"AttributeBegin\n  Surface \"constant\"\n  Attribute \"light\" \"shadows\" \"on\"\n  Attribute \"light\" \"nsamples\" [4]\n");
          fprintf(outfile,"  AreaLightSource \"arealight\" 1 \"intensity\" %f\n",p->Mat->Emission.R);
          for(fi=0,i=0; i<(p->nf); i++){
		    k=(p->pnv)? p->pnv[i] : 3;
		    fprintf(outfile,"  Polygon \"P\" [");
			for(j=0; j<k; j++){
              fprintf(outfile," %f %f %f",p->pts[p->pfv[fi+j]].x, p->pts[p->pfv[fi+j]].z, p->pts[p->pfv[fi+j]].y);
			}
			fprintf(outfile,"]\n"); fi+=k;
		  }
          fprintf(outfile,"AttributeEnd\nIlluminate 1 1\n\n");
		  lights++;
        }
      }
      Prim=Prim->succ;
    }

	/*if(!lights){
	  float  d = Betrag(V->Target - V->Location);
      vector N = unit  (V->Target - V->Location);
      vector U = N*V->Up;
      vector L = V->Location - d*N + d*U + d*V->Up;
      fprintf(outfile,"light_source { < %f, %f, %f > color red 1.0 green 1.0 blue 1.0 }\n\n",L.x,L.y,L.z);
	}*/

    TextWin.Output(1,"\n          Objects "); Prim = VMF.Base;
    while(Prim){
      TextWin.Output(1,".");
      switch (Prim->idnt) {
        case ROOT     : break;
        case SPHERE   : {
          fprintf(outfile,"AttributeBegin\n  Attribute \"identifier\" \"name\" \"%s\"\n",((Surface *)Prim)->name);
	      OutputRIBMat(((Surface *)Prim)->Mat,outfile);
          OutputRIBTranss((Surface *)Prim, outfile);
          fprintf(outfile,"  Sphere 1 -1 1 360\nAttributeEnd\n\n");
          break;
        }
        case CYLINDER : {
          fprintf(outfile,"AttributeBegin\n  Attribute \"identifier\" \"name\" \"%s\"\n",((Surface *)Prim)->name);
	      OutputRIBMat(((Surface *)Prim)->Mat,outfile);
          OutputRIBTranss((Surface *)Prim, outfile);
          fprintf(outfile,"  Rotate   -90 1 0 0\n");
          fprintf(outfile,"  Cylinder     1 0 1 360\nAttributeEnd\n\n");
          break;
        }
        case TORUS    : {
          fprintf(outfile,"AttributeBegin\n  Attribute \"identifier\" \"name\" \"%s\"\n",((Surface *)Prim)->name);
	      OutputRIBMat(((Surface *)Prim)->Mat,outfile);
          OutputRIBTranss((Surface *)Prim, outfile);
          fprintf(outfile,"  Rotate   -90 1 0 0\n");
          fprintf(outfile,"  Torus %f %f 0 90 360\nAttributeEnd\n\n", ((Torus *)Prim)->R1, ((Torus *)Prim)->R2);
          break;
        }
        case POLYMESH : {
          int      fi,i,j,k;
          vector   p0,n0;
          PolyMesh *p = (PolyMesh *)Prim;
		  if(p->Mat->Emission.R<0.01f){
            fprintf(outfile,"AttributeBegin\n  Attribute \"identifier\" \"name\" \"%s\"\n",p->name);
		    if(p->nm > 1){
			} else {
			  OutputRIBMat(p->Mat,outfile);
              for(fi=0,i=0; i<(p->nf); i++){
			    k=(p->pnv)? p->pnv[i] : 3;
			    fprintf(outfile,"  Polygon \"P\" [");
			    for(j=0; j<k; j++){
                  fprintf(outfile," %f %f %f",p->pts[p->pfv[fi+j]].x, p->pts[p->pfv[fi+j]].z, p->pts[p->pfv[fi+j]].y);
				}
			    fprintf(outfile,"]\n"); fi+=k;
			  }
              fprintf(outfile,"AttributeEnd\n\n");
			}
		  }
          break;
        }
      }
      Prim=Prim->succ;
    }
    fprintf(outfile,"WorldEnd\n");
    fclose(outfile);
    TextWin.Output(1,"\n  done.\n");
  } else TextWin.Output(3,"- can't open file!\n");
}


void OutputVRMLTranss(Surface *S, FILE *outfile){
  vector scl; angles rot; vector trn;
  matrix trf;
  if(RFTrans2(&(S->TM), &scl, &rot, &trn)<0.01f){
    trf.InitTransformation(scl.x, scl.z, scl.y, rot.x, rot.z, rot.y, -trn.x, trn.z, trn.y);
	fprintf(outfile,"    MatrixTransform {\n");
    fprintf(outfile,"      matrix %f %f %f 0.0\n",trf.M11, trf.M12, trf.M13);
    fprintf(outfile,"             %f %f %f 0.0\n",trf.M21, trf.M22, trf.M23);
    fprintf(outfile,"             %f %f %f 0.0\n",trf.M31, trf.M32, trf.M33);
    fprintf(outfile,"             %f %f %f 1.0\n",trf.M41, trf.M42, trf.M43);
    fprintf(outfile,"    }\n");
  }
}

void OutputVRMLMat(MatPtr mp, MatPtr Env, FILE *outfile){
  fprintf(outfile,"    Material {\n");
  if(mp->Reflection.G<0.0f){
    fprintf(outfile,"      diffuseColor  %f %f %f\n", mp->Reflection.R, mp->Reflection.R, mp->Reflection.R);
  } else {
    /*if(mp->Reflection.R<0.0f){
      fprintf(outfile,"  pigment { color red 1.0 green 1.0 blue 1.0 filter %f }\n",mp->Trns(555.0f, mp->IOR(555.0f), VMDoc->Environment.IOR(555.0f), 1.0f));
	} else {
	  if(mp->Texture){
        fprintf(outfile,"  pigment { image_map { bmp \"%s\" } }\n", mp->Texture->tname);
	  } else {*/
        fprintf(outfile,"      diffuseColor  %f %f %f\n", mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
	  //}
	//}
  }
  fprintf(outfile,"      emissiveColor %f %f %f\n", mp->Emission.R, mp->Emission.R, mp->Emission.R);
  fprintf(outfile,"      transparency  %f\n",mp->Trns(555.0f, mp->IOR(555.0f), Env->IOR(555.0f), 1.0f));
  fprintf(outfile,"    }\n");
  if(mp->Texture){
    fprintf(outfile,"    Texture2 {\n");
    fprintf(outfile,"      filename \"%s\"\n",mp->Texture->tname);
    fprintf(outfile,"    }\n");
  }
}

void CVModelDoc::ExportVRML(char *filename, ViewPtr V)
{
  FILE   *outfile;
  PrimPtr Prim;
  int     lights=0;

  outfile = fopen(filename,"wt");
  if (outfile){
    TextWin.Output(1,"- writing VRML header ...");
    fprintf(outfile,"#VRML V1.0 ascii\n");
    fprintf(outfile,"#\n");
    fprintf(outfile,"#converted by VModel V0.6 (C)2002 Michael Granz\n\n");
    vector v1   = vector(-V->Target.x, V->Target.z, V->Target.y)-vector(-V->Location.x, V->Location.z, V->Location.y);
    vector v2   = vector(0.0f, 0.0f, 1.0f);
	vector axis = unit(v1*v2);
	double ang  = acos(unit(v1)%unit(v2));
	fprintf(outfile,"Separator {\n");
    fprintf(outfile,"  PerspectiveCamera {\n");
    fprintf(outfile,"    position     %f %f %f\n", -V->Location.x, V->Location.z, V->Location.y);
    fprintf(outfile,"    orientation  %f %f %f %f\n", axis.x, axis.y, axis.z, ang);
    fprintf(outfile,"    heightAngle  %f\n",2.0*atan(V->h/2.0));
    fprintf(outfile,"  }\n\n");

    /*fprintf(outfile,"background { color red %f green %f blue %f }\n\n",Background.R,Background.G,Background.B);
    */
    /*Prim = Base;
    while(Prim){
      if(Prim->idnt==POLYMESH){
        PolyMesh *p = (PolyMesh *)Prim;
		if(p->Mat->Emission.R>0.01f){
		  vector base=p->pts[p->pfv[0]];
		  vector p1=p->pts[p->pfv[1]]-base;
		  vector p2=p->pts[p->pfv[(p->pnv)? p->pnv[0]-1 : 2]]-base;
          vector L = base+0.5f*p1+0.5f*p2-0.01f*unit(p1*p2);
		  fprintf(outfile,"light_source {\n  < %f, %f, %f > color red 1.0 green 1.0 blue 1.0\n",L.x,L.y,L.z);
          fprintf(outfile,"  //area_light <%f, %f, %f>, <%f, %f ,%f>, 5, 5\n",p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);
          fprintf(outfile,"  //adaptive 1 jitter\n}\n\n");
		  lights++;
        }
      }
      Prim=Prim->succ;
    }*/

	/*if(!lights){
	  float  d = Betrag(V->Target - V->Location);
      vector N = unit  (V->Target - V->Location);
      vector U = N*V->Up;
      vector L = V->Location - d*N + d*U + d*V->Up;
      fprintf(outfile,"light_source { < %f, %f, %f > color red 1.0 green 1.0 blue 1.0 }\n\n",L.x,L.y,L.z);
	}*/

    /*TextWin.Output(1,"\n          Materials ");
    MatPtr mp = MatList;
    while(mp){
      TextWin.Output(1,".");
	  povnam(newnam,mp->name);
      fprintf(outfile,"#declare MAT_%s = texture {\n",newnam);
      if(mp->Reflection.G<0.0f){
        fprintf(outfile,"  pigment { color red %f green %f blue %f }\n", mp->Reflection.R, mp->Reflection.R, mp->Reflection.R);
	  } else {
        if(mp->Reflection.R<0.0f){
          fprintf(outfile,"  pigment { color red 1.0 green 1.0 blue 1.0 filter %f }\n",mp->Trns(555.0f, mp->IOR(555.0f), VMDoc->Environment.IOR(555.0f), 1.0f));
		} else {
		  if(mp->Texture){
            fprintf(outfile,"  pigment { image_map { bmp \"%s\" } }\n", mp->Texture->tname);
		  } else {
            fprintf(outfile,"  pigment { color red %f green %f blue %f }\n", mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
		  }
		}
	  }
	  fprintf(outfile,"  finish  {\n");
      fprintf(outfile,"    ambient           %f\n",(Ambient.R+Ambient.G+Ambient.B)/3.0F);
      fprintf(outfile,"    diffuse           %f\n",1.0f-mp->Specularity);
      fprintf(outfile,"    phong             %f phong_size %d\n",mp->Specularity,(int)(mp->PBExp? mp->PBExp : 50.0f));
      if(mp->Specularity>=0.3f) fprintf(outfile,"    reflection        %f\n",mp->Spec(555.0f, mp->IOR(555.0f), VMDoc->Environment.IOR(555.0f), 1.0f));
      if((mp->Reflection.R < 0.0f) || (mp->Transmission.R > 0.001)) fprintf(outfile,"    refraction on ior %f\n",mp->IOR(555.0f));
      fprintf(outfile,"  }\n");
      fprintf(outfile,"}\n\n");
      mp = mp->next;
    }*/
    TextWin.Output(1,"\n          Objects ");
    Prim = VMF.Base;
    while(Prim){
      TextWin.Output(1,".");
      switch (Prim->idnt) {
        case ROOT     : break;
        case CUBE     : {
          fprintf(outfile,"  Separator {\n");
          if(((Surface *)Prim)->name){
	        fprintf(outfile,"    Info {\n");
            fprintf(outfile,"      string \"%s\"\n",((Surface *)Prim)->name);
            fprintf(outfile,"    }\n");
		  }
          OutputVRMLMat(((Surface *)Prim)->Mat, &(VMF.Environment), outfile);
          OutputVRMLTranss((Surface *)Prim, outfile);
          fprintf(outfile,"    Cube {}\n");
          fprintf(outfile,"  }\n\n");
          break;
        }
        case SPHERE   : {
          fprintf(outfile,"  Separator {\n");
          if(((Surface *)Prim)->name){
	        fprintf(outfile,"    Info {\n");
            fprintf(outfile,"      string \"%s\"\n",((Surface *)Prim)->name);
            fprintf(outfile,"    }\n");
		  }
          OutputVRMLMat(((Surface *)Prim)->Mat, &(VMF.Environment), outfile);
          OutputVRMLTranss((Surface *)Prim, outfile);
          fprintf(outfile,"    Sphere {\n      radius 1.0\n    }\n");
          fprintf(outfile,"  }\n\n");
          break;
        }
        case CYLINDER : {
          fprintf(outfile,"  Separator {\n");
          if(((Surface *)Prim)->name){
	        fprintf(outfile,"    Info {\n");
            fprintf(outfile,"      string \"%s\"\n",((Surface *)Prim)->name);
            fprintf(outfile,"    }\n");
		  }
          OutputVRMLMat(((Surface *)Prim)->Mat, &(VMF.Environment), outfile);
          matrix H,B;
          H.M11=  1.0f; H.M12=  0.0f; H.M13=  0.0f;
          H.M21=  0.0f; H.M22=  1.0f; H.M23=  0.0f;
          H.M31=  0.0f; H.M32=  0.0f; H.M33=  1.0f;
          H.M41=  0.0f; H.M42=  0.0f; H.M43=  0.5f;
		  B=((Surface *)Prim)->TM; ((Surface *)Prim)->TM=H*((Surface *)Prim)->TM;
          OutputVRMLTranss((Surface *)Prim, outfile);
		  ((Surface *)Prim)->TM=B;
          fprintf(outfile,"    Cylinder {\n      parts  SIDES\n      radius 1.0\n      height 1.0\n    }\n");
          fprintf(outfile,"  }\n\n");
          break;
        }
        /*case TORUS    : {
          fprintf(outfile,"torus {\n  %f, %f\n", ((Torus *)Prim)->R1, ((Torus *)Prim)->R2);
          OutputTransformations((Surface *)Prim, outfile);
          fprintf(outfile,"  texture { MAT_%s }\n",((Surface *)Prim)->Mat->name);
          fprintf(outfile,"}\n\n");
          break;
        }*/
        case POLYMESH : {
          int      fi,i,j,k;
          vector   p0,n0;
          PolyMesh *p = (PolyMesh *)Prim;
          fprintf(outfile,"  Separator {\n");
          if(p->name){
	        fprintf(outfile,"    Info {\n");
            fprintf(outfile,"      string \"%s\"\n",p->name);
            fprintf(outfile,"    }\n");
		  }
          OutputVRMLMat(p->Mat, &(VMF.Environment), outfile);
          fprintf(outfile,"    Coordinate3 {\n");
		  for(i=0; i<(p->nv); i++){
			fprintf(outfile,i==0? "      point [ " : "              ");
			fprintf(outfile,"%f %f %f", -p->pts[i].x, p->pts[i].z, p->pts[i].y);
			fprintf(outfile,i==(p->nv-1)? " ]\n" : ",\n");
		  }
          fprintf(outfile,"    }\n");
		  if((p->txc) && (p->Mat->Texture)){
            fprintf(outfile,"    TextureCoordinate2 {\n");
		    for(i=0; i<(p->nt); i++){
			  fprintf(outfile,i==0? "      point [ " : "              ");
			  fprintf(outfile,"%f %f", p->txc[i].x, p->txc[i].y);
			  fprintf(outfile,i==(p->nt-1)? " ]\n" : ",\n");
			}
            fprintf(outfile,"    }\n");
		  }
		  if(p->nm > 1){
          } else {
		    fprintf(outfile,"    IndexedFaceSet {\n");
            for(fi=0,i=0; i<(p->nf); i++){
			  k=(p->pnv)? p->pnv[i] : 3;
			  fprintf(outfile,i==0? "      coordIndex        [ " : "                          ");
			  for(j=0; j<k; j++){
                fprintf(outfile,"%d",p->pfv[fi+(k-1-j)]);
			    fprintf(outfile,j==(k-1)? ", -1" : ", ");
			  }
			  fi+=k;
			  fprintf(outfile,i==(p->nf-1)? " ]\n" : ",\n");
			}
		    if((p->txc) && (p->Mat->Texture) && (p->pft)){
              for(fi=0,i=0; i<(p->nf); i++){
			    k=(p->pnv)? p->pnv[i] : 3;
			    fprintf(outfile,i==0? "      textureCoordIndex [ " : "                          ");
			    for(j=0; j<k; j++){
                  fprintf(outfile,"%d",p->pft[fi+(k-1-j)]);
			      fprintf(outfile,j==(k-1)? ", -1" : ", ");
				}
			    fi+=k;
			    fprintf(outfile,i==(p->nf-1)? " ]\n" : ",\n");
			  }
			}
            fprintf(outfile,"    }\n");
		  }
          fprintf(outfile,"  }\n\n");
          break;
        }
      }
      Prim=Prim->succ;
    }
    fprintf(outfile,"}\n"); fclose(outfile);
    TextWin.Output(1,"\n  done.\n");
  } else TextWin.Output(3,"- can't open file!\n");
}

void CVModelDoc::ExportVIB(char *filename, ViewPtr V)
{
  FILE   *outfile;
  int     lights=0;
  char    newnam[64];
  PrimPtr Prim;

  outfile = fopen(filename,"wt");
  if (outfile){
    TextWin.Output(1,"- writing VirtuaLight file header ...");
    fprintf(outfile,"// VirtuaLight VIB-Structure\n");
    fprintf(outfile,"//\n");
    fprintf(outfile,"// converted by VModel V0.60\n\n");
    fprintf(outfile,"ReadArchive \"statics.vib\"\n\n");

    TextWin.Output(1,"\n          Materials ");
    MatPtr mp = VMF.MatList;
    while(mp){
      TextWin.Output(1,".");
	  povnam(newnam,mp->name);
      fprintf(outfile,"Declare MAT_%s = Shader [ %s [ ", newnam, (mp->Texture)? "FunctionalSurface" : "PlainSurface");
      if(mp->Reflection.G<0.0f){
        fprintf(outfile,"Color '%f,%f,%f' ", mp->Reflection.R, mp->Reflection.R, mp->Reflection.R);
	  } else {
        if(mp->Reflection.R<0.0f){
          fprintf(outfile,"Color '1.0,1.0,1.0' ");
		} else {
		  if(mp->Texture){
			fprintf(outfile,"Color UVImageMapping(ImageFile(\"%s\", BILINEAR)) ",mp->Texture->tname);
		  } else {
            fprintf(outfile,"Color '%f,%f,%f' ", mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
		  }
		}
	  }
      fprintf(outfile,"Kd %f ",1.0f-mp->Specularity);
      if(mp->Specularity>=0.2f){
        if((mp->Reflection.R < 0.0f) || (mp->Transmission.R > 0.001)){
		  fprintf(outfile,"Ks 0 Kr %f ",mp->Spec(555.0f, mp->IOR(555.0f), VMF.Environment.IOR(555.0f), 1.0f));
		  fprintf(outfile,"Kt (%f, %f) ",mp->Trns(555.0f, mp->IOR(555.0f), VMF.Environment.IOR(555.0f), 1.0f), mp->IOR(555.0f));
		} else {
		  fprintf(outfile,"Ks 0 Kr %f ",mp->Specularity);
		}
		fprintf(outfile,"Caustics 0.5 ");
	  } else {
		fprintf(outfile,"Ks %f ",mp->Specularity);
	  }
      fprintf(outfile,"] ]\n\n");
      mp = mp->next;
    }

    fprintf(outfile,"Camera [ (%f, %f, %f), (%f, %f, %f), (%f, %f, %f) FieldOfView %f Format(480,360) FrameAspectRatio -(480/360) ]\n\n", -V->Location.x, V->Location.z, V->Location.y, -V->Target.x, V->Target.z, V->Target.y, -V->Up.x, V->Up.z, V->Up.y, (float)(Rad2Deg*2.0*atan(V->h/2.0)));

    fprintf(outfile,"Options [ IrradianceSamples 100 IrradianceBounceLevel 3 IrradianceMaxError 0.6 ]\n\nComputeIrradiance 16\n\nPhotons(30000, 0, 0)\n\n");

    Prim = VMF.Base;
    while(Prim){
      if(Prim->idnt==POLYMESH){
        PolyMesh *p = (PolyMesh *)Prim;
		if(p->Mat->Emission.R>0.01f){
		  vector base=p->pts[p->pfv[0]];
		  vector p1=p->pts[p->pfv[1]]-base;
		  vector p2=p->pts[p->pfv[(p->pnv)? p->pnv[0]-1 : 2]]-base;
          vector L = base+0.5f*p1+0.5f*p2-0.01f*unit(p1*p2);
          fprintf(outfile,"AreaLight [ Plane(6, 6, 0.1) Scale (%f, 1, %f) Translate (%f,%f,%f) Intensity %f Decay QUADRATIC ]\n\n",Betrag(p1)/6.0f,Betrag(p2)/6.0f,-L.x,L.z,L.y,p->Mat->Emission.R);
		  lights++;
        }
      }
      Prim=Prim->succ;
    }

	/*if(!lights){
	  float  d = Betrag(V->Target - V->Location);
      vector N = unit  (V->Target - V->Location);
      vector U = N*V->Up;
      vector L = V->Location - d*N + d*U + d*V->Up;
      fprintf(outfile,"light_source { < %f, %f, %f > color red 1.0 green 1.0 blue 1.0 }\n\n",L.x,L.y,L.z);
	}*/

    TextWin.Output(1,"\n          Objects "); Prim = VMF.Base;
    while(Prim){
      TextWin.Output(1,".");
      switch (Prim->idnt) {
        case ROOT     : break;
        case SPHERE   : {
          vector scl; angles rot; vector trn;
          if(RFTrans2(&(((Surface *)Prim)->TM), &scl, &rot, &trn)<0.01f){
	        povnam(newnam,((Surface *)Prim)->Mat->name);
            fprintf(outfile,"// %s\n",((Surface *)Prim)->name);
            fprintf(outfile,"Shape [ Sphere((%f, %f, %f), %f) MAT_%s ]\n", -trn.x, trn.z, trn.y, scl.x, newnam);
		  }
          break;
        }
        case CYLINDER : {
          vector scl; angles rot; vector trn;
          if(RFTrans2(&(((Surface *)Prim)->TM), &scl, &rot, &trn)<0.01f){
	        povnam(newnam,((Surface *)Prim)->Mat->name);
            fprintf(outfile,"// %s\n",((Surface *)Prim)->name);
			vector p2=vector(0.0f, 0.0f, 1.0f)*((Surface *)Prim)->TM;
            fprintf(outfile,"Shape [ Cylinder((%f, %f, %f), (%f, %f, %f), %f) MAT_%s ]\n", -trn.x, trn.z, trn.y, -p2.x, p2.z, p2.y, scl.x, newnam);
		  }
          break;
        }
        case TORUS    : {
          vector scl; angles rot; vector trn;
          if(RFTrans2(&(((Surface *)Prim)->TM), &scl, &rot, &trn)<0.01f){
	        povnam(newnam,((Surface *)Prim)->Mat->name);
            fprintf(outfile,"// %s\n",((Surface *)Prim)->name);
			vector p2=unit(vector(0.0f, 0.0f, 1.0f)*((Surface *)Prim)->NM);
            fprintf(outfile,"Shape [ Torus((%f, %f, %f), (%f, %f, %f), %f, %f) MAT_%s ]\n", -trn.x, trn.z, trn.y, -p2.x, p2.z, p2.y, ((Torus *)Prim)->R1, ((Torus *)Prim)->R2, newnam);
		  }
          break;
        }
        case POLYMESH : {
          int      fi,i,j,k;
          vector   p0,n0;
          PolyMesh *p = (PolyMesh *)Prim;
          fprintf(outfile,"// %s\n",p->name);
		  if(p->nm > 1){
		  } else {
	        povnam(newnam,p->Mat->name);
			for(fi=0,i=0; i<(p->nf); i++){
		      k=(p->pnv)? p->pnv[i] : 3;
              if((p->txc) && (p->Mat->Texture)){
			    fprintf(outfile,"Shape [ TriangularPatch(");
			    for(j=0; j<k; j++){
                  fprintf(outfile,"(%f, %f, %f)",-p->pts[p->pfv[fi+j]].x, p->pts[p->pfv[fi+j]].z, p->pts[p->pfv[fi+j]].y);
                  if(p->nn){
					fprintf(outfile,", (%f, %f, %f)",-p->nrm[p->pfn[fi+j]].x, p->nrm[p->pfn[fi+j]].z, p->nrm[p->pfn[fi+j]].y);
				  } else {
					vector N = unit((p->pts[p->pfv[fi+2]] - p->pts[p->pfv[fi]]) * (p->pts[p->pfv[fi+1]] - p->pts[p->pfv[fi]]));
					fprintf(outfile,", (%f, %f, %f)",-N.x, N.z, N.y);
				  }
				  fprintf(outfile," UV=%f,%f", p->txc[p->pft[fi+j]].x, p->txc[p->pft[fi+j]].y);
			      if(j<(k-1)) fprintf(outfile,", ");
				}
			  } else {
			    fprintf(outfile,"Shape [ Polygon(%d, ",k);
			    for(j=0; j<k; j++){
                  fprintf(outfile,"(%f, %f, %f)",-p->pts[p->pfv[fi+j]].x, p->pts[p->pfv[fi+j]].z, p->pts[p->pfv[fi+j]].y);
			      if(j<(k-1)) fprintf(outfile,", ");
				}
			  }
			  fprintf(outfile,") MAT_%s ]\n",newnam); fi+=k;
			}
		  }
          break;
        }
      }
      Prim=Prim->succ;
    }
    fclose(outfile);
    TextWin.Output(1,"\n  done.\n");
  } else TextWin.Output(3,"- can't open file!\n");
}
