
#include "stdafx.h"
#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "GraphWin.h"

#include "gl\gl.h"
#include "gl\glu.h"

#include "RaVi.h"

extern "C" {
  #include "OpenImage.h"
}

extern char filename[320];

extern float wlcoff(float cR, float cG, float cB, float v);
extern float wavlen;

int         UseTexture;
TexMapPtr   ltexture;


void glError(GLenum Ern, char *Text, ...)
{
  if(Ern!=GL_NO_ERROR){
    char buf1[128];
	char buf2[128];
    // convert format-string with variable-length argument list to string
    va_list vl; va_start(vl, Text); vsprintf(buf1,Text,vl); va_end(vl);
	sprintf(buf2,"Error '%s' after %s", gluErrorString(Ern), buf1);
	MessageBox(NULL, buf2, "OpenGL Error", MB_OK);
  }
}


// Utility to scale the number x to the next lower power of 2.
// For example, 300 would get scaled to 256...
int scale_down_to_power_of_2(long int x)
{
  int i;
  long int shifted_x;

  shifted_x = x;
  for(i = 0; ((i < 16) && (shifted_x != 0)); i++) shifted_x = shifted_x >> 1;

  if(i == 0)
    return 0;
  else
    return ((int)(pow(2.0, (double)(i-1))));
}


void SetTexture(TexMapPtr map, int TexMode, int refmap)
{
  if(map != ltexture){
    if(!(map->tbits)){
      char texpath[260];
	  if(strchr(map->tname, '\\')){
	    strcpy(texpath, map->tname);
	  } else {
	    strcpy(texpath, filename);
        char *c = strrchr(texpath, '\\'); if(!c) c = strrchr(texpath, ':'); c = c? c+1 : texpath;
        strcpy(c, map->tname);
	  }
	  map->tbits=LoadImageFile(texpath, &(map->tinfo));
      if(map->tbits){
        int maxtxsiz; glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxtxsiz);
	    int newwidth = min(maxtxsiz, scale_down_to_power_of_2(map->tinfo.bmiHeader.biWidth ) );
	    int newheigh = min(maxtxsiz, scale_down_to_power_of_2(map->tinfo.bmiHeader.biHeight) );
        int ScLen = ((newwidth*3+3) & 0xFFFFFFFC);
	    map->glbits = new(byte[newheigh*ScLen]);
		map->glwidth     = newwidth;
		map->glheight    = newheigh;
        if((map->tinfo.bmiHeader.biWidth != newwidth) || (map->tinfo.bmiHeader.biHeight != newheigh)){
		  gluScaleImage(GL_BGR_EXT, map->tinfo.bmiHeader.biWidth, map->tinfo.bmiHeader.biHeight, GL_UNSIGNED_BYTE, map->tbits, newwidth, newheigh, GL_UNSIGNED_BYTE, map->glbits);
		  glError(glGetError(),"ScaleImage <%s> (%d->%d),(%d->%d)",map->tname,map->tinfo.bmiHeader.biWidth, newwidth, map->tinfo.bmiHeader.biHeight, newheigh);
        } else memcpy(map->glbits, map->tbits, newheigh*ScLen);
        // convert Windows DIB BGR to OpenGL RGB
		ushort x,y; byte t; byte *bitptr = map->glbits;
		for(y=0; y<newheigh; y++){
		  for(x=0; x<newwidth; x++){
		    t=bitptr[3*x]; bitptr[3*x]=bitptr[3*x+2]; bitptr[3*x+2]=t;
		  }
		  bitptr += ScLen;
		}
		map->glname=OGLT++;
	    glBindTexture (GL_TEXTURE_2D, map->glname); 
        glPixelStorei(GL_UNPACK_ALIGNMENT, 4); //glError(glGetError(),"PixelStore");
		//glError(glGetError(),"TexImage2D (%d,%d)", map->tinfo.bmiHeader.biWidth, map->tinfo.bmiHeader.biHeight);

        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, TexMode);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0, 3, map->glwidth, map->glheight, 0, GL_RGB, GL_UNSIGNED_BYTE, map->glbits);
        glEnable(GL_TEXTURE_2D); //glError(glGetError(),"Enable Texture2D");

        if(refmap){
		  glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
          glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
          glEnable(GL_TEXTURE_GEN_S);
          glEnable(GL_TEXTURE_GEN_T);
		} else {
          glDisable(GL_TEXTURE_GEN_S);
          glDisable(GL_TEXTURE_GEN_T);
		}
      }// else TextWin.Output(3,"- Texture '%s' not found!\n",map->tname);
	}
	glBindTexture (GL_TEXTURE_2D, map->glname); 
	ltexture = map;
  }
  glEnable (GL_TEXTURE_2D); UseTexture=1;
}

void SetMaterial(MatPtr mat, int hl)
{
  GLfloat mat_diffuse[]    = { 0.7f, 0.6f, 0.5f, 1.0f };
  GLfloat mat_specular[]   = { 1.0f, 1.0f, 1.0f, 1.0f };
  GLfloat mat_emission[]   = { 0.0f, 0.0f, 0.0f, 1.0f };

  mat_specular[0] = mat->Specularity; 
  mat_specular[1] = mat->Specularity; 
  mat_specular[2] = mat->Specularity;
  if(hl){
    mat_emission[0] = 0.50f; 
    mat_emission[1] = 0.50f; 
    mat_emission[2] = 0.25f;
  } else {
    mat_emission[0]  =                                           mat->Emission.R; 
    mat_emission[1]  = (mat->Emission.G<0.0f)? mat->Emission.R : mat->Emission.G; 
    mat_emission[2]  = (mat->Emission.G<0.0f)? mat->Emission.R : mat->Emission.B; 
  }

  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR,  mat_specular);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION,  mat_emission);
  glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, mat->PBExp? mat->PBExp :30.0f);

  if(mat->Texture && (GraphWin.DisplayMode & VM_TEXTURED)){
    mat_diffuse[0] = 1.0f; 
    mat_diffuse[1] = 1.0f; 
    mat_diffuse[2] = 1.0f; 
	SetTexture(mat->Texture, GL_MODULATE, 0);
  } else {
    if(mat->ReflMap && (GraphWin.DisplayMode & VM_TEXTURED)){
      mat_diffuse[0] = 1.0f;
      mat_diffuse[1] = 1.0f;
      mat_diffuse[2] = 1.0f;
	  SetTexture(mat->ReflMap, GL_MODULATE, 1);
	} else {
      glDisable(GL_TEXTURE_2D); UseTexture=0;
	  if(mat->Reflection.R<0.0f){
        mat_diffuse[0]  = 0.2f+mat->Reflection.G; 
        mat_diffuse[1]  = 0.2f+mat->Reflection.G; 
        mat_diffuse[2]  = 0.2f+mat->Reflection.G; 
	  } else {
        mat_diffuse[0]  =                                               mat->Reflection.R; 
        mat_diffuse[1]  = (mat->Reflection.G<0.0f)? mat->Reflection.R : mat->Reflection.G; 
        mat_diffuse[2]  = (mat->Reflection.G<0.0f)? mat->Reflection.R : mat->Reflection.B; 
	  }
	}
  }
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat_diffuse);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat_diffuse);
}


void SetTransform(matrix *TM)
{
  float m[16];
  m[ 0]=TM->M11; m[ 1]=TM->M12; m[ 2]=TM->M13; m[ 3]=0.0f;
  m[ 4]=TM->M21; m[ 5]=TM->M22; m[ 6]=TM->M23; m[ 7]=0.0f;
  m[ 8]=TM->M31; m[ 9]=TM->M32; m[10]=TM->M33; m[11]=0.0f;
  m[12]=TM->M41; m[13]=TM->M42; m[14]=TM->M43; m[15]=1.0f;
  glMatrixMode(GL_MODELVIEW); glMultMatrixf(m);
}


void Surface::SetLine(float x1, float y1, float z1, float x2, float y2, float z2, float n1x, float n1y, float n1z, float n2x, float n2y, float n2z, int edge, int flags, int hl)
{
  vector pm = unit(((GraphWin.GScreen.V.h)? vector((x1+x2)/2.0f, (y1+y2)/2.0f, (z1+z2)/2.0f)*TM : GraphWin.GScreen.V.Target) - GraphWin.GScreen.V.Location);
  float  w1 = pm % unit(((flags & LINE_INVERSE)? vector(-n1x, -n1y, -n1z) : vector(n1x, n1y, n1z))*NM); BOOL eatf=FALSE; BOOL skip=FALSE;
  if((n2x==0.0f) && (n2y==0.0f) && (n2z==0.0f)){
    if( (w1<0.0f) || (flags & LINE_DOUBLESIDED) ) eatf=TRUE;
  } else {
    float  w2 = pm % unit(((flags & LINE_INVERSE)? vector(-n2x, -n2y, -n2z) : vector(n2x, n2y, n2z))*NM);
    if( (w1<0.0f) || (w2<0.0f) || (flags & LINE_DOUBLESIDED) ) eatf=TRUE;
    if(!edge && ( ((w1<0.0f) && (w2<0.0f)) || ((w1>0.0f) && (w2>0.0f)) )) skip=TRUE;
  }
  if(!skip){
	if(hl){
      if(eatf) glColor3f(0.75f, 0.30f, 0.30f); else  glColor3f(0.85f, 0.70f, 0.70f);
	} else {
	  if(flags & LINE_CROSSLINE){
        if(eatf) glColor3f(0.2f, 0.7f, 0.2f); else  glColor3f(0.45f, 0.75f, 0.45f);
	  } else {
	    if(eatf) glColor3f(0.30f, 0.30f, 0.30f); else  glColor3f(0.75f, 0.75f, 0.75f);
	  }
	}
    glVertex3f(x1, y1, z1);
    glVertex3f(x2, y2, z2);
  }
}

void Surface::DrawWire(int hl)
{
  BOOL   eatf,skip;
  vector V1,V2;
  vector pm;
  float  w1,w2;
  int    hl1;

  glDisable(GL_LIGHTING); glDisable(GL_TEXTURE_2D);
  glBegin(GL_LINES);
    for(int i=0; i<mesh.nl; i++){
	  if(!(mesh.pts)){
        V1 = ((PolyMesh *)this)->pts[mesh.lin[i].v1]; V2 = ((PolyMesh *)this)->pts[mesh.lin[i].v2];
	  } else {
        V1 = mesh.pts[mesh.lin[i].v1]; V2 = mesh.pts[mesh.lin[i].v2];
	  }
      pm = unit(((GraphWin.GScreen.V.h)? ((V1+V2)/2.0f)*TM : GraphWin.GScreen.V.Target) - GraphWin.GScreen.V.Location);
      w1 = pm % unit(mesh.snm[mesh.lin[i].n1]*NM); eatf=FALSE; skip=FALSE;
      if(mesh.lin[i].n2 == NO_NORMAL){
        if( (w1<0.0f) || (flags & SURF_DOUBLESIDED) ) eatf=TRUE;
	  } else {
        w2 = pm % unit(mesh.snm[mesh.lin[i].n2]*NM);
        if( (w1<0.0f) || (w2<0.0f) || (mesh.lin[i].flags & SURF_DOUBLESIDED) ) eatf=TRUE;
        if(!(mesh.lin[i].flags & 64) && ( ((w1<0.0f) && (w2<0.0f)) || ((w1>0.0f) && (w2>0.0f)) )) skip=TRUE;
	  }
      if(!skip){
	    if(hl==2){
		  if(((PolyMesh *)this)->flg && ((((PolyMesh *)this)->flg[mesh.lin[i].n1] & SELECTED) || ((mesh.lin[i].n2 != NO_NORMAL) && (((PolyMesh *)this)->flg[mesh.lin[i].n2] & SELECTED)))) hl1=1; else hl1=0;
		} else hl1=hl;
	    if(hl1){
          if(eatf) glColor3f(0.75f, 0.30f, 0.30f); else  glColor3f(0.85f, 0.70f, 0.70f);
		} else {
	      if(mesh.lin[i].flags & LINE_CROSSLINE){
            if(eatf) glColor3f(0.2f, 0.7f, 0.2f); else  glColor3f(0.45f, 0.75f, 0.45f);
		  } else {
	        if(eatf) glColor3f(0.30f, 0.30f, 0.30f); else  glColor3f(0.75f, 0.75f, 0.75f);
		  }
		}
        glVertex3f(V1.x, V1.y, V1.z);
        glVertex3f(V2.x, V2.y, V2.z);
	  }
	}
  glEnd();
}

typedef struct LNode *LineList;

struct LNode  { LineList left,right; ushort p1,p2,n1,n2,i,c; byte flags; };

int    LL_Count2;     // element-counter for line-lists

LineList LL_Add2(LineList l, ushort v1, ushort v2, ushort n, byte flags)
{
  ushort c=v1^v2;

  if(!l){
    l=new LNode;
    l->right = NULL; l->left  = NULL;
    l->p1    = v1;   l->p2    = v2;
    l->n1    = n;    l->n2    = NO_NORMAL;
    l->c     = c;    l->i     = LL_Count2;
    l->flags = flags;
    LL_Count2++;
    return l;
  }
  if((l->c)==c){
    if( (((l->p1)==v1) && ((l->p2)==v2)) || (((l->p1)==v2) && ((l->p2)==v1)) ){
      l->flags |= flags;
      if(l->n2==NO_NORMAL) l->n2=n; else l->flags |= (LINE_DOUBLESIDED | LINE_CROSSLINE);
      return l;
    }
  }
  if((l->c)<c){
    l->right = LL_Add2(l->right,v1,v2,n,flags);
  } else {
    l->left  = LL_Add2(l->left ,v1,v2,n,flags);
  }
  return l;
}

void LL_Collect2(LineList l, llent *line_array)
{
  if(l){
    if(l->right) LL_Collect2(l->right,line_array);
    if(l->left)  LL_Collect2(l->left ,line_array);
    line_array[l->i].v1    = l->p1;
	line_array[l->i].v2    = l->p2;
    line_array[l->i].n1    = l->n1;
	line_array[l->i].n2    = l->n2;
    line_array[l->i].flags = l->flags | 64;
    delete(l);
  }
}


float CalcStep(double h)
{
  /*GLdouble   x1,y1,z1;
  GLdouble   x2,y2,z2;
  GLdouble   mm[16];
  GLdouble   mp[16];
  GLint      mv[16];
  float      d1,d2,d;

  glGetDoublev(GL_MODELVIEW_MATRIX,mm);
  glGetDoublev(GL_PROJECTION_MATRIX,mp);
  glGetIntegerv(GL_VIEWPORT,mv);
  gluProject(-1.0, h,  0.0, mm, mp, mv, &x1, &y1, &z1);
  gluProject(+1.0, h,  0.0, mm, mp, mv, &x2, &y2, &z2);
  d1=sqrt(sqr((float)(x2-x1))+sqr((float)(y2-y1)));
  gluProject( 0.0, h, -1.0, mm, mp, mv, &x1, &y1, &z1);
  gluProject( 0.0, h, +1.0, mm, mp, mv, &x2, &y2, &z2);
  d2=sqrt(sqr((float)(x2-x1))+sqr((float)(y2-y1)));
  d=max(d1,d2);
  
  if(d>150)*/
	return 0.19634f;
  /*else if(d>60)
	return 0.26179f;
  else if(d>40)
	return 0.39269f;
  else
    return 0.52359f;*/
}


void Surface::MeshPts(float ustart, float ustop, float ustep, ushort uloop, float vstart, float vstop, float vstep, ushort vloop)
{
  float u,v; int i=0, j=0; mesh.uc=0; mesh.vc=0;
  mesh.nv = ((int)((vstop-vstart)/vstep)+3)*((int)((ustop-ustart)/ustep)+3);
  mesh.pts=new vector[mesh.nv]; mesh.vnm=new vector[mesh.nv]; mesh.tx1=new float[mesh.nv]; mesh.ty1=new float[mesh.nv];
  mesh.pts[i] = PPoint(ustart, vstart); mesh.vnm[i]=PNormal(ustart, vstart); mesh.vc++; i++;
  mesh.tx1[j] = txc? bilinear(0.0f, 0.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) : 0.0f;
  mesh.ty1[j] = txc? bilinear(0.0f, 0.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : 0.0f; j++;
  for(v=vstart+vstep; v<(vstop-0.2*vstep); v+=vstep){
	mesh.pts[i] = PPoint(ustart,v); mesh.vnm[i]=PNormal(ustart,v); mesh.vc++; i++;
	mesh.tx1[j] = txc? bilinear(0.0f, (v-vstart)/(vstop-vstart), txc[3].x, txc[2].x, txc[1].x, txc[0].x) : (v-vstart)/(vstop-vstart);
	mesh.ty1[j] = txc? bilinear(0.0f, (v-vstart)/(vstop-vstart), txc[3].y, txc[2].y, txc[1].y, txc[0].y) :  0.0f; j++;
  }
  if(!vloop){
	mesh.pts[i] = PPoint(ustart,vstop); mesh.vnm[i]=PNormal(ustart, vstop); mesh.vc++; i++;
  }
  mesh.tx1[j] = txc? bilinear(0.0f, 1.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) : 1.0f;
  mesh.ty1[j] = txc? bilinear(0.0f, 1.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : 0.0f; j++;
  mesh.uc++;
  if(idnt==SPHERE){
	u=((ustart>-1.5708)? ((int)((ustart+Pi_2)/ustep)+1)*ustep - Pi_2 : ustep);
  } else {
	u=ustart+ustep;
  }
  for(; u<(ustop-0.2*ustep); u+=ustep){
	mesh.pts[i] = PPoint(u, vstart); mesh.vnm[i] = PNormal(u, vstart); i++;
    mesh.tx1[j] = txc? bilinear((u-ustart)/(ustop-ustart), 0.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) :     0.0f                 ;
    mesh.ty1[j] = txc? bilinear((u-ustart)/(ustop-ustart), 0.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : (u-ustart)/(ustop-ustart); j++;
	for(v=vstart+vstep; v<(vstop-0.2*vstep); v+=vstep){
	  mesh.pts[i] = PPoint(u,v); mesh.vnm[i] = PNormal(u,v); i++;
      mesh.tx1[j] = txc? bilinear((u-ustart)/(ustop-ustart), (v-vstart)/(vstop-vstart), txc[3].x, txc[2].x, txc[1].x, txc[0].x) : (v-vstart)/(vstop-vstart);
      mesh.ty1[j] = txc? bilinear((u-ustart)/(ustop-ustart), (v-vstart)/(vstop-vstart), txc[3].y, txc[2].y, txc[1].y, txc[0].y) : (u-ustart)/(ustop-ustart); j++;
    }
	if(!vloop){
	  mesh.pts[i] = PPoint(u,vstop); mesh.vnm[i] = PNormal(u, vstop); i++;
	}
    mesh.tx1[j] = txc? bilinear((u-ustart)/(ustop-ustart), 1.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) :     1.0f                 ;
    mesh.ty1[j] = txc? bilinear((u-ustart)/(ustop-ustart), 1.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : (u-ustart)/(ustop-ustart); j++;
	mesh.uc++;
  }
  if(!uloop){
    mesh.pts[i] = PPoint(ustop, vstart); mesh.vnm[i] = PNormal(ustop, vstart); i++;
    mesh.tx1[j] = txc? bilinear(1.0f, 0.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) : 0.0f;
    mesh.ty1[j] = txc? bilinear(1.0f, 0.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : 1.0f; j++;
    for(v=vstart+vstep; v<(vstop-0.2*vstep); v+=vstep){
      mesh.pts[i] = PPoint(ustop,v); mesh.vnm[i] = PNormal(ustop, v); i++;
      mesh.tx1[j] = txc? bilinear(1.0f, (v-vstart)/(vstop-vstart), txc[3].x, txc[2].x, txc[1].x, txc[0].x) : (v-vstart)/(vstop-vstart);
      mesh.ty1[j] = txc? bilinear(1.0f, (v-vstart)/(vstop-vstart), txc[3].y, txc[2].y, txc[1].y, txc[0].y) :                    1.0f  ; j++;
	}
    if(!vloop){
	  mesh.pts[i] = PPoint(ustop,vstop); mesh.vnm[i] = PNormal(ustop, vstop); i++;
	}
    mesh.tx1[j] = txc? bilinear(1.0f, 1.0f, txc[3].x, txc[2].x, txc[1].x, txc[0].x) : 1.0f;
    mesh.ty1[j] = txc? bilinear(1.0f, 1.0f, txc[3].y, txc[2].y, txc[1].y, txc[0].y) : 1.0f; j++;
    mesh.uc++;
  }
  mesh.nv=mesh.uc*mesh.vc;
}

void Surface::CompleteRMesh(int u_steps, int v_steps, int u_closed, int v_closed, int inverse)
{
  int lp=0,sp=0;
  mesh.lin=new llent [mesh.nl=(2*(u_steps-1)+1)*(v_steps-1)+(v_closed? 2*(u_steps-1)+1 : (u_steps-1))+(u_closed? v_steps   : 0) ];
  mesh.snm=new vector[           (u_steps-1)   *(v_steps-1)+(v_closed?   (u_steps-1)   :  0         )+(u_closed? v_steps-1 : 0) ];
  for(int v=0; v<(v_steps-1); v++){
    for(int u=0; u<(u_steps-1); u++){
	  mesh.snm[sp]=unit( (mesh.pts[(u+1)+(v+1)*u_steps]-mesh.pts[u+v*u_steps]) * (mesh.pts[u+(v+1)*u_steps]-mesh.pts[(u+1)+v*u_steps]) )*NM; if(inverse) mesh.snm[sp]=-mesh.snm[sp];
      mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2= u   +(v+1)*u_steps; mesh.lin[lp].n1=sp; mesh.lin[lp].n2=(u==0? (u_closed? sp+u_steps-1              : NO_NORMAL) : sp-                        1  ); mesh.lin[lp].flags=((u==0 && !u_closed)? 64 : 0); lp++;
      mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=(u+1)+ v   *u_steps; mesh.lin[lp].n1=sp; mesh.lin[lp].n2=(v==0? (v_closed? u+(u_steps-1)*(v_steps-1) : NO_NORMAL) : sp-(u_steps-(u_closed? 0 : 1))); mesh.lin[lp].flags=((v==0 && !v_closed)? 64 : 0); lp++;
	  sp++;
	}
	if(u_closed){
	  mesh.snm[sp]=unit( (mesh.pts[(v+1)*u_steps]-mesh.pts[u+v*u_steps]) * (mesh.pts[u+(v+1)*u_steps]-mesh.pts[v*u_steps]) )*NM; if(inverse) mesh.snm[sp]=-mesh.snm[sp]; sp++;
      mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=v*u_steps; mesh.lin[lp].n1=sp-1; mesh.lin[lp].n2=(v==0? NO_NORMAL : sp-u_steps-1); mesh.lin[lp].flags=(v==0? 64 : 0); lp++;
	}
    mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=u+(v+1)*u_steps; mesh.lin[lp].n1=sp-1; mesh.lin[lp].n2=(u_closed? sp-2 : NO_NORMAL); mesh.lin[lp].flags=(u_closed? 0 : 64); lp++;
  }
  for(int u=0; u<(u_steps-1); u++){
	if(v_closed){
	  mesh.snm[sp]=unit( (mesh.pts[u+1]-mesh.pts[u+v*u_steps]) * (mesh.pts[u]-mesh.pts[(u+1)+v*u_steps]) )*NM; if(inverse) mesh.snm[sp]=-mesh.snm[sp];
      mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=u; mesh.lin[lp].n1=sp; mesh.lin[lp].n2=(u==0? NO_NORMAL : sp-1 ); mesh.lin[lp].flags=(u==0? 64 : 0); lp++;
	}
    mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=(u+1)+ v   *u_steps; mesh.lin[lp].n1=u+(v-1)*(u_steps-(u_closed? 0 : 1)); mesh.lin[lp].n2=(v_closed? sp : NO_NORMAL); mesh.lin[lp].flags=(v_closed? 0 : 64); lp++;
	if(v_closed) sp++;
  }
  if(u_closed){ mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=v*u_steps; mesh.lin[lp].n1=sp-1; mesh.lin[lp].n2=NO_NORMAL; mesh.lin[lp].flags=64; lp++; }
  if(v_closed){ mesh.lin[lp].v1=u+v*u_steps; mesh.lin[lp].v2=  u      ; mesh.lin[lp].n1=sp-1; mesh.lin[lp].n2=NO_NORMAL; mesh.lin[lp].flags=64; lp++; }
}


void Surface::ShadeRMesh(int u_steps, int v_steps, int u_closed, int v_closed, int inverse)
{
  vector p,n;
  int ut_steps = u_closed? u_steps+1 : u_steps;

  if(flags & SURF_DOUBLESIDED){
    glDisable(GL_CULL_FACE);
  } else {
    glEnable(GL_CULL_FACE); glCullFace(GL_BACK);
  }
  glBegin(GL_QUADS);
    for(int v=0; v<(v_steps-1); v++){
      for(int u=0; u<(u_steps-1); u++){
		if(inverse){
          p=mesh.pts[ u   + v   *u_steps]; n=mesh.vnm[ u   + v   *u_steps]; glTexCoord2d(mesh.tx1[ u   + v   *ut_steps], mesh.ty1[ u   + v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[ u   +(v+1)*u_steps]; n=mesh.vnm[ u   +(v+1)*u_steps]; glTexCoord2d(mesh.tx1[ u   +(v+1)*ut_steps], mesh.ty1[ u   +(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)+(v+1)*u_steps]; n=mesh.vnm[(u+1)+(v+1)*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+(v+1)*ut_steps], mesh.ty1[(u+1)+(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)+ v   *u_steps]; n=mesh.vnm[(u+1)+ v   *u_steps]; glTexCoord2d(mesh.tx1[(u+1)+ v   *ut_steps], mesh.ty1[(u+1)+ v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		} else {
	      p=mesh.pts[(u+1)+ v   *u_steps]; n=mesh.vnm[(u+1)+ v   *u_steps]; glTexCoord2d(mesh.tx1[(u+1)+ v   *ut_steps], mesh.ty1[(u+1)+ v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)+(v+1)*u_steps]; n=mesh.vnm[(u+1)+(v+1)*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+(v+1)*ut_steps], mesh.ty1[(u+1)+(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[ u   +(v+1)*u_steps]; n=mesh.vnm[ u   +(v+1)*u_steps]; glTexCoord2d(mesh.tx1[ u   +(v+1)*ut_steps], mesh.ty1[ u   +(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
          p=mesh.pts[ u   + v   *u_steps]; n=mesh.vnm[ u   + v   *u_steps]; glTexCoord2d(mesh.tx1[ u   + v   *ut_steps], mesh.ty1[ u   + v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		}
	  }
	  if(u_closed){
		if(inverse){
          p=mesh.pts[u+ v   *u_steps]; n=mesh.vnm[u+ v   *u_steps]; glTexCoord2d(mesh.tx1[ u   + v   *ut_steps], mesh.ty1[ u   + v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[u+(v+1)*u_steps]; n=mesh.vnm[u+(v+1)*u_steps]; glTexCoord2d(mesh.tx1[ u   +(v+1)*ut_steps], mesh.ty1[ u   +(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[  (v+1)*u_steps]; n=mesh.vnm[  (v+1)*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+(v+1)*ut_steps], mesh.ty1[(u+1)+(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[   v   *u_steps]; n=mesh.vnm[   v   *u_steps]; glTexCoord2d(mesh.tx1[(u+1)+ v   *ut_steps], mesh.ty1[(u+1)+ v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		} else {
	      p=mesh.pts[   v   *u_steps]; n=mesh.vnm[   v   *u_steps]; glTexCoord2d(mesh.tx1[(u+1)+ v   *ut_steps], mesh.ty1[(u+1)+ v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[  (v+1)*u_steps]; n=mesh.vnm[  (v+1)*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+(v+1)*ut_steps], mesh.ty1[(u+1)+(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[u+(v+1)*u_steps]; n=mesh.vnm[u+(v+1)*u_steps]; glTexCoord2d(mesh.tx1[ u   +(v+1)*ut_steps], mesh.ty1[ u   +(v+1)*ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
          p=mesh.pts[u+ v   *u_steps]; n=mesh.vnm[u+ v   *u_steps]; glTexCoord2d(mesh.tx1[ u   + v   *ut_steps], mesh.ty1[ u   + v   *ut_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		}
	  }
	}
    if(v_closed){
      for(int u=0; u<(u_steps-1); u++){
		if(inverse){
          p=mesh.pts[ u   +v*u_steps]; n=mesh.vnm[ u   +v*u_steps]; glTexCoord2d(mesh.tx1[ u   +v*u_steps], mesh.ty1[ u   +v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[ u             ]; n=mesh.vnm[ u             ]; glTexCoord2d(mesh.tx1[ u             ], mesh.ty1[ u             ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)          ]; n=mesh.vnm[(u+1)          ]; glTexCoord2d(mesh.tx1[(u+1)          ], mesh.ty1[(u+1)          ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)+v*u_steps]; n=mesh.vnm[(u+1)+v*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+v*u_steps], mesh.ty1[(u+1)+v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		} else {
	      p=mesh.pts[(u+1)+v*u_steps]; n=mesh.vnm[(u+1)+v*u_steps]; glTexCoord2d(mesh.tx1[(u+1)+v*u_steps], mesh.ty1[(u+1)+v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[(u+1)          ]; n=mesh.vnm[(u+1)          ]; glTexCoord2d(mesh.tx1[(u+1)          ], mesh.ty1[(u+1)          ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[ u             ]; n=mesh.vnm[ u             ]; glTexCoord2d(mesh.tx1[ u             ], mesh.ty1[ u             ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
          p=mesh.pts[ u   +v*u_steps]; n=mesh.vnm[ u   +v*u_steps]; glTexCoord2d(mesh.tx1[ u   +v*u_steps], mesh.ty1[ u   +v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		}
	  }
	  if(u_closed){
		if(inverse){
          p=mesh.pts[u+v*u_steps]; n=mesh.vnm[u+v*u_steps]; glTexCoord2d(mesh.tx1[u+v*u_steps], mesh.ty1[u+v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[u          ]; n=mesh.vnm[u          ]; glTexCoord2d(mesh.tx1[u          ], mesh.ty1[u          ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[  0        ]; n=mesh.vnm[  0        ]; glTexCoord2d(mesh.tx1[  0        ], mesh.ty1[  0        ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[  v*u_steps]; n=mesh.vnm[  v*u_steps]; glTexCoord2d(mesh.tx1[  v*u_steps], mesh.ty1[  v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		} else {
	      p=mesh.pts[  v*u_steps]; n=mesh.vnm[  v*u_steps]; glTexCoord2d(mesh.tx1[  v*u_steps], mesh.ty1[  v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[  0        ]; n=mesh.vnm[  0        ]; glTexCoord2d(mesh.tx1[  0        ], mesh.ty1[  0        ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
	      p=mesh.pts[u          ]; n=mesh.vnm[u          ]; glTexCoord2d(mesh.tx1[u          ], mesh.ty1[u          ]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
          p=mesh.pts[u+v*u_steps]; n=mesh.vnm[u+v*u_steps]; glTexCoord2d(mesh.tx1[u+v*u_steps], mesh.ty1[u+v*u_steps]); glNormal3f(n.x, n.y, n.z); glVertex3f(p.x, p.y, p.z);
		}
	  }
	}
  glEnd();
}


void PolyMesh::BuildGL(int hl)
{
  int hl1,i;
  if(!mesh.snm) CalcSurfaceNormals();
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM);
  if(GraphWin.DisplayMode & VM_SHADED){
    int        j,ne;
    long int   ii;
    vector     V,N;
    MatPtr    *matl;
    byte      *mats;

    if((pmflags) & SURF_DOUBLESIDED){
	  glDisable(GL_CULL_FACE);
    } else {
	  glEnable(GL_CULL_FACE); glCullFace(GL_BACK);
    }
    if(nm > 1){
      matl = mtl;
	  mats = mat;
    //} else {
    //  if(hl!=2) SetMaterial(Mat, hl);
    }
    if(((pmflags & SURF_ROUNDED) || (pmflags & SURF_GROUPED)) && (!nrm)){
      CalcSmoothingNormals();
	}
    if(hl==2){ if(flg && (flg[0] & SELECTED)) hl1=1; else hl1=0; } else hl1=hl;
    if(nm > 1) SetMaterial(matl[mats[0]], hl1);
    if(!pnv) glBegin(GL_TRIANGLES);
    for(ii=0,i=0; i<nf; i++) {
      ne = pnv? pnv[i] : 3;
	  if(hl==2){ if(flg && (flg[i] & SELECTED)) hl1=1; else hl1=0; } else hl1=hl;
      if(nm > 1){
        if((i>0) || (mats[i]!=mats[i-1])){
          if(!pnv) glEnd();
		  SetMaterial(matl[mats[i]], hl1);
          if(!pnv) glBegin(GL_TRIANGLES);
		}
	  } else {
	    //if(hl==2) SetMaterial(Mat, hl1);
	  }
	  if(pnv) glBegin(GL_POLYGON);
      if(!nrm){ N=mesh.snm[i]; glNormal3f(N.x, N.y, N.z); }
      for(j=0; j<ne; j++){
        if(nrm){
		  N = pfn? nrm[pfn[ii+ne-1-j]] : nrm[pfv[ii+ne-1-j]];
          glNormal3f(N.x, N.y, N.z);
		}
        V = pts[pfv[ii+ne-1-j]];
        if(UseTexture){
          if(txc){
            // use stored texture coords
            glTexCoord2d(txc[pft[ii+ne-1-j]].x, txc[pft[ii+ne-1-j]].y);
		  } else {
            // calculate view-dependent texture coords for reflection mapping
            vector R =  unit(V - GraphWin.GScreen.V.Location);
  	        vector S = -unit(2*(N % R)*N-R);
            float a = ((fabs(S.x)>0.001f)? (float)atan(S.y/S.x) : Pi_2) + ((S.x<0.0f)? Pi : 0.0f);
            float b = (float)asin(S.z);
            glTexCoord2d(a/TwoPi, b/TwoPi);
		  }
		}
        glVertex3f(V.x, V.y, V.z);
	  }
      if(pnv) glEnd();
      ii += ne;
	}
    if(!pnv) glEnd();
  } else {
    if(!mesh.lin){
      // count and collect unique polygon edges in binary tree
      LineList ll = NULL;
      byte     f  = (pmflags & SURF_ROUNDED);
      long int      ap,i,j,pn;
      for(LL_Count2=0,ap=0,i=0; i<nf; i++) {
        pn=(pnv? pnv[i] : 3);
        for(j=0; j<(pn-1); j++){
          ll=LL_Add2(ll, pfv[ap+j], pfv[ap+j+1], (ushort)i, f);
        }
        ll=LL_Add2(ll, pfv[ap+j], pfv[ap], (ushort)i, f);
        ap+=pn;
      }
      // create line-array and collect line-entries from tree into the array
      mesh.nl=LL_Count2; mesh.lin = new llent[mesh.nl];
      LL_Collect2(ll, mesh.lin);
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}


void Cube::BuildGL(int hl)
{
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM);
  if(GraphWin.DisplayMode & VM_SHADED){
    glEnable(GL_CULL_FACE); glCullFace(GL_BACK);
    glBegin(GL_QUADS);
      glNormal3f( 0.0f, 0.0f,  1.0f);
      glVertex3f(-0.5f,-0.5f,  0.5f); glVertex3f( 0.5f,-0.5f,  0.5f); glVertex3f( 0.5f, 0.5f,  0.5f); glVertex3f(-0.5f, 0.5f,  0.5f);
      glNormal3f( 1.0f, 0.0f,  0.0f);
      glVertex3f( 0.5f,-0.5f,  0.5f); glVertex3f( 0.5f,-0.5f, -0.5f); glVertex3f( 0.5f, 0.5f, -0.5f); glVertex3f( 0.5f, 0.5f,  0.5f);
      glNormal3f( 0.0f, 0.0f, -1.0f);
      glVertex3f(-0.5f, 0.5f, -0.5f); glVertex3f( 0.5f, 0.5f, -0.5f); glVertex3f( 0.5f,-0.5f, -0.5f); glVertex3f(-0.5f,-0.5f, -0.5f);
      glNormal3f(-1.0f, 0.0f,  0.0f);
      glVertex3f(-0.5f, 0.5f,  0.5f); glVertex3f(-0.5f, 0.5f, -0.5f); glVertex3f(-0.5f,-0.5f, -0.5f); glVertex3f(-0.5f,-0.5f,  0.5f);
      glNormal3f( 0.0f, 1.0f,  0.0f);
      glVertex3f(-0.5f, 0.5f,  0.5f); glVertex3f( 0.5f, 0.5f,  0.5f); glVertex3f( 0.5f, 0.5f, -0.5f); glVertex3f(-0.5f, 0.5f, -0.5f);
      glNormal3f( 0.0f,-1.0f,  0.0f);
      glVertex3f(-0.5f,-0.5f, -0.5f); glVertex3f( 0.5f,-0.5f, -0.5f); glVertex3f( 0.5f,-0.5f,  0.5f); glVertex3f(-0.5f,-0.5f,  0.5f);
    glEnd();
  } else {
	if(!(mesh.pts)){
	  mesh.nv=8; mesh.pts=new vector[8];
	  mesh.pts[0]=vector(-0.5f, -0.5f, -0.5f);
	  mesh.pts[1]=vector( 0.5f, -0.5f, -0.5f);
	  mesh.pts[2]=vector(-0.5f, -0.5f,  0.5f);
	  mesh.pts[3]=vector( 0.5f, -0.5f,  0.5f);
	  mesh.pts[4]=vector(-0.5f,  0.5f, -0.5f);
	  mesh.pts[5]=vector( 0.5f,  0.5f, -0.5f);
	  mesh.pts[6]=vector(-0.5f,  0.5f,  0.5f);
	  mesh.pts[7]=vector( 0.5f,  0.5f,  0.5f);
	}
	if(!(mesh.snm)){
	  mesh.snm=new vector[6];
	  if(flags & SURF_INVERSE){
	    mesh.snm[0]=vector( 1.0f,  0.0f,  0.0f)*NM;
	    mesh.snm[1]=vector( 0.0f,  1.0f,  0.0f)*NM;
	    mesh.snm[2]=vector( 0.0f,  0.0f,  1.0f)*NM;
	    mesh.snm[3]=vector(-1.0f,  0.0f,  0.0f)*NM;
	    mesh.snm[4]=vector( 0.0f, -1.0f,  0.0f)*NM;
	    mesh.snm[5]=vector( 0.0f,  0.0f, -1.0f)*NM;
	  } else {
	    mesh.snm[0]=vector(-1.0f,  0.0f,  0.0f)*NM;
	    mesh.snm[1]=vector( 0.0f, -1.0f,  0.0f)*NM;
	    mesh.snm[2]=vector( 0.0f,  0.0f, -1.0f)*NM;
	    mesh.snm[3]=vector( 1.0f,  0.0f,  0.0f)*NM;
	    mesh.snm[4]=vector( 0.0f,  1.0f,  0.0f)*NM;
	    mesh.snm[5]=vector( 0.0f,  0.0f,  1.0f)*NM;
	  }
	}
	if(!(mesh.lin)){
	  mesh.nl=12; mesh.lin=new llent[12];
	  mesh.lin[ 0].v1=0; mesh.lin[ 0].v2=1; mesh.lin[ 0].n1=1; mesh.lin[ 0].n2=2; mesh.lin[ 0].flags=64;
	  mesh.lin[ 1].v1=1; mesh.lin[ 1].v2=3; mesh.lin[ 1].n1=1; mesh.lin[ 1].n2=3; mesh.lin[ 1].flags=64;
	  mesh.lin[ 2].v1=3; mesh.lin[ 2].v2=2; mesh.lin[ 2].n1=1; mesh.lin[ 2].n2=5; mesh.lin[ 2].flags=64;
	  mesh.lin[ 3].v1=2; mesh.lin[ 3].v2=0; mesh.lin[ 3].n1=1; mesh.lin[ 3].n2=0; mesh.lin[ 3].flags=64;
	  mesh.lin[ 4].v1=4; mesh.lin[ 4].v2=5; mesh.lin[ 4].n1=4; mesh.lin[ 4].n2=2; mesh.lin[ 4].flags=64;
	  mesh.lin[ 5].v1=5; mesh.lin[ 5].v2=7; mesh.lin[ 5].n1=4; mesh.lin[ 5].n2=3; mesh.lin[ 5].flags=64;
	  mesh.lin[ 6].v1=7; mesh.lin[ 6].v2=6; mesh.lin[ 6].n1=4; mesh.lin[ 6].n2=5; mesh.lin[ 6].flags=64;
	  mesh.lin[ 7].v1=6; mesh.lin[ 7].v2=4; mesh.lin[ 7].n1=4; mesh.lin[ 7].n2=0; mesh.lin[ 7].flags=64;
	  mesh.lin[ 8].v1=0; mesh.lin[ 8].v2=4; mesh.lin[ 8].n1=0; mesh.lin[ 8].n2=2; mesh.lin[ 8].flags=64;
	  mesh.lin[ 9].v1=2; mesh.lin[ 9].v2=6; mesh.lin[ 9].n1=0; mesh.lin[ 9].n2=5; mesh.lin[ 9].flags=64;
	  mesh.lin[10].v1=1; mesh.lin[10].v2=5; mesh.lin[10].n1=3; mesh.lin[10].n2=2; mesh.lin[10].flags=64;
	  mesh.lin[11].v1=3; mesh.lin[11].v2=7; mesh.lin[11].n1=3; mesh.lin[11].n2=5; mesh.lin[11].flags=64;
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}


void Cylinder::BuildGL(int hl)
{
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM);
  float step=CalcStep(0.5); float stop = angle? angle-0.2f*step : 6.283185f-0.2f*step;

  if((!mesh.pts) || (!mesh.vnm)) MeshPts(0.0f, 1.0f, 1.0f, 0, 0.0f, (angle? angle : TwoPi), step, (!angle) || (angle>6.2831));
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, (!angle) || (angle>6.2831), 0, (flags & SURF_INVERSE));
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, (!angle) || (angle>6.2831), 0, (flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}

void Sphere::BuildGL(int hl)
{  
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM); float vstep, ustep; vstep=ustep=CalcStep(0.0);

  if((!mesh.pts) || (!mesh.vnm)) MeshPts(ustart, ustop, ustep, 0, 0.0f, (vangle? vangle : TwoPi), vstep, (!vangle) || (vangle>6.2831));
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, ((!vangle) || (vangle>6.2831)), 0, (flags & SURF_INVERSE));
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, ((!vangle) || (vangle>6.2831)), 0, (flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}


void Ring::BuildGL(int hl)
{
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM);
  float step=CalcStep(0.5);

  if((!mesh.pts) || (!mesh.vnm)) MeshPts(r_i, 1.0f, 0.1f, 0, 0.0f, (angle? angle : TwoPi), step, (!angle) || (angle>6.2831));
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, ((!angle) || (angle>6.2831)), 0, !(flags & SURF_INVERSE));
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, ((!angle) || (angle>6.2831)), 0, !(flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}

void Paraboloid::BuildGL(int hl)
{  
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM); float vstep, ustep; vstep=ustep=CalcStep(0.0);

  if((!mesh.pts) || (!mesh.vnm)) MeshPts(0.0f, (float)sqrt(height), 0.5f*ustep, 0, 0.0f, (angle? angle : TwoPi), vstep, (!angle) || (angle>6.2831));
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, ((!angle) || (angle>6.2831)), 0, (flags & SURF_INVERSE));
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, ((!angle) || (angle>6.2831)), 0, (flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}

void Function::BuildGL(int hl)
{
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM);
  int uc=64; int vc=64;
  if((!mesh.pts) || (!mesh.vnm)) MeshPts(Min.x, Max.x, (Max.x-Min.x)/uc, 0, Min.y, Max.y, (Max.y-Min.y)/vc, 0);
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, 0, 0, 1/*(flags & INVERSE)*/);
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, 0, 0, !(flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}


void Torus::BuildGL(int hl)
{ 
  glPushMatrix(); // Isolation of the transformation for each object (julien)
  SetTransform(&TM); float vstep, ustep; vstep=ustep=CalcStep(0.0);

  if((!mesh.pts) || (!mesh.vnm)) MeshPts(ustart, ustop, ustep, (ustop>6.2831), 0.0f, (vangle? vangle : TwoPi), vstep, (!vangle) || (vangle>6.2831));
  if(GraphWin.DisplayMode & VM_SHADED){
	ShadeRMesh(mesh.vc, mesh.uc, ((!vangle) || (vangle>6.2831)), (ustop>6.2831), (flags & SURF_INVERSE));
  } else {
	if(!(mesh.snm) || !(mesh.lin)){
	  CompleteRMesh(mesh.vc, mesh.uc, ((!vangle) || (vangle>6.2831)), (ustop>6.2831), (flags & SURF_INVERSE));
	}
	DrawWire(hl);
  }
  glPopMatrix(); // Return to the original transformation for the MODELVIEW matrix (julien)
}


void CreateGL1(PrimPtr pr, int hl)
{
  Surface   *p = (Surface *)pr;

  if((p->idnt==POLYMESH) || (p->idnt==CUBE) || (p->idnt==CYLINDER) || (p->idnt==SPHERE) || (p->idnt==TORUS) || (p->idnt==FUNCTION) || (p->idnt==PARABOLOID) || (p->idnt==RING)){
    // create texture-display-lists for multimaterial objects before object-display-list
	if((GraphWin.DisplayMode & VM_SHADED) && (GraphWin.DisplayMode & VM_TEXTURED) && (p->idnt==POLYMESH) && (((PolyMesh *)p)->nm > 1)){
	  PolyMesh *t = (PolyMesh *)p;
	  for(int i=0; i<(t->nm); i++){
		if(t->mtl[i]->Texture){
		  if(!(t->mtl[i]->Texture->tbits)) SetMaterial(t->mtl[i], 0);
          //glError(glGetError(),"preSetMaterial %s",t->mtl[i]->name);
		}
	  }
	}
    glNewList (p->OGLL, GL_COMPILE); glError(glGetError(),"NewList %d",p->OGLL);
	  p->BuildGL(hl);
    glEndList(); glError(glGetError(),"EndList");
    p->dflags |= 1;
    if(GraphWin.DisplayMode & VM_SHADED) p->dflags |= 2; else p->dflags &= ~2;
    if(hl) p->dflags |= 4; else p->dflags &= ~4;
  }
}


void CreateGL(PrimPtr pr)
{
  while(pr){
	CreateGL1(pr,0);
    pr = pr->succ;
  }
}

void InitRay(vector Org)
{
  glVertex3f(Org.x, Org.y, Org.z);
}

float R_fnf(float n1, float n2, double ce1);

RGB_Color RayGL(RV_PrP OrgObj, MatPtr mat, vector *Org, vector *Ray, IntPar IntP, float em, float dist, int level, Thrdata *Thread)
{
  GetIPN(Org, Ray, &IntP); MatPtr m = IntP.p->Mat;
  glVertex3f(IntP.IP.x, IntP.IP.y, IntP.IP.z);

  // specular reflection/refraction
  if((level<20) && (m->Specularity>ray_thresh)){
    float  spc = 0.0f, trn=0.0f;
	float  ca  = IntP.N % *Ray;
	vector OutDir;

	if(m->Transmission.R < 0.0f){
	  // fresnel reflection?
	  float n1,n2;
  	  if(ca<0.0f){
	    n1=mat->IOR(wavlen); n2=            m->IOR(wavlen);
	  } else {
        n1=m  ->IOR(wavlen); n2=RV_Environment.IOR(wavlen);
	  }
	  spc = m->Specularity * R_fnf(n1,n2,ca); trn = m->Specularity-spc;
	} else {
	  // simple specular reflection/refraction?
	  if((m->Specularity > ray_thresh) && (m->type!=Shader_Phong)){
	    spc = m->Specularity * ((m->Reflection.G < 0.0f)? m->Reflection.R : wlcoff(m->Reflection.R, m->Reflection.G, m->Reflection.B, wavlen));
        trn = m->Transmission.R;
	  }
	}

	// collect specular reflected light
	if(spc > ray_thresh) {
	  // mirror direction
	  OutDir = -unit(2.0f * ca * IntP.N - *Ray); //if(m->PBExp > 0.001f) OutDir = rndpdir(OutDir, 1.0f/m->PBExp);
      RayTrace(Thread, IntP.p, &InitRay, &(VMDoc->VMF.Environment), &(IntP.IP), &OutDir, 1.0f, 0.0f, &RayGL, level+1, 20);
	}
    // collect transmitted (refracted) light
	if(trn > ray_thresh) {
	  if(ca<0.0f){
	    OutDir=refract(*Ray,  IntP.N, mat->IOR(wavlen),             m->IOR(wavlen));
        RayTrace(Thread, IntP.p, &InitRay,                           m                    , &(IntP.IP), &OutDir, 1.0f, 0.0f, &RayGL, level+1, 20);
	  } else {
	    OutDir=refract(*Ray, -IntP.N, mat->IOR(wavlen), RV_Environment.IOR(wavlen));
        RayTrace(Thread, IntP.p, &InitRay, ((OutDir % IntP.N)<0.0f)? m : &(RV_Environment), &(IntP.IP), &OutDir, 1.0f, 0.0f, &RayGL, level+1, 20);
	  }
	}
  }
  return RGB_Color(-1.0f, -1.0f, -1.0f);
}

extern void InitRaVi(); Thrdata ThD[4];

void DrawObj(PrimPtr p)
{
  if((p->idnt==POLYMESH) || (p->idnt==CUBE) || (p->idnt==SPHERE) || (p->idnt==CYLINDER) || (p->idnt==TORUS) || (p->idnt==FUNCTION) || (p->idnt==PARABOLOID) || (p->idnt==RING)){
    if((GraphWin.DisplayMode & VM_SHADED) && ((p->idnt!=POLYMESH) || (((PolyMesh *)p)->nm < 2)) && ((Surface *)p)->Mat) SetMaterial(((Surface *)p)->Mat, 0);
    //glError(glGetError(),"after SetMaterial");
    if((GraphWin.DisplayMode & VM_REBUILD) || (((Surface *)p)->dflags & 4) || (!(((Surface *)p)->dflags & 1)) || (!(GraphWin.DisplayMode & VM_SHADED)) || (!(((Surface *)p)->dflags & 2)) ){
  	  CreateGL1(p,0); glError(glGetError(),"CreateGL1 '%s'",((Surface *)p)->name);
    }
    glCallList(((Surface *)p)->OGLL); //glError(glGetError(),"CallList %d",((Surface *)p)->OGLL);
  } else if(p->idnt==RAY) {
	Ray *r = (Ray *)p; if(!RV_SceneTree) InitRaVi();
	glDisable(GL_LIGHTING); glDisable(GL_TEXTURE_2D);
    glColor3f(wlcoff(1.0f, 0.0f, 0.0f, r->wavelength), wlcoff(0.0f, 1.0f, 0.0f, r->wavelength), wlcoff(0.0f, 0.0f, 1.0f, r->wavelength));
    glBegin(GL_LINE_STRIP);
      wavlen=r->wavelength; ray_thresh=0.3f;
      RayTrace(ThD, NULL, &InitRay, &(VMDoc->VMF.Environment), &(r->origin), &(r->direction), 0.0f, 0.0f, &RayGL, 0, 20);
    glEnd();
  }
}
