// GraphWin.cpp : implementation file
//

#include "stdafx.h"
#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "GraphWin.h"

#include "gl\gl.h"
#include "gl\glu.h"

extern void DrawObj(PrimPtr p);

#ifdef _DEBUG
  #define new DEBUG_NEW
  #undef THIS_FILE
  static char THIS_FILE[] = __FILE__;
#endif

unsigned char threeto8[8] = { 0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377 };

unsigned char twoto8[4]   = { 0, 0x55, 0xaa, 0xff };

unsigned char oneto8[2]   = { 0, 255 };

static int defaultOverride[13] = { 0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91 };

vector LPos;

void glError(GLenum Ern, char *Text, ...);

static PALETTEENTRY defaultPalEntry[20] = 
{
  { 0,   0,   0,    0 },
  { 0x80,0,   0,    0 },
  { 0,   0x80,0,    0 },
  { 0x80,0x80,0,    0 },
  { 0,   0,   0x80, 0 },
  { 0x80,0,   0x80, 0 },
  { 0,   0x80,0x80, 0 },
  { 0xC0,0xC0,0xC0, 0 },

  { 192, 220, 192,  0 },
  { 166, 202, 240,  0 },
  { 255, 251, 240,  0 },
  { 160, 160, 164,  0 },

  { 0x80,0x80,0x80, 0 },
  { 0xFF,0,   0,    0 },
  { 0,   0xFF,0,    0 },
  { 0xFF,0xFF,0,    0 },
  { 0,   0,   0xFF, 0 },
  { 0xFF,0,   0xFF, 0 },
  { 0,   0xFF,0xFF, 0 },
  { 0xFF,0xFF,0xFF, 0 }
};

/////////////////////////////////////////////////////////////////////////////
// CGraphWin

CGraphWin::CGraphWin()  { }

CGraphWin::~CGraphWin() { }

BEGIN_MESSAGE_MAP(CGraphWin, CWnd)
	//{{AFX_MSG_MAP(CGraphWin)
	ON_WM_ERASEBKGND()
	ON_WM_CREATE()
	ON_WM_PAINT()
	ON_WM_DESTROY()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_MOUSEMOVE()
	ON_WM_RBUTTONDOWN()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


void CGraphWin::Init()
{
  PIXELFORMATDESCRIPTOR  pfd;
  int                    n;
  HGLRC                  hrc;

  m_pDC = new CClientDC(this);

  ASSERT(m_pDC != NULL);

  if (!bSetupPixelFormat()) return;

  n = ::GetPixelFormat(m_pDC->GetSafeHdc());
  ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

  CreateRGBPalette(); DisplayMode=VM_TEXTURED;

  hrc = wglCreateContext(m_pDC->GetSafeHdc());
  wglMakeCurrent(m_pDC->GetSafeHdc(), hrc);

  glNewList(1, GL_COMPILE);
  glEndList();
}


BOOL CGraphWin::bSetupPixelFormat()
{
  static PIXELFORMATDESCRIPTOR pfd = {
    sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
    1,                              // version number
    PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
                                    // support window, OpenGL, double buffered
    PFD_TYPE_RGBA,                  // RGBA type
    24,                             // 24-bit true color depth
    0, 0, 0, 0, 0, 0,               // color bits ignored
    0,                              // no alpha buffer
    0,                              // shift bit ignored
    24,                             // true color accumulation buffer for full scene antialiasing
    0, 0, 0, 0,                     // accum bits ignored
    32,                             // 32-bit z-buffer
    0,                              // no stencil buffer
    0,                              // no auxiliary buffer
    PFD_MAIN_PLANE,                 // main layer
    0,                              // reserved
    0, 0, 0                         // layer masks ignored
  };
  int pixelformat;

  if ( (pixelformat = ChoosePixelFormat(m_pDC->GetSafeHdc(), &pfd)) == 0 ) {
    MessageBox("ChoosePixelFormat failed");
    return FALSE;
  }

  if (SetPixelFormat(m_pDC->GetSafeHdc(), pixelformat, &pfd) == FALSE) {
    MessageBox("SetPixelFormat failed");
    return FALSE;
  }

  return TRUE;
}


unsigned char CGraphWin::ComponentFromIndex(int i, UINT nbits, UINT shift)
{
  unsigned char val;

  val = (unsigned char) (i >> shift);
  switch (nbits) {
    case 1:  val &= 0x1; return oneto8[val];
    case 2:  val &= 0x3; return twoto8[val];
    case 3:  val &= 0x7; return threeto8[val];
    default:             return 0;
  }
}


void CGraphWin::CreateRGBPalette()
{
  PIXELFORMATDESCRIPTOR  pfd;
  LOGPALETTE            *pPal;
  int                    n,i;
 
  n = ::GetPixelFormat(m_pDC->GetSafeHdc());
  ::DescribePixelFormat(m_pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

  if (pfd.dwFlags & PFD_NEED_PALETTE) {
    n = 1 << pfd.cColorBits;
    pPal = (PLOGPALETTE) new char[sizeof(LOGPALETTE) + n * sizeof(PALETTEENTRY)];

    ASSERT(pPal != NULL);

    pPal->palVersion = 0x300;
    pPal->palNumEntries = n;
    for(i=0; i<n; i++) {
      pPal->palPalEntry[i].peRed   = ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
      pPal->palPalEntry[i].peGreen = ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
      pPal->palPalEntry[i].peBlue  = ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
      pPal->palPalEntry[i].peFlags = 0;
    }

    // fix up the palette to include the default GDI palette
    if((pfd.cColorBits == 8) && (pfd.cRedBits == 3) && (pfd.cRedShift == 0) && (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) && (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6) ) {
      for (i = 1 ; i <= 12 ; i++) pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
    }

    m_cPalette.CreatePalette(pPal);
    delete pPal;

    m_pOldPalette = m_pDC->SelectPalette(&m_cPalette, FALSE);
    m_pDC->RealizePalette();
  }
}

#define PI_ 3.14159265358979323846

/*  accFrustum()
 *  The first 6 arguments are identical to the glFrustum() call.
 *  
 *  pixdx and pixdy are anti-alias jitter in pixels. 
 *  Set both equal to 0.0 for no anti-alias jitter.
 *  eyedx and eyedy are depth-of field jitter in pixels. 
 *  Set both equal to 0.0 for no depth of field effects.
 *
 *  focus is distance from eye to plane in focus. 
 *  focus must be greater than, but not equal to 0.0.
 *
 *  Note that accFrustum() calls glTranslatef().  You will 
 *  probably want to insure that your ModelView matrix has been 
 *  initialized to identity before calling accFrustum().
 */
void accFrustum(GLdouble left, GLdouble right, GLdouble bottom, GLdouble top,
    GLdouble znear, GLdouble zfar, GLdouble pixdx, GLdouble pixdy)
{
    GLdouble xwsize, ywsize; 
    GLdouble dx, dy;
    GLint viewport[4];

    glGetIntegerv (GL_VIEWPORT, viewport);
    
    xwsize = right - left;
    ywsize = top - bottom;
    
    dx = -(pixdx*xwsize/(GLdouble) viewport[2]);
    dy = -(pixdy*ywsize/(GLdouble) viewport[3]);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(left + dx, right + dx, bottom + dy, top + dy, znear, zfar);
}

/*  accPerspective()
 * 
 *  The first 4 arguments are identical to the gluPerspective() call.
 *  pixdx and pixdy are anti-alias jitter in pixels. 
 *  Set both equal to 0.0 for no anti-alias jitter.
 *  eyedx and eyedy are depth-of field jitter in pixels. 
 *  Set both equal to 0.0 for no depth of field effects.
 *
 *  focus is distance from eye to plane in focus. 
 *  focus must be greater than, but not equal to 0.0.
 *
 *  Note that accPerspective() calls accFrustum().
 */
void accPerspective(GLdouble fovy, GLdouble aspect, 
    GLdouble znear, GLdouble zfar, GLdouble pixdx, GLdouble pixdy)
{
    GLdouble fov2,left,right,bottom,top;

    fov2 = ((fovy*PI_) / 180.0) / 2.0;

    top = znear / (cos(fov2) / sin(fov2));
    bottom = -top;

    right = top * aspect;
    left = -right;

    accFrustum (left, right, bottom, top, znear, zfar, pixdx, pixdy);
}

extern byte       *BMPbitsB;
extern BITMAPINFO  BMPinfoB;

extern void DrawVLink(CDC *hdc, vector pg, vector pp);
extern void DrawELink(CDC *hdc, vector p1, vector p2, vector pp);
extern void	SetTexture(TexMapPtr map, int TexMode, int refmap);

void CGraphWin::DrawScene(CDC *hdc)
{
  GLdouble  Aspect;
  static    BOOL bBusy = FALSE;
  int       DrwB;

  if(bBusy) return; else bBusy = TRUE;

  GetClientRect(&m_oldRect); glClearDepth(1.0f);

  if(BMPbitsB || ((DisplayMode & VM_CHANGED) && !(DisplayMode & VM_ANTIALIAS))){
	glDrawBuffer(GL_FRONT); DisplayMode &= ~VM_CHANGED;
  } else {
    glDrawBuffer(GL_BACK);
  }

  // don't divide by zero, not that we should ever run into that...
  if (m_oldRect.bottom) Aspect = (GLdouble)m_oldRect.right/m_oldRect.bottom; else Aspect = 1.0;

  glViewport(0, 0, m_oldRect.right, m_oldRect.bottom);
  
  if(DisplayMode & VM_SHADED){
    glClearColor(VMDoc->VMF.Background.R, VMDoc->VMF.Background.G, VMDoc->VMF.Background.B, 1.0f);
  } else {
	glClearColor(0.9f, 0.9f, 0.9f, 1.0f);
  }
  
  if(DisplayMode & VM_ANTIALIAS) glClear(GL_ACCUM_BUFFER_BIT);
  for(int aa=0; aa<((DisplayMode & VM_ANTIALIAS)? 4 : 1); aa++){
  
    if(BMPbitsB){
	  float fx=(float)m_oldRect.right /(float)BMPinfoB.bmiHeader.biWidth;
	  float fy=(float)m_oldRect.bottom/(float)BMPinfoB.bmiHeader.biHeight;
	  // the following crashes on Riva128 and produces bad quality on Voodoo3
	  // but worked very well on Permedia2 for example ...
      /*glMatrixMode(GL_MODELVIEW); glLoadIdentity();
      glMatrixMode(GL_PROJECTION); glLoadIdentity();
	  if(fy>fx){
	    glRasterPos3f(-1.0f, -1.0f+((m_oldRect.bottom-fx*BMPinfoB.bmiHeader.biHeight)/m_oldRect.bottom), 0.0f); glPixelZoom(fx, fx);
	  } else {
	    glRasterPos3f(-1.0f+((m_oldRect.right -fy*BMPinfoB.bmiHeader.biWidth )/m_oldRect.right ), -1.0f, 0.0f); glPixelZoom(fy, fy);
	  }
	  glDrawPixels(BMPinfoB.bmiHeader.biWidth, BMPinfoB.bmiHeader.biHeight, GL_BGR_EXT, GL_UNSIGNED_BYTE, BMPbitsB);
      glError(glGetError(),"DrawPixels");
      glClear(GL_DEPTH_BUFFER_BIT);*/
	  if(fy>fx){
	    StretchDIBits(hdc->m_hDC, 1, (int)(0.5f*(m_oldRect.bottom-fx*BMPinfoB.bmiHeader.biHeight))+1, (int)(fx*BMPinfoB.bmiHeader.biWidth+0.5f), (int)(fx*BMPinfoB.bmiHeader.biHeight+0.5f), 0, 0, BMPinfoB.bmiHeader.biWidth, BMPinfoB.bmiHeader.biHeight, BMPbitsB, &BMPinfoB, DIB_RGB_COLORS, SRCCOPY);
	  } else {
	    StretchDIBits(hdc->m_hDC, (int)(0.5f*(m_oldRect.right -fy*BMPinfoB.bmiHeader.biWidth ))+1, 1, (int)(fy*BMPinfoB.bmiHeader.biWidth+0.5f), (int)(fy*BMPinfoB.bmiHeader.biHeight+0.5f), 0, 0, BMPinfoB.bmiHeader.biWidth, BMPinfoB.bmiHeader.biHeight, BMPbitsB, &BMPinfoB, DIB_RGB_COLORS, SRCCOPY);
	  }
      glClear(GL_DEPTH_BUFFER_BIT);
    } else {
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

    glMatrixMode(GL_PROJECTION); glLoadIdentity();

    if(VMDoc->VMF.Sky){
      if((DisplayMode & VM_SHADED) && (DisplayMode & VM_TEXTURED)){
	    if(GScreen.V.h){
          if(DisplayMode & VM_ANTIALIAS){
	        switch(aa){
	          case 0: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)), -0.25, -0.25); break;
	          case 1: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)),  0.25, -0.25); break;
	          case 2: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)),  0.25,  0.25); break;
	          case 3: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)), -0.25,  0.25); break;
			}
		  } else gluPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)));
		} else {
	      glOrtho(-(GScreen.V.Zoom), (GScreen.V.Zoom), -(GScreen.V.Zoom)/Aspect, (GScreen.V.Zoom)/Aspect, max(1.0, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-10.0f*(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+10.0f*(VMDoc->WRD)));
		}

        glMatrixMode(GL_MODELVIEW); glLoadIdentity();

        //gluLookAt applied on the MODELVIEW MATRIX (julien)
        gluLookAt(GScreen.V.Location.x, GScreen.V.Location.y, GScreen.V.Location.z,
                  GScreen.V.Target.x,   GScreen.V.Target.y,   GScreen.V.Target.z,
  	      	      GScreen.V.Up.x,       GScreen.V.Up.y,       GScreen.V.Up.z);

        glDisable(GL_DEPTH_TEST);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL   );
          SetTexture(VMDoc->VMF.Sky_map, GL_DECAL, 0);
		  DrawObj(VMDoc->VMF.Sky);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	  }
	}
    
	if(GScreen.V.h){
      if(DisplayMode & VM_ANTIALIAS){
	    switch(aa){
	      case 0: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)), -0.25, -0.25); break;
	      case 1: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)),  0.25, -0.25); break;
	      case 2: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)),  0.25,  0.25); break;
	      case 3: accPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)), -0.25,  0.25); break;
	    }
	  } else gluPerspective(Rad2Deg*2.0*atan(GScreen.V.h/2.0), Aspect, max(0.1, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)));
    } else {
	  glOrtho(-(GScreen.V.Zoom), (GScreen.V.Zoom), -(GScreen.V.Zoom)/Aspect, (GScreen.V.Zoom)/Aspect, max(1.0, 0.8*(Betrag(GScreen.V.Location-(VMDoc->WMP))-(VMDoc->WRD))), 1.25*(Betrag(GScreen.V.Location-(VMDoc->WMP))+(VMDoc->WRD)));
    }

    glMatrixMode(GL_MODELVIEW); glLoadIdentity();

    //gluLookAt applied on the MODELVIEW MATRIX (julien)
    gluLookAt(GScreen.V.Location.x, GScreen.V.Location.y, GScreen.V.Location.z,
			  GScreen.V.Target.x,   GScreen.V.Target.y,   GScreen.V.Target.z,
  			  GScreen.V.Up.x,       GScreen.V.Up.y,       GScreen.V.Up.z);

	glEnable(GL_DEPTH_TEST); glDepthFunc(GL_LESS);

    LtPtr Lt;
	int   Li;

	if(!(VMDoc->Lights)){
	  PrimPtr p = VMDoc->VMF.Base;
      while(p){
        if((p->idnt==POLYMESH) || (p->idnt==CUBE) || (p->idnt==SPHERE) || (p->idnt==CYLINDER) || (p->idnt==TORUS) || (p->idnt==FUNCTION) || (p->idnt==PARABOLOID) || (p->idnt==RING)){
		  if(((Surface *)p)->Mat->Emission.R > 0.01f){
            if(p->idnt==SPHERE){
			  Lt=new Lite;
			  Lt->p=vector(0.0f, 0.0f, 0.0f)*((Surface *)p)->TM;
			  Lt->v=((Surface *)p)->Mat->Emission.R;
			  Lt->next=VMDoc->Lights; VMDoc->Lights=Lt;
			}
            if(p->idnt==POLYMESH){
			  Lt=new Lite; Lt->p=vector(0.0f, 0.0f, 0.0f);
			  for(int i=0; i<((PolyMesh *)p)->nv; i++){
			    Lt->p = Lt->p + ((PolyMesh *)p)->pts[i];
			  }
			  Lt->p = Lt->p / ((PolyMesh *)p)->nv * ((Surface *)p)->TM;
			  Lt->v=((Surface *)p)->Mat->Emission.R;
			  Lt->next=VMDoc->Lights; VMDoc->Lights=Lt;
			}
		  }
		}
        p=p->succ;
	  }
      if(!(VMDoc->Lights)){
		VMDoc->Lights=new Lite; VMDoc->Lights->next=NULL; VMDoc->Lights->v=-1.0f;
	  }
	}
	if(VMDoc->Lights->v < 0.0f){
	  float   d = Betrag(GScreen.V.Target - GScreen.V.Location);
      vector  N = unit  (GScreen.V.Target - GScreen.V.Location);
      vector  U = N*GScreen.V.Up;
      VMDoc->Lights->p = LPos = GScreen.V.Location - d*N + d*U + d*GScreen.V.Up;
	} else LPos = VMDoc->Lights->p;
  
    int maxlights; 
	GLfloat light_position[] = { 1.0f, 1.0f, 1.0f, 1.0f };
    GLfloat light_ambient[]  = { 0.0f, 0.0f, 0.0f, 1.0f };
    GLfloat light_diffuse[]  = { 1.0f, 1.0f, 1.0f, 1.0f };
    light_ambient[0]=VMDoc->VMF.Ambient.R;
    light_ambient[1]=VMDoc->VMF.Ambient.G;
    light_ambient[2]=VMDoc->VMF.Ambient.B;
    glLightModeli (GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
    glLightModeli (GL_LIGHT_MODEL_TWO_SIDE, 1);
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient);
	glGetIntegerv(GL_MAX_LIGHTS, &maxlights);
    glEnable(GL_NORMALIZE);
    glEnable(GL_LIGHTING);

    for(Lt=VMDoc->Lights,Li=0; Lt; Lt=Lt->next,Li++){
	  light_position[0]=Lt->p.x; light_position[1]=Lt->p.y; light_position[2]=Lt->p.z;
	  light_diffuse[0]=((Lt->v < 0.0f)? 1.0f : Lt->v);
	  light_diffuse[1]=((Lt->v < 0.0f)? 1.0f : Lt->v);
	  light_diffuse[2]=((Lt->v < 0.0f)? 1.0f : Lt->v);
	  if(Li<maxlights){
        glLightfv(GL_LIGHT0+Li, GL_POSITION, light_position);
        glLightfv(GL_LIGHT0+Li, GL_DIFFUSE,  light_diffuse );
        glLightfv(GL_LIGHT0+Li, GL_SPECULAR, light_diffuse );
        glEnable(GL_LIGHT0+Li);
	  }
	}

	PrimPtr p = VMDoc->VMF.Base;
    while(p){
      DrawObj(p);
      p=p->succ;
    }
    if(DisplayMode & VM_ANTIALIAS){ glReadBuffer(GL_BACK); glAccum(GL_ACCUM, 0.25f); }
  }
  glFinish();
  if(DisplayMode & VM_ANTIALIAS){
    glDrawBuffer(GL_FRONT);
	glAccum(GL_RETURN, 1.0f);
  } else {
    glGetIntegerv(GL_DRAW_BUFFER, &DrwB);
    if(DrwB==GL_BACK) SwapBuffers(wglGetCurrentDC());
  }

  DisplayMode &= ~VM_REBUILD; bBusy = FALSE;
}


CGraphWin GraphWin;


/////////////////////////////////////////////////////////////////////////////
// CGraphWin message handlers

BOOL CGraphWin::OnEraseBkgnd(CDC* pDC) { return TRUE; }


int CGraphWin::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
  if (CWnd::OnCreate(lpCreateStruct) == -1) return -1;
  Init();   // initialize OpenGL
  CDC *hdc; hdc = GetDC(); CBM.CreateCompatibleBitmap(hdc,16,16); ReleaseDC(hdc);
  return 0;
}


void CGraphWin::OnPaint() 
{
  CPaintDC dc(this); // device context for painting

  DrawScene(&dc);
  
  if(VMDoc->VLList){
	VLPtr vl = VMDoc->VLList;
	while(vl){
	  if(vl->edg){
	    DrawELink(&dc, vl->pg, vl->pe, vl->pp);
	  } else {
	    DrawVLink(&dc, vl->pg, vl->pp);
	  }
	  vl = vl->next;
	}
  }

}


void CGraphWin::OnDestroy() 
{
  HGLRC   hrc;

  hrc = ::wglGetCurrentContext();
  ::wglMakeCurrent(NULL, NULL);
  if (hrc) ::wglDeleteContext(hrc);
  if (m_pOldPalette) m_pDC->SelectPalette(m_pOldPalette, FALSE);
  if (m_pDC) delete m_pDC;
  CWnd::OnDestroy();
}

