// Virtual Modeller V0.5GL  for MS Visual C++ V4.2   27.10.1998  Michael Granz

#include "stdafx.h"
#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "TextWin.h"

extern void DelIsoVert(PrimPtr p, BOOL stat);

typedef struct LayList   *LayPtr;
struct LayList { char name[32]; LayPtr next; };

typedef struct FaceList  *FacePtr;
struct FaceList { vector p0,p1,p2,p3; LayPtr Lay; FacePtr next; };

void CVModelDoc::Import3DS(char *filename, ViewPtr V)
{
  FILE                *infile;
  FILE                *lgfile;
  unsigned short int   magic;
  unsigned short int   num;
  long           int   size,r1,r2,r3,r4;
  char                 name[256];
  byte                 bits[96];
  MatPtr               mats[256];
  matrix               tfm;
  int                  Maps=0, mgf=0, mgo=0, mo, mc;
  int                  i,j;
  short int            mi;

  #ifdef _DEBUG
    int log=1;
  #else
	int log=0;
  #endif

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  infile = fopen(filename,"rb");
  if(log) lgfile = fopen("D:\\Temp\\3DS_import_log.txt","wt");
  if(infile){
    TextWin.Output(0,"reading file ...\n"); if(log) fprintf(lgfile,"reading file '%s'...\n",filename);
    fread(&magic,2,1,infile); fread(&size,4,1,infile);
    if((magic == 0x4D4D) || (magic == 0xC23D)){
      //TextWin.Output(0,"- filesize: %ld bytes.\n",size);
      if(log) fprintf(lgfile,"- filesize: %ld bytes.\n",size);
      while(!feof(infile)){
        fread(&magic,2,1,infile); fread(&size,4,1,infile);
        if(!feof(infile) && size>=6){
          //TextWin.Output(0,"- Chunk '%4x', size %ld bytes\n",magic,size);
          if(log) fprintf(lgfile,"- Chunk '%4x', size %ld bytes\n",magic,size);
          if(magic == 0x3D3D){
            r1=size-6;
            while((!feof(infile)) && (r1>0)){
              fread(&magic,2,1,infile); fread(&size,4,1,infile); r1-=size;
              if(!feof(infile) && size>=6){
                switch(magic){
				  case 0xAFFF:{
                    //TextWin.Output(0,"  - Chunk '%4x', size %ld bytes\n",magic,size);
                    if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes:",magic,size);
                    MatPtr mat = new Material; mat->next=VMF.MatList; VMF.MatList=mat;
                    // 3D-Studio default-material:
                    mat->Reflection.R = 0.69803F; mat->Reflection.G = 0.69803F; mat->Reflection.B = 0.69803F;
                    mat->Specularity  = 0.89803F; mat->PBExp = 36.0F;
                    r2=size-6;
                    while((!feof(infile)) && (r2>0)){
                      fread(&magic,2,1,infile); fread(&size,4,1,infile); r2-=size;
                      if(!feof(infile)){
                        switch(magic){
				          case 0xA000: // Material-Name
                            i=0; do { fread(&(name[i]),1,1,infile); } while(name[i++]!=0);
                            TextWin.Output(0,"  - Material '%s'\n",name);
                            if(log) fprintf(lgfile," Material '%s'\n",name);
							mat->name=strdup(name);
							break;
				          case 0xA020: // Diffuse Color
                            fread(&bits,size-6,1,infile);
							mat->Reflection.R = (float)(bits[6])/255.0f;
							mat->Reflection.G = (float)(bits[7])/255.0f;
							mat->Reflection.B = (float)(bits[8])/255.0f;
							break;
				          case 0xA030: // Specular Color
                            fread(&bits,size-6,1,infile);
							//mat->Os.R = (float)(bits[6])/255.0f;
							//mat->Os.G = (float)(bits[7])/255.0f;
							//mat->Os.B = (float)(bits[8])/255.0f;
							break;
				          case 0xA050: // Transparency
                            fread(&bits,size-6,1,infile);
							//mat->Kt   = (float)(bits[6])/100.0f;
							break;
				          case 0xA040: // Shininess
                            fread(&bits,size-6,1,infile);
							mat->Specularity = (float)(bits[6])/100.0f;
							break;
				          case 0xA041: // Shininess Intensity
                            fread(&bits,size-6,1,infile);
							if(bits[6]<20){
							  mat->PBExp = 2.0f+(float)(bits[6])/5.0f;
							} else {
							  if(bits[6]>50){
							    mat->PBExp = 50.0f+(float)pow((double)(bits[6])/50.0, 3.4);
							  } else {
							    mat->PBExp = 1.5f*(float)(bits[6])-24.0f;
							  }
							}
							break;
				          case 0xA081: // Doublesided
							//mat->R = 0.001f;
							break;
				          case 0xA200: // Texture Map
                            fread(&bits,size-6,1,infile);
							if(bits[6]){
			                  TexMapPtr t = VMF.TexList; while(t && stricmp((char *)&(bits[14]), t->tname)) t=t->next;
			                  if(!t){
			                    t = new TexMap; t->next=VMF.TexList; VMF.TexList=t;
			                    t->tname=strdup((char *)&(bits[14])); t->tbits=NULL; t->glbits=NULL; t->tflags=0;
			                  }
			                  mat->Texture=t;
                              TextWin.Output(0,"    - Texture-Map    '%s'\n", &(bits[14]));
							}
                            if(log) fprintf(lgfile,"    - Texture-Map '%s' (%d%%)\n", &(bits[14]), bits[6]);
							break;
				          case 0xA230: // Bump Map
                            fread(&bits,size-6,1,infile);
							if(bits[6]){
							  Maps=1;
                              TextWin.Output(0,"    - Bump-Map       '%s'\n", &(bits[14]));
							}
                            if(log) fprintf(lgfile,"    - Bump-Map '%s' (%d%%)\n", &(bits[14]), bits[6]);
							break;
				          case 0xA220: // Reflection Map
                            fread(&bits,size-6,1,infile);
							if(bits[6]){
							  Maps=1;
                              TextWin.Output(0,"    - Reflection-Map '%s'\n", &(bits[14]));
							}
                            if(log) fprintf(lgfile,"    - Reflection-Map '%s' (%d%%)\n", &(bits[14]), bits[6]);
							break;
						  default:
                            //TextWin.Output(0,"    - Chunk '%4x', size %ld bytes\n",magic,size);
                            if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes: ",magic,size); byte byt[32];
			                for(i=0; i<min(28,size-6); i++){
				              fread(&(byt[i]),1,1,infile);
				              if(log) fprintf(lgfile," %3d", byt[i]);
							} byt[i]=0;
			                if(log) fprintf(lgfile," '");
							for(j=0; j<i; j++){
			                  if(log) fprintf(lgfile,"%c",byt[j]);
							}
			                if(log) fprintf(lgfile,"'\n");
                            if(size-6-i > 0) fseek(infile,size-6-i,SEEK_CUR);
						}
					  }
					}
					break;
                  }
				  case 0x4000:
                    i=0; mo=0; mc=0; do fread(&(name[i]),1,1,infile); while(name[i++]!=0);
                    //TextWin.Output(0,"  - Chunk '%4x', size %ld bytes: '%s'\n",magic,size,name);
                    if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes: '%s'\n",magic,size,name);
                    r2=size-6-i;
                    while((!feof(infile)) && (r2>0)){
                      fread(&magic,2,1,infile); fread(&size,4,1,infile); r2-=size;
                      if(!feof(infile)){
                        if(magic == 0x4100){
                          //TextWin.Output(0,"    - Chunk '%4x', size %ld bytes\n",magic,size);
                          if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes\n",magic,size);
                          TextWin.Output(0,"  - Object '%s'",name);
                          r3=size-6;
                          int triflags;
						  MatPtr  mat = NULL;
                          PolyMesh *p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
                          p->name=strdup(name); tfm.InitIdentMat();
                          while((!feof(infile)) && (r3>0)){
                            fread(&magic,2,1,infile); fread(&size,4,1,infile); r3-=size;
                            if(!feof(infile)){
                              switch(magic){
                                case 0x4110: // Vertex-List
                                  fread(&num,2,1,infile);
                                  //TextWin.Output(0,"      - Chunk '%4x', size %ld bytes, %d vertices\n",magic,size,num);
                                  if(log) fprintf(lgfile,"      - Chunk '%4x', size %ld bytes, %d vertices\n",magic,size,num);
                                  TextWin.Output(0,": %d vertices",num);
                                  p->nv=num; p->pts=new(vector[p->nv]);
                                  for(i=0; i<(p->nv); i++){
                                    fread(&(p->pts[i].x),4,1,infile);
                                    fread(&(p->pts[i].y),4,1,infile);
                                    fread(&(p->pts[i].z),4,1,infile);
                                  }
                                  break;
                                case 0x4140: // Texture Coords
                                  fread(&num,2,1,infile);
                                  //TextWin.Output(0,"      - Chunk '%4x', size %ld bytes, %d vertices\n",magic,size,num);
                                  if(log) fprintf(lgfile,"      - Chunk '%4x', size %ld bytes, %d texcoords\n",magic,size,num);
                                  TextWin.Output(0,", %d texture coords",num);
                                  p->nt=num; p->txc=new(vector[p->nt]);
                                  for(i=0; i<(p->nt); i++){
                                    fread(&(p->txc[i].x),4,1,infile);
                                    fread(&(p->txc[i].y),4,1,infile);
									        p->txc[i].z = 0.0f;
                                  }
                                  break;
                                case 0x4160:
                                  if(log) fprintf(lgfile,"      - Chunk '%4x', size %ld bytes (transformation matrix)\n",magic,size);
                                  fread(&(tfm.M11),1,4,infile); fread(&(tfm.M12),1,4,infile); fread(&(tfm.M13),1,4,infile);
                                  fread(&(tfm.M21),1,4,infile); fread(&(tfm.M22),1,4,infile); fread(&(tfm.M23),1,4,infile);
                                  fread(&(tfm.M31),1,4,infile); fread(&(tfm.M32),1,4,infile); fread(&(tfm.M33),1,4,infile);
                                  fread(&(tfm.M41),1,4,infile); fread(&(tfm.M42),1,4,infile); fread(&(tfm.M43),1,4,infile);
                                  if(log) fprintf(lgfile,"        / %8.3f, %8.3f, %8.3f \\\n",tfm.M11,tfm.M12,tfm.M13);
                                  if(log) fprintf(lgfile,"        | %8.3f, %8.3f, %8.3f |\n", tfm.M21,tfm.M22,tfm.M23);
                                  if(log) fprintf(lgfile,"        | %8.3f, %8.3f, %8.3f |\n", tfm.M31,tfm.M32,tfm.M33);
                                  if(log) fprintf(lgfile,"        \\ %8.3f, %8.3f, %8.3f /\n",tfm.M41,tfm.M42,tfm.M43);
								  if(p->pts){
									matrix Mi=tfm.Inverse();
									for(i=0; i<(p->nv); i++){
									  p->pts[i]=p->pts[i]*Mi;
									}
								  }
                                  break;
                                case 0x4120: // TriFace-List
                                  fread(&num,2,1,infile);
                                  //TextWin.Output(0,"      - Chunk '%4x', size %ld bytes, %d faces\n",magic,size,num);
                                  if(log) fprintf(lgfile,"      - Chunk '%4x', size %ld bytes, %d faces\n",magic,size,num);
                                  TextWin.Output(0,", %d faces",num);
                                  p->nf = num; p->pfv=new(ushort[3*(p->nf)]);
								  for(i=0; i<(p->nf); i++){
                                    fread(&(p->pfv[3*i+2]),2,1,infile);
                                    fread(&(p->pfv[3*i+1]),2,1,infile);
                                    fread(&(p->pfv[3*i  ]),2,1,infile);
                                    fread(&triflags       ,2,1,infile);
                                  }
								  if(p->txc){
                                    p->pft=new(ushort[3*(p->nf)]);
									memcpy(p->pft, p->pfv, 3*(p->nf)*sizeof(ushort));
								  }
                                  r4=size-6-2-8*num;
                                  while((!feof(infile)) && (r4>0)){
                                    fread(&magic,2,1,infile); fread(&size,4,1,infile); r4-=size;
                                    if(!feof(infile)){
									  long unsigned int sgrps; int ogrps;
                                      switch(magic){
                                        case 0x4130: // Material-Application
                                          //TextWin.Output(0,"        - Chunk '%4x', size %ld bytes\n",magic,size);
                                          if(log) fprintf(lgfile,"        - Chunk '%4x', size %ld bytes (material application)\n",magic,size);
                                          i=0; do { fread(&(name[i]),1,1,infile); } while(name[i++]!=0);
                                          TextWin.Output(0,", '%s'",name);
                                          mat = VMF.MatList; while(mat && strcmp(mat->name,name)) mat=mat->next;
										  mats[mc]=mat;
										  if(mc){
                                            fread(&mi,2,1,infile);
											for(j=0; j<((size-8-i)/2); j++){
                                              fread(&mi,2,1,infile);
											  p->mat[mi]=mc;
											}
										    mc++;
										  } else {
											p->mat=new(byte[p->nf]);
											for(j=0; j<(p->nf); j++) p->mat[j]=0;
										    fseek(infile,size-6-i,SEEK_CUR); mc=1;
										  }
										  break;
                                        case 0x4150: // Smoothing-Groups
                                          //TextWin.Output(0,"        - Chunk '%4x', size %ld bytes\n",magic,size);
                                          if(log) fprintf(lgfile,"        - Chunk '%4x', size %ld bytes (smoothing groups)\n",magic,size);
                                          p->grp=new(byte[p->nf]);
								          for(i=0; i<(p->nf); i++){
                                            fread(&sgrps,4,1,infile);
										    if(sgrps){
										  	  for(j=1; !(sgrps & 1); j++) sgrps = sgrps>>1;
										      if(sgrps>1){ mgf++; mo=1; }
											  p->grp[i]=j;
										    } else {
										      p->grp[i]=0;
										    }
										    if(i){
										      if(p->grp[i] != ogrps) ogrps=-1;
										    } else {
										      ogrps=p->grp[i];
										    }
                                          }
										  if(ogrps==-1){
										    p->pmflags |= SURF_GROUPED;
										  } else {
										    if(ogrps) p->pmflags |= SURF_ROUNDED;
										    delete p->grp; p->grp=NULL;
										  }
                                          break;
                                        default: /*TextWin.Output(0,"        - Chunk '%4x', size %ld bytes\n",magic,size);*/ fseek(infile,size-6,SEEK_CUR);
                                                 if(log) fprintf(lgfile,"        - Chunk '%4x', size %ld bytes\n",magic,size);
                                      }
                                    }
                                  }
                                  break;
                                default: /*TextWin.Output(0,"      - Chunk '%4x', size %ld bytes\n",magic,size);*/ fseek(infile,size-6,SEEK_CUR);
                                         if(log) fprintf(lgfile,"      - Chunk '%4x', size %ld bytes\n",magic,size);
                              }
                            }
                          }

                          if(p->pts && p->pfv){
						    if(!mc){
						      mat = VMF.MatList; while(mat && strcmp(mat->name,"default")) mat=mat->next;
                              if(!mat){
                                mat=new Material; mat->next=VMF.MatList; VMF.MatList=mat;
                                mat->name=strdup("default");
                                // 3D-Studio default-material:
                                mat->Reflection.R = 0.69803F; mat->Reflection.G = 0.69803F; mat->Reflection.B = 0.69803F;
                                mat->Specularity  = 0.89803F; mat->PBExp = 36.0F;
                              }
                              p->Mat=mat;
						    } else {
                              if(mc>1){
						        p->mtl = new(MatPtr[mc]); p->nm=mc;
							    for(i=0; i<mc; i++) p->mtl[i]=mats[i];
						      } else {
							    delete(p->mat); p->mat=NULL;
						      }
						      p->Mat=mats[0];
							}

							//if(p->Mat->R > 0.0f) p->pmflags |= DOUBLESIDED;

                            p->TM=tfm;               // transformation
                            p->Tb = p->TM.Inverse(); // calculate backtransformation matrix
                            p->NM = p->TM.NTrans();  // calculate transformation matrix for normals

                            TextWin.Output(0," - done.\n");
						  } else {
							if(p){
                              VMF.Base=p->succ; if(VMF.Base) VMF.Base->pred=NULL;
							  Kill1(p);
							}
							TextWin.Output(0," - skipped.\n");
						  }
                        } else {
                          //TextWin.Output(0,"    - Chunk '%4x', size %ld bytes\n",magic,size);
                          if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes\n",magic,size);
                          fseek(infile,size-6,SEEK_CUR);
                        }
                      }
                    }
                    if(mo) mgo++;
					break;
                  default:
                    //TextWin.Output(0,"  - Chunk '%4x', size %ld bytes\n",magic,size);
                    if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes\n",magic,size);
                    fseek(infile,size-6,SEEK_CUR);
                }
              }
            }
          } else if(magic == 0xB000){
            r1=size-6; vector pivot=vector(0.0f, 0.0f, 0.0f); vector pos=vector(0.0f, 0.0f, 0.0f); vector scale=vector(1.0f, 1.0f, 1.0f); vector axis; float rot=0.0f; char nam[24]; nam[0]=0;
            while((!feof(infile)) && (r1>0)){
              fread(&magic,2,1,infile); fread(&size,4,1,infile); r1-=size;
              if(!feof(infile) && size>=6){
				if(magic == 0xB00A){
                  if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				  for(i=0; i<size-6; i++){
				    fread(&(byt[i]),1,1,infile);
				  } if(log) fprintf(lgfile," %d, %d, '%s'\n",byt[0], byt[1], &(byt[2]));
				} else if(magic == 0xB002){
                  if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes\n",magic,size);
                  r2=size-6;
                  while((!feof(infile)) && (r2>0)){
                    fread(&magic,2,1,infile); fread(&size,4,1,infile); r2-=size;
                    if(!feof(infile) && size>=6){
				      if(magic == 0xB013){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size);
				        fread(&(pivot.x),4,1,infile);
				        fread(&(pivot.y),4,1,infile);
				        fread(&(pivot.z),4,1,infile);
						if(log) fprintf(lgfile," pivot = (%f, %f, %f)\n",pivot.x, pivot.y, pivot.z);
					  } else if(magic == 0xB020){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				        for(i=0; i<20; i++){
				          fread(&(byt[i]),1,1,infile);
                          if(log) fprintf(lgfile," %d",byt[i]);
						}
				        fread(&(pos.x),4,1,infile);
				        fread(&(pos.y),4,1,infile);
				        fread(&(pos.z),4,1,infile);
						if(log) fprintf(lgfile," position = (%f, %f, %f)\n",pos.x, pos.y, pos.z);
					  } else if(magic == 0xB021){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				        for(i=0; i<20; i++){
				          fread(&(byt[i]),1,1,infile);
                          if(log) fprintf(lgfile," %d",byt[i]);
						}
				        fread(&(rot) ,4,1,infile);
				        fread(&(axis.x),4,1,infile);
				        fread(&(axis.y),4,1,infile);
				        fread(&(axis.z),4,1,infile);
						if(log) fprintf(lgfile," rotation = angle: %f, axis = (%f, %f, %f)\n",rot, axis.x, axis.y, axis.z);
                        if(size>42) fseek(infile,size-42,SEEK_CUR);
					  } else if(magic == 0xB022){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				        for(i=0; i<20; i++){
				          fread(&(byt[i]),1,1,infile);
                          if(log) fprintf(lgfile," %d",byt[i]);
						}
				        fread(&(scale.x),4,1,infile);
				        fread(&(scale.y),4,1,infile);
				        fread(&(scale.z),4,1,infile);
						if(log) fprintf(lgfile," scale = (%f, %f, %f)\n",scale.x, scale.y, scale.z);
					  } else if(magic == 0xB014){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); float v;
				        fread(&v,4,1,infile); if(log) fprintf(lgfile," %f"  ,v);
				        fread(&v,4,1,infile); if(log) fprintf(lgfile,",%f"  ,v);
				        fread(&v,4,1,infile); if(log) fprintf(lgfile,",%f"  ,v);
				        fread(&v,4,1,infile); if(log) fprintf(lgfile,",%f"  ,v);
				        fread(&v,4,1,infile); if(log) fprintf(lgfile,",%f"  ,v);
				        fread(&v,4,1,infile); if(log) fprintf(lgfile,",%f\n",v);
					  } else if(magic == 0xB010){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); int i;
				        for(i=0; i<size-6; i++){
				          fread(&(nam[i]),1,1,infile);
						} if(log) fprintf(lgfile," '%s'\n",nam);
					  } else if(magic == 0xB011){
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); char des[24]; int i;
				        for(i=0; i<size-6; i++){
				          fread(&(des[i]),1,1,infile);
						} if(log) fprintf(lgfile," '%s'\n",des);
					  } else {
                        if(log) fprintf(lgfile,"    - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				        for(i=0; i<size-6; i++){
				          fread(&(byt[i]),1,1,infile);
                          if(log) fprintf(lgfile," %d",byt[i]);
						} if(log) fprintf(lgfile," (%s)\n",byt);
					  }
					}
				  }
				  if(nam[0]){
                    SurfPtr p = (SurfPtr)VMF.Base;
					while(p){
					  if(!stricmp(p->name,nam)){
						p->TM.InitIdentMat();
                        p->TM.Translate(-pivot.x, -pivot.y, -pivot.z); p->TM.Scale(scale.x, scale.y, scale.z); p->TM.Translate(pivot.x, pivot.y, pivot.z);
						if(rot) p->TM=p->TM*InitAxisRot(pivot.x, pivot.y, pivot.z, axis.x, axis.y, axis.z, rot);
						p->TM.Translate(pos.x, pos.y, pos.z); p->TM.Translate(-pivot.x, -pivot.y, -pivot.z);
                        p->Tb = p->TM.Inverse(); // calculate backtransformation matrix
                        p->NM = p->TM.NTrans();  // calculate transformation matrix for normals
					  }
					  p=(SurfPtr)p->succ;
					}
				  }
				} else {
                  if(log) fprintf(lgfile,"  - Chunk '%4x', size %ld bytes:",magic,size); byte byt[24]; int i;
				  for(i=0; i<min(20,size-6); i++){
				    fread(&(byt[i]),1,1,infile);
                    if(log) fprintf(lgfile," %d",byt[i]);
				  } if(log) fprintf(lgfile," (%s)\n",&(byt[14]));
                  fseek(infile,size-6-i,SEEK_CUR);
				}
			  }
			}
          } else fseek(infile,size-6,SEEK_CUR);
        }
      }
      PrimPtr p = VMF.Base;
      while(p){
        if(p->idnt==POLYMESH){
          for(int i=0; i<((PolyMesh *)p)->nv; i++){
			vector v=((PolyMesh *)p)->pts[i]*((PolyMesh *)p)->TM;
            if(v.x > maxx) maxx=v.x; if(v.x < minx) minx=v.x;
            if(v.y > maxy) maxy=v.y; if(v.y < miny) miny=v.y;
            if(v.z > maxz) maxz=v.z; if(v.z < minz) minz=v.z;
		  }
		}
        p=p->succ;
	  }
      vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
      float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
      V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
      V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
	  V->SetViewAngle(0.5F);
      vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
      strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
      WorldValid=FALSE;
	  if(mgf ) TextWin.Output(3,"- WARNING: %d faces in %d objects had multiple smoothing groups assigned.\n",mgf,mgo);
      if(Maps) TextWin.Output(3,"- WARNING: Scene contains unhandled maps!\n");
    } else TextWin.Output(3,"- ERROR: no 3D-Studio Mesh or Project!\n");
    fclose(infile);
  } else TextWin.Output(3,"- ERROR: couldn't open input-file!\n");
  if(log) fclose(lgfile);
}


void GetDEMToken(char *a, FILE *file)
{
  char c; int i=0;
  do c=fgetc(file); while( c==' ' || c==',' || c==':' || c=='\t' || c=='\n' );
  if (!(feof(file)))
    do{ a[i++]=c; c=fgetc(file); }while( c!=' ' && c!=',' && c!=':' && c!='\t' && c!='\n' );
  a[i]=0;
}

void CVModelDoc::Import4DS(char *filename, ViewPtr V)
{
  FILE *fdsfile;
  unsigned char byt;
  float num;
  int i,x,y;
  int mp=0; MatPtr mats[80];
  MatPtr  mat = NULL;
  PolyMesh *p;
  unsigned char nmat,objs,facs;
  int nv; vector *pts;
  int nt; vector *txc;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  fdsfile = fopen(filename,"rb");
  if(fdsfile){
    for(x=0; x<14; x++){
      fread(&byt,1,1,fdsfile);
	}
    fread(&nmat,1,1,fdsfile); TextWin.Output(0," - %d Materials:\n",(int)nmat);
    fread(&byt,1,1,fdsfile);  unsigned char tlen; char tnam[24];
    for(i=0; i<nmat; i++){
      for(x=0; x<44; x++){
        fread(&byt,1,1,fdsfile);
	  }
      fread(&tlen,1,1,fdsfile); fread(tnam,1,tlen,fdsfile); tnam[tlen]=0;
      mat=new Material; mat->next=VMF.MatList; VMF.MatList=mat; mat->name=strdup(tnam);
	  TexMapPtr t = new TexMap; t->next=VMF.TexList; VMF.TexList=t;
	  t->tname=strdup("C:\\Spiele\\Mafia\\maps\\12345678.bmp"); strcpy(t->tname+21,tnam); t->tbits=NULL; t->tflags=0;
	  mat->Texture=t;
	  mats[mp++]=mat; TextWin.Output(0,"   '%s'\n",tnam);
	}
    for(x=0; x<49; x++){
      fread(&byt,1,1,fdsfile);
	}
    unsigned char nlen; fread(&nlen,1,1,fdsfile); char onam[24]; fread(onam,1,nlen,fdsfile); onam[nlen]=0; //p->name=strdup("4DS");
    for(x=0; x<8; x++){
      fread(&byt,1,1,fdsfile);
	}        
    fread(&byt,1,1,fdsfile); nv=byt; pts=new(vector[nv]);
                             nt=byt; txc=new(vector[nt]);
    TextWin.Output(0," - %d Vertices\n",nv);
    fread(&byt,1,1,fdsfile);
    for(i=0; i<nv; i++){
      fread(&(pts[i].x),1,4,fdsfile);
      fread(&(pts[i].y),1,4,fdsfile);
      fread(&(pts[i].z),1,4,fdsfile);
      fread(&num,1,4,fdsfile);
      fread(&num,1,4,fdsfile);
      fread(&num,1,4,fdsfile);
      fread(&(txc[i].x),1,4,fdsfile);
      fread(&(txc[i].y),1,4,fdsfile); txc[i].y = 1.0f-txc[i].y; txc[i].z = 0.0f;
      if(pts[i].x > maxx) maxx=pts[i].x;
      if(pts[i].y > maxy) maxy=pts[i].y;
      if(pts[i].z > maxz) maxz=pts[i].z;
      if(pts[i].x < minx) minx=pts[i].x;
      if(pts[i].y < miny) miny=pts[i].y;
      if(pts[i].z < minz) minz=pts[i].z;
	}
    fread(&objs,1,1,fdsfile); TextWin.Output(0," - %d Objects:\n",(int)objs);
    for(i=0; i<objs; i++){
      p = new PolyMesh; p->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=p; VMDoc->VMF.Base=p;
      fread(&facs,1,1,fdsfile); TextWin.Output(0,"   %d Triangles",(int)facs);
      p->nv=nv  ; p->pts=new(vector[  p->nv]); memcpy(p->pts, pts, nv*sizeof(vector));
      p->nt=nt  ; p->txc=new(vector[  p->nt]); memcpy(p->txc, txc, nt*sizeof(vector));
      p->nf=facs; p->pfv=new(ushort[3*p->nf]); p->name=strdup(onam);
                  p->pft=new(ushort[3*p->nf]);
      for(y=0; y<facs; y++){
        fread(&byt,1,1,fdsfile);
        fread(&byt,1,1,fdsfile); p->pft[3*y+2]=p->pfv[3*y+2]=byt;
        fread(&byt,1,1,fdsfile);
        fread(&byt,1,1,fdsfile); p->pft[3*y+1]=p->pfv[3*y+1]=byt;
        fread(&byt,1,1,fdsfile);
        fread(&byt,1,1,fdsfile); p->pft[3*y  ]=p->pfv[3*y  ]=byt;
	  }
      fread(&byt,1,1,fdsfile);
      fread(&byt,1,1,fdsfile); TextWin.Output(0," - Material %d\n",(int)byt); p->Mat=mats[byt-1];
      fread(&byt,1,1,fdsfile);
      p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);
	}
    fclose(fdsfile);

    // set up a view
    vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
    float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
    V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
    V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
    V->SetViewAngle(0.5F);
    vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
    strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
  } else TextWin.Output(3,"- ERROR: couldn't open input-file!\n");
}

void CVModelDoc::ImportDEM(char *filename, ViewPtr V)
{
  FILE      *infile;
  PolyMesh  *p;
  char       tk[80];
  int        i,x,y,u,v;
  long int   hv;

  int   rng= 8;
  float scl=20.0f*rng;
  int   scp=1200 /rng;
  /*int   rng= 3;
  float scl=30.0f*rng;
  int   scp=600 /rng;*/

  infile = fopen(filename,"rt");
  if(infile){
	short int *hg=new(short int[1201*1201]);
    p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
    p->name=strdup("DEM"); p->pmflags |= SURF_ROUNDED;
	p->nv=scp*scp; p->pts=new(vector[p->nv]);
    p->nf=(scp-1)*(scp-1); p->pnv=new(byte[p->nf]); p->pfv=new(ushort[4*p->nf]);

    for(i=0; i<42; i++) GetDEMToken(tk,infile);

    for(y=0; y<1201; y++) {
      for(i=0; i<9; i++) GetDEMToken(tk,infile);
      for(x=0; x<1201; x++){
	    GetDEMToken(tk,infile); hg[1201*      y +x]=atoi(tk);
	  //GetDEMToken(tk,infile); hg[1201*(1200-y)+x]=atoi(tk);
	  }
	}

    for(y=0; y<scp; y++) {
      for(x=0; x<scp; x++){
		hv=0;
		for(u=0; u<rng; u++) {
		  for(v=0; v<rng; v++) {
			hv+=hg[1201*(rng*y+u)+(rng*x+v)];
		  }
		}
		hv/=(rng*rng);
		p->pts[scp*y+x]=vector(scl*x, scl*y, (float)hv);
		if((x>0) && (y>0)){
		  p->pnv[   (scp-1)*(y-1)+(x-1)   ]=4;
		  p->pfv[4*((scp-1)*(y-1)+(x-1))+3]=scp*(y-1)+x-1;
		  p->pfv[4*((scp-1)*(y-1)+(x-1))+2]=scp*(y-1)+x;
		  p->pfv[4*((scp-1)*(y-1)+(x-1))+1]=scp* y   +x;
		  p->pfv[4*((scp-1)*(y-1)+(x-1))  ]=scp* y   +x-1;
		}
	  }
	}

	delete hg;
    VMF.MatList=new Material; VMF.MatList->name=strdup("default");
    p->Mat=VMF.MatList;

    // default-transformation:
    p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);

	// set up a view
	vector tp = 0.5 * vector(scp*scl, scp*scl, 0.0f);
    V->Locate(tp+vector(1.0F*scp*scl, 1.5F*scp*scl, 2.5F*scp*scl));
    V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
	V->SetViewAngle(0.5F); WorldValid=FALSE;
    vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
    strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");

	fclose(infile);
  } else TextWin.Output(3,"- ERROR: couldn't open input-file!\n");
}


int intread(FILE *infile)
{
  unsigned char a[2];
  fread(a,2,1,infile);
  return a[1]+256*a[0];
}

long longread(FILE *infile)
{
  unsigned char a[4];
  long l;
  fread(a,4,1,infile);
  l=(long)a[3]+(((long)a[2])<<8)+(((long)a[1])<<16)+(((long)a[0])<<24);
  return l;
}

float floatread(FILE *infile)
{
  float f1,f2;
  fread(&f1,4,1,infile);
  *(((byte *)(&f2))  ) = *(((byte *)(&f1))+3);
  *(((byte *)(&f2))+1) = *(((byte *)(&f1))+2);
  *(((byte *)(&f2))+2) = *(((byte *)(&f1))+1);
  *(((byte *)(&f2))+3) = *(((byte *)(&f1))  );
  return f2;
}

void CVModelDoc::ImportLWO(char *filename, ViewPtr V)
{
  FILE      *infile;
  FILE      *lgfile;
  char       chunk[10];
  char       snm[128];
  MatPtr     mr[128];
  long       fpos,fl,cl,cp;
  int        fi,i,j,pn,vc,fc,nv,mc=0,mnv=0,tn=0;
  PolyMesh  *p;
  vector    *vert;
  vector    *texc;
  int       *tci;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  #ifdef _DEBUG
    int log=1;
  #else
	int log=0;
  #endif

  if(log) lgfile = fopen("D:\\Temp\\LWO_import_log.txt","wt");
  infile = fopen(filename,"rb");
  if(infile){
    TextWin.Output(0,"reading file ...\n");
    if(log) fprintf(lgfile,"reading file '%s'...\n",filename);
    fgets(chunk,5,infile); fl=longread(infile);
    if(!strcmp(chunk,"FORM")){
      fgets(chunk,5,infile);
      //TextWin.Output(0,"- IFF-Formlength %ld bytes\n",fl);
      if(log) fprintf(lgfile,"- IFF-Formlength %ld bytes\n",fl);
      if(!strcmp(chunk,"LWOB")){
        while(!feof(infile)){
          fgets(chunk,5,infile); cl=longread(infile); if(cl & 1) cl++;
          if(!feof(infile)){
            if(!strcmp(chunk,"PNTS")){
			  TextWin.Output(0,"- %ld vertices\n",pn=(cl/12));
			  if(log) fprintf(lgfile,"- %ld vertices\n",pn=(cl/12));
              vert=new(vector[pn]);
              for(i=0; i<pn; i++){
				(vert[i]).x= floatread(infile);
				(vert[i]).y= floatread(infile);
				(vert[i]).z=-floatread(infile);
                if(vert[i].x > maxx) maxx=vert[i].x;
                if(vert[i].y > maxy) maxy=vert[i].y;
                if(vert[i].z > maxz) maxz=vert[i].z;
                if(vert[i].x < minx) minx=vert[i].x;
                if(vert[i].y < miny) miny=vert[i].y;
                if(vert[i].z < minz) minz=vert[i].z;
			  }
			} else if(!strcmp(chunk,"SRFS")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes\n",chunk,cl);
			  int cp=0;
			  for(i=0; i<cl; i++){
                fread(&(snm[cp]),1,1,infile);
				if(snm[cp]==0){
				  if(cp>0){
					MatPtr mat = new Material; mat->next=VMF.MatList; VMF.MatList=mat;
			        mat->name=strdup(snm); mr[mc++]=mat;
				    if(log) fprintf(lgfile,"  '%s'\n",snm);
				  } cp=0;
				} else cp++;
			  }
			} else if(!strcmp(chunk,"SURF")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes\n",chunk,cl);
			  char snm[30]; int cp=0;
			  do{
                fread(&(snm[cp++]),1,1,infile);
			  } while(snm[cp-1]);
			  if(log) fprintf(lgfile,"  '%s'\n",snm);
			  MatPtr mp=VMF.MatList; byte byt;
              if(mp){ while(((mp->next)!=NULL)&&(strcmp(mp->name,snm)!=0)) mp=mp->next; }
              //if(strcmp(mp->name,a)!=0) TextWin.Output(3,"- Material '%s' nicht gefunden!\n",a);
              if(cp & 1){ fread(&byt,1,1,infile); cp++; }
              char subch[10]; int sl;
			  while(cp<cl){
                fgets(subch,5,infile);
				sl=intread(infile); if(sl & 1) sl++;
			    if(!strcmp(subch,"COLR")){
                  fread(&byt,1,1,infile); mp->Reflection.R=(float)byt/255.0f;
                  fread(&byt,1,1,infile); mp->Reflection.G=(float)byt/255.0f;
                  fread(&byt,1,1,infile); mp->Reflection.B=(float)byt/255.0f;
                  fread(&byt,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Color: (%f, %f, %f)\n",mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
				} else if(!strcmp(subch,"VDIF")){
				  float VDif=floatread(infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - VDIF: %f\n",VDif);
				} else if(!strcmp(subch,"SMAN")){
				  float SmAn=floatread(infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Smoothing angle: %f (%6.2f)\n",SmAn,180.0f*SmAn/3.14159265f);
				} else if(!strcmp(subch,"BTEX")){
				  char ctex[80];
				  for(i=0; i<sl; i++) fread(&(ctex[i]),1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - BTEX '%s'\n",ctex);
				} else if(!strcmp(subch,"CTEX")){
				  char ctex[80];
				  for(i=0; i<sl; i++) fread(&(ctex[i]),1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - CTEX '%s'\n",ctex);
				} else if(!strcmp(subch,"RIMG")){
				  char ctex[80];
				  for(i=0; i<sl; i++) fread(&(ctex[i]),1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - RIMG '%s'\n",ctex);
				} else if(!strcmp(subch,"TIMG")){
				  char ctex[80];
				  for(i=0; i<sl; i++) fread(&(ctex[i]),1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - TIMG '%s'\n",ctex);
				} else if(!strcmp(subch,"TALP")){
				  char ctex[80];
				  for(i=0; i<sl; i++) fread(&(ctex[i]),1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - TALP '%s'\n",ctex);
				} else {
			      if(log) fprintf(lgfile,"  - Chunk '%s', length %ld bytes: ",subch,sl);
                  /*fseek(infile,sl,SEEK_CUR);*/ cp += sl+6;
			      for(i=0; i<sl; i++){
                    fread(&byt,1,1,infile);
				    if(log) fprintf(lgfile," %d",byt);
				  }
			      if(log) fprintf(lgfile,"\n");
				}
			  }
			} else if(!strcmp(chunk,"POLS")){
              fpos=ftell(infile); cp=cl; fc=0; vc=0;
			  while(cp>0){
			    nv=intread(infile); vc+=nv; fc++;
                fseek(infile,2*nv+2,SEEK_CUR); cp -= (2*nv+4);
			  }
			  TextWin.Output(0,"- %d faces with %d vertices",fc,vc);
			  if(log) fprintf(lgfile,"- %d faces with %d vertices\n",fc,vc);
              fseek(infile,fpos,SEEK_SET);
              p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
              p->name=strdup("LW-Object");
			  p->nv=pn; p->pts=vert;
              p->nf = fc; p->pnv=new(byte[fc]); p->pfv=new(ushort[vc]); p->mat=new(byte[fc]); p->nm=mc;
              p->mtl=new(MatPtr[mc]); for(i=0; i<mc; i++) p->mtl[i]=mr[i];

			  for(fi=0,i=0; i<(p->nf); i++){
				p->pnv[i]=intread(infile); if((p->pnv[i])>mnv) mnv=(p->pnv[i]);
			    for(j=0; j<(p->pnv[i]); j++){
			      p->pfv[fi++]=intread(infile);
			    }
			    p->mat[i]=intread(infile)-1;
			  }
			  if(mnv==3){ delete p->pnv; p->pnv=NULL; }
              if(!VMF.MatList) { VMF.MatList=new Material; strcpy(VMF.MatList->name,"default"); }
              p->Mat=VMF.MatList;

              // default-transformation:
              p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);

              TextWin.Output(0," - done.\n");
			} else {
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes\n",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes\n",chunk,cl);
              fseek(infile,cl,SEEK_CUR);
			}
          }
        }
      } else if(!strcmp(chunk,"LWO2")){
        char  cf[80][240];
		int  *fix;
        while(!feof(infile)){
          fgets(chunk,5,infile); cl=longread(infile); if(cl & 1) cl++;
          if(!feof(infile)){
            if(!strcmp(chunk,"PNTS")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl);
			  TextWin.Output(0,"- %ld vertices\n",pn=(cl/12));
			  if(log) fprintf(lgfile," %ld vertices\n",pn);
              vert=new(vector[pn]);
              for(i=0; i<pn; i++){
				(vert[i]).x= floatread(infile);
				(vert[i]).y= floatread(infile);
				(vert[i]).z=-floatread(infile);
                if(vert[i].x > maxx) maxx=vert[i].x;
                if(vert[i].y > maxy) maxy=vert[i].y;
                if(vert[i].z > maxz) maxz=vert[i].z;
                if(vert[i].x < minx) minx=vert[i].x;
                if(vert[i].y < miny) miny=vert[i].y;
                if(vert[i].z < minz) minz=vert[i].z;
			  }
			} else if(!strcmp(chunk,"TAGS")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes\n",chunk,cl);
			  int cp=0;
			  for(i=0; i<cl; i++){
                fread(&(snm[cp]),1,1,infile);
				if(snm[cp]==0){
				  if(cp>0){
					MatPtr mat = new Material; mat->next=VMF.MatList; VMF.MatList=mat;
			        mat->name=strdup(snm); mr[mc++]=mat;
				    if(log) fprintf(lgfile,"  '%s'\n",snm);
				  } cp=0;
				} else cp++;
			  }
			} else if(!strcmp(chunk,"VMAP")){
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes.\n",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl);
              char subch[10]; fgets(subch,5,infile);
			  if(log) fprintf(lgfile," Type '%s':",subch); cp=4;
			  int   n1=intread(infile); cp+=2;
			  char  mpnm[24]; int ni=0;
			  do{
                fread(&(mpnm[ni++]),1,1,infile);
			  } while(mpnm[ni-1]);
              while(ni & 1) fread(&(mpnm[ni++]),1,1,infile); cp+=ni;
			  if(log) fprintf(lgfile," %d, '%s':",n1,mpnm);
			  tn=(cl-cp)/10; int vi;
			  if(log) fprintf(lgfile," %d texture coords\n",tn);
              texc=new(vector[tn]); tci=new(int[pn]); for(i=0; i<pn; i++) tci[i]=0;
              for(i=0; i<tn; i++){
				vi          =   intread(infile); tci[vi]=i;
				(texc[i]).x = floatread(infile);
				(texc[i]).y = floatread(infile);
				(texc[i]).z = 0.0f;
			  }
			} else if(!strcmp(chunk,"POLS")){
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes:",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl);
              char subch[10]; fgets(subch,5,infile);
			  if(log) fprintf(lgfile," Type '%s':",subch); cp=4;
              fpos=ftell(infile); fc=0; vc=0;
			  while(cp<cl){
			    nv=intread(infile); vc+=nv; fc++; cp+=2;
			    //if(log) fprintf(lgfile,"\n    %d :",nv);
				for(i=0; i<nv; i++){
			      intread(infile); cp+=2;
			      //if(log) fprintf(lgfile," %2d",sl);
				}
			  }
			  TextWin.Output(0," %d faces with %d vertices\n",fc,vc);
			  if(log) fprintf(lgfile," %d faces with %d vertices\n",fc,vc);
              p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
              p->name=strdup("LW-Object");
			  p->nv=pn; p->pts=vert; if(tn){ p->nt=tn; p->txc=texc; p->pft=new(ushort[vc]); }
              p->nf = fc; p->pnv=new(byte[fc]); p->pfv=new(ushort[vc]); fix=new(int[fc]);
              fseek(infile,fpos,SEEK_SET); cp=4; fc=0; fi=0;
			  while(cp<cl){
			    fix[fc]=fi; p->pnv[fc++]=nv=intread(infile); cp+=2;
				for(i=0; i<nv; i++){
			             p->pfv[fi]=intread(infile); cp+=2;
				  if(tn) p->pft[fi]=tci[p->pfv[fi]]; fi++;
				}
			  }
              if(!VMF.MatList) { VMF.MatList=new Material; VMF.MatList->name=strdup("default"); }
              p->Mat=VMF.MatList;
              // default-transformation:
              p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);
			} else if(!strcmp(chunk,"PTAG")){
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes.\n",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl);
              char subch[10]; fgets(subch,5,infile); int fn,mn;
			  if(log) fprintf(lgfile," Type '%s'\n",subch); cp=4;
              p->mtl=new(MatPtr[mc]); for(i=0; i<mc; i++) p->mtl[i]=mr[i];
              p->mat=new(byte[p->nf]); p->nm=mc;
			  while(cp<cl){
			    fn=intread(infile);
				mn=intread(infile);
				p->mat[fn]=mn; cp+=4;
			  }
			} else if(!strcmp(chunk,"VMAD")){
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes.\n",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl);
              char subch[10]; fgets(subch,5,infile);
			  if(log) fprintf(lgfile," Type '%s':",subch); cp=4;
			  int   vn=intread(infile); cp+=2;
			  char  mpnm[24]; int ni=0;
			  do{
                fread(&(mpnm[ni++]),1,1,infile);
			  } while(mpnm[ni-1]);
              while(ni & 1) fread(&(mpnm[ni++]),1,1,infile); cp+=ni;
			  if(log) fprintf(lgfile," %d, '%s'\n",vn,mpnm);
			  int fn; int vtn=(cl-cp)/12;
              p->txc=new(vector[p->nt=tn+vtn]); for(i=0; i<tn; i++) p->txc[i]=texc[i];
              for(i=0; i<vtn; i++){
			    vn=intread(infile);
			    fn=intread(infile);
			    p->txc[tn+i].x=floatread(infile);
			    p->txc[tn+i].y=floatread(infile);
				p->txc[tn+i].z=0.0f;
				for(j=0; j<p->pnv[fn]; j++){
				  if(p->pfv[fix[fn]+j]==vn){
					p->pft[fix[fn]+j]=tn+i;
				  }
				}
			    if(log) fprintf(lgfile,"  %2d,%2d :  %f,%f\n",vn,fn,p->txc[tn+i].x,p->txc[tn+i].y);
			  }
			} else if(!strcmp(chunk,"CLIP")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes:",chunk,cl);
			  int  cn=longread(infile);
			  char tp[8]; fread(tp,4,1,infile); tp[4]=0;
			  int  sn=intread(infile);
			  for(i=0; i<sn; i++){
                fread(&(cf[cn][i]),1,1,infile);
			  } cf[cn][i]=0;
			  if(log) fprintf(lgfile," %2d '%s': '%s'\n",cn,tp,cf[cn]);
			} else if(!strcmp(chunk,"SURF")){
			  if(log) fprintf(lgfile,"- Chunk '%s', length %ld bytes:",chunk,cl);
			  char sn[80]; char dummy; cp=0;
			  do {
                fread(&(sn[cp++]),1,1,infile);
			  } while(sn[cp-1]);
			  if(cp & 1){ fread(&dummy,1,1,infile); cp++; }
			  fread(&dummy,1,1,infile); cp++;
			  fread(&dummy,1,1,infile); cp++;
			  if(log) fprintf(lgfile," '%s'\n",sn);
			  MatPtr mp=VMF.MatList; if(mp){ while(((mp->next)!=NULL)&&(strcmp(mp->name,sn)!=0)) mp=mp->next; }
              char subch[10]; int sl;
			  while(cp<cl){
                fgets(subch,5,infile);
				sl=intread(infile); if(sl & 1) sl++;
			    if(!strcmp(subch,"COLR")){
                  mp->Reflection.R=floatread(infile);
                  mp->Reflection.G=floatread(infile);
                  mp->Reflection.B=floatread(infile);
                  fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Color: (%f, %f, %f)\n",mp->Reflection.R, mp->Reflection.G, mp->Reflection.B);
				} else if(!strcmp(subch,"DIFF")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Diffuse: %f\n",a);
				} else if(!strcmp(subch,"SPEC")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Specular: %f\n",a);
				} else if(!strcmp(subch,"GLOS")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - GLOS: %f\n",a);
				} else if(!strcmp(subch,"BUMP")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - BUMP: %f\n",a);
				} else if(!strcmp(subch,"TRAN")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - TRAN: %f\n",a);
				} else if(!strcmp(subch,"RIND")){
                  float a = floatread(infile); fread(&dummy,1,1,infile); fread(&dummy,1,1,infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - RIND: %f\n",a);
				} else if(!strcmp(subch,"SMAN")){
                  float a = floatread(infile); cp += sl+6;
			      if(log) fprintf(lgfile,"  - Smoothing angle: %f (%6.2f)\n",a,180.0f*a/3.14159265f);
				} else if(!strcmp(subch,"BLOK")){
			      if(log) fprintf(lgfile,"  - Chunk '%s', length %ld bytes:\n",subch,sl);
                  char subch2[10]; int sl2; int cp2=0;
			      while(cp2<sl){
                    fgets(subch2,5,infile);
				    sl2=intread(infile); if(sl2 & 1) sl2++;
			        if(!strcmp(subch2,"TMAP")){
			          if(log) fprintf(lgfile,"    - Chunk '%s', length %ld bytes:\n",subch2,sl2);
                      char subch3[10]; int sl3; int cp3=0;
			          while(cp3<sl2){
                        fgets(subch3,5,infile);
				        sl3=intread(infile); if(sl3 & 1) sl3++;
			            //if(!strcmp(subch2,"TMAP")){
						//} else {
			              if(log) fprintf(lgfile,"      - Chunk '%s', length %ld bytes: ",subch3,sl3); byte byt[24];
                          cp3 += sl3+6;
			              for(i=0; i<min(20,sl3); i++){
				            fread(&(byt[i]),1,1,infile);
				            if(log) fprintf(lgfile," %d", byt[i]);
						  } byt[i]=0;
			              if(log) fprintf(lgfile," '%s'\n",byt);
                          if(sl3-i > 0) fseek(infile,sl3-i,SEEK_CUR);
					    //}
					  } cp2 += sl2+6;
					} else if(!strcmp(subch2,"IMAG")){
			          if(log) fprintf(lgfile,"    - Chunk '%s', length %ld bytes:",subch2,sl2);
					  int in=intread(infile);
			          if(log) fprintf(lgfile," %d = '%s'\n",in,cf[in]);
					  mp->Texture = new TexMap;
				      mp->Texture->tname =strdup(cf[in]); mp->Texture->tbits=NULL; mp->Texture->glbits=NULL; mp->Texture->tflags=0;
					  cp2 += sl2+6;
                    } else {
			          if(log) fprintf(lgfile,"    - Chunk '%s', length %ld bytes: ",subch2,sl2); byte byt[24];
                      cp2 += sl2+6;
			          for(i=0; i<min(20,sl2); i++){
				        fread(&(byt[i]),1,1,infile);
				        if(log) fprintf(lgfile," %d", byt[i]);
					  } byt[i]=0;
			          if(log) fprintf(lgfile," '%s'\n",byt);
                      if(sl2-i > 0) fseek(infile,sl2-i,SEEK_CUR);
                    }
				  } cp += sl+6;
				} else {
			      if(log) fprintf(lgfile,"  - Chunk '%s', length %ld bytes: ",subch,sl); byte byt[24];
                  cp += sl+6;
			      for(i=0; i<min(20,sl); i++){
				    fread(&(byt[i]),1,1,infile);
				    if(log) fprintf(lgfile," %d", byt[i]);
				  } byt[i]=0;
			      if(log) fprintf(lgfile," '%s'\n",byt);
                  if(sl-i > 0) fseek(infile,sl-i,SEEK_CUR);
				}
			  }
			} else {
			  TextWin.Output(0,"- Chunk '%s', length %ld bytes.\n",chunk,cl);
			  if(log) fprintf(lgfile,"- Chunk '%s', length %3ld bytes:",chunk,cl); byte byt[24];
			  for(i=0; i<min(20,cl); i++){
				fread(&(byt[i]),1,1,infile);
				if(log) fprintf(lgfile," %3d", byt[i]);
			  } byt[i]=0;
			  if(log) fprintf(lgfile," '%s'\n",byt);
              if(cl-i > 0) fseek(infile,cl-i,SEEK_CUR);
			}
          }
        }
      } else TextWin.Output(3,"- ERROR: no Lightwave-Object-format!\n");
      // set up a view
      vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
      float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
      V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
      V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
      V->SetViewAngle(0.5F); WorldValid=FALSE;
      vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
      strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
    } else TextWin.Output(3,"- ERROR: no IFF-Form!\n");
    fclose(infile);
  } else TextWin.Output(3,"- ERROR: couldn't open input-file!\n");
  if(log) fclose(lgfile);
}


int   GEOF;
int   indent;

void GetGToken(char *a, CFile *file)
{
  char c; int i,n=0; indent=0;
  do{
    while( n==1 && c!=13 && c!=10){
	  n=file->Read(&c,1);
	} i=0;
    do { n=file->Read(&c,1); indent++; } while( n==1 && (c==' ' || c==',' || c==':' || c=='=' || c==';' || c=='\t' || c=='(' || c==')' || c==13 || c==10 ));
    if(n==1){
      if(c=='\"'){
	    n=file->Read(&c,1);
	    while(n==1 && (c!='\"' && c!=13 && c!=10)){
	      a[i++]=c; n=file->Read(&c,1);
	    }
      } else {
	    do{ a[i++]=c; n=file->Read(&c,1); }while( n==1 && (c!=' ' && c!=',' && c!='=' && c!=';' && c!=':' && c!='\t' && c!='(' && c!=')' && c!=13 && c!=10));
      }
    }
    a[i]=0;
  } while(a[0]=='!');
  GEOF = n==1? 0 : 1;
}

void CVModelDoc::ImportGEO(char *filename, ViewPtr V)
{
  PolyMesh  *p;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  int    fc,vc,lc,i,j;
  char   a[80];
  CFile  file;
  
  int pr[80],pn[80]; int pp=0;

  if(file.Open(filename, CFile::modeRead | CFile::shareDenyWrite)){
    GetGToken(a,&file);
    if(!stricmp(a,"3DG1")){
      p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
      p->name=strdup("GEO-Object");
      GetGToken(a,&file); p->nv=atoi(a); p->pts=new(vector[p->nv]); fc=0; vc=0; lc=0;
      TextWin.Output(1,"- %d vertices\n",p->nv);
	  for(i=0; i<p->nv; i++){
        GetGToken(a,&file); p->pts[i].x=(float)atof(a);
        GetGToken(a,&file); p->pts[i].y=(float)atof(a);
        GetGToken(a,&file); p->pts[i].z=(float)atof(a);
        if(p->pts[i].x > maxx) maxx=p->pts[i].x;
        if(p->pts[i].y > maxy) maxy=p->pts[i].y;
        if(p->pts[i].z > maxz) maxz=p->pts[i].z;
        if(p->pts[i].x < minx) minx=p->pts[i].x;
        if(p->pts[i].y < miny) miny=p->pts[i].y;
        if(p->pts[i].z < minz) minz=p->pts[i].z;
	  }
	  while(!GEOF){
        GetGToken(a,&file);
	    if(indent==1){
		  int nv=atoi(a);
		  for(i=0; i<nv; i++){
            GetGToken(a,&file);		  
		  }
          GetGToken(a,&file); int p=atoi(a);
          //TextWin.Output(1,"- face with %d vertices, property %d\n",nv,atoi(a));
		  for(i=0; i<pp && pr[i]!=p; i++); if(i>=pp){ pn[pp]=1; pr[pp++]=p; } else pn[i]++;
		  if(nv>2){ fc++; vc+=nv; } else lc++;
		} else if(indent==3){
		  int ns=atoi(a);
		  for(j=0; j<ns; j++){
            GetGToken(a,&file);
	        if(indent==5){
		      int nv=atoi(a);
		      for(i=0; i<nv; i++){
                GetGToken(a,&file);		  
			  }
              GetGToken(a,&file);
              //TextWin.Output(1,"  - subface with %d vertices, property %d\n",nv,atoi(a));
			}
		  }
		}
	  }
      TextWin.Output(1,"- %d faces with %d vertices (%d lines)\n",fc,vc,lc);
	  TextWin.Output(1,"- %d properties: ",pp);
	  p->nm=pp; p->mtl=new(MatPtr[p->nm]);
	  for(i=0; i<pp; i++){
		TextWin.Output(1,", %d(%dx)",pr[i],pn[i]);
	    MatPtr mat = new Material; mat->next=VMF.MatList; VMF.MatList=mat;
		char mtn[16]; sprintf(mtn,"%d",pr[i]);
		mat->name=strdup(mtn); p->mtl[i]=mat;
		if(abs(pr[i])== 0){ mat->Reflection.R=0.1f; mat->Reflection.G=0.1f; mat->Reflection.B=0.1f; } // black
		if(abs(pr[i])== 1){ mat->Reflection.R=0.1f; mat->Reflection.G=0.9f; mat->Reflection.B=0.1f; } // green
		if(abs(pr[i])== 2){ mat->Reflection.R=0.1f; mat->Reflection.G=0.7f; mat->Reflection.B=0.1f; } // green
		if(abs(pr[i])== 6){ mat->Reflection.R=0.7f; mat->Reflection.G=0.5f; mat->Reflection.B=0.1f; } // brown
		if(abs(pr[i])== 7){ mat->Reflection.R=0.5f; mat->Reflection.G=0.5f; mat->Reflection.B=0.5f; } // gray
		if(abs(pr[i])==10){ mat->Reflection.R=0.1f; mat->Reflection.G=0.5f; mat->Reflection.B=0.1f; } // green
		if(abs(pr[i])==12){ mat->Reflection.R=0.9f; mat->Reflection.G=0.1f; mat->Reflection.B=0.1f; } // red
		if(abs(pr[i])==15){ mat->Reflection.R=0.9f; mat->Reflection.G=0.9f; mat->Reflection.B=0.9f; } // white
		if(abs(pr[i])==17){ mat->Reflection.R=0.1f; mat->Reflection.G=0.5f; mat->Reflection.B=0.9f; } // blau
		if(abs(pr[i])==28){ mat->Reflection.R=0.9f; mat->Reflection.G=0.1f; mat->Reflection.B=0.1f; } // red
		if(abs(pr[i])==30){ mat->Reflection.R=0.9f; mat->Reflection.G=0.9f; mat->Reflection.B=0.1f; } // yellow
		if(abs(pr[i])==31){ mat->Reflection.R=0.9f; mat->Reflection.G=0.9f; mat->Reflection.B=0.9f; } // white
		if(abs(pr[i])==32){ mat->Reflection.R=0.3f; mat->Reflection.G=0.3f; mat->Reflection.B=0.3f; } // dark gray
	  } TextWin.Output(1,"\n");
      p->nf = fc; p->pnv=new(byte[fc]); p->mat=new(byte[fc]); p->pfv=new(ushort[vc]); fc=0; vc=0;
      // read faces-chunk again to fill array
  	  file.Close(); file.Open(filename, CFile::modeRead | CFile::shareDenyWrite);
      GetGToken(a,&file); GetGToken(a,&file);
	  for(i=0; i<p->nv; i++){
        GetGToken(a,&file);
        GetGToken(a,&file);
        GetGToken(a,&file);
	  }
	  while(!GEOF){
        GetGToken(a,&file);
	    if(indent==1){
		  int nv=atoi(a); if(nv>2) p->pnv[fc]=nv;
		  for(i=0; i<nv; i++){
            GetGToken(a,&file); if(nv>2) p->pfv[vc+nv-1-i]=atoi(a);
		  }
          GetGToken(a,&file); int pp=atoi(a);
          if(nv>2){
		    for(i=0; pr[i]!=pp; i++); p->mat[fc]=i;
            fc++; vc+=nv;
		  }
		} else if(indent==3){
		  int ns=atoi(a);
		  for(j=0; j<ns; j++){
            GetGToken(a,&file);
	        if(indent==5){
		      int nv=atoi(a);
		      for(i=0; i<nv; i++){
                GetGToken(a,&file);		  
			  }
              GetGToken(a,&file);
              //TextWin.Output(1,"  - subface with %d vertices, property %d\n",nv,atoi(a));
			}
		  }
		}
	  }
      TextWin.Output(1,"- %d faces with %d vertices\n",fc,vc);
      if(!VMF.MatList) { VMF.MatList=new Material; VMF.MatList->name=strdup("default"); }
      p->Mat=VMF.MatList;

      // default-transformation:
      p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);

      // set up a view
      vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
      float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
      V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
      V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
      V->SetViewAngle(0.5F); WorldValid=FALSE;
      vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
      strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
	} else {
	  TextWin.Output(3,"- ERROR: no VideoScape-GEO-format!\n");
	}
	file.Close();
  } else {
	TextWin.Output(3,"- file '%s' could not be opened!\n",filename);
  }
}


int   VFEOF,ln;

void GetVToken(char *a, CFile *file)
{
  char c; int i,n=0;
  do{
	// read comment till end of line, skipped at first entry
    while( n==1 && c!=13 && c!=10){
	  n=file->Read(&c,1);
	} i=0;
    // trim leading blanks
	do n=file->Read(&c,1); while( n==1 && (c==' ' || c==',' || c==':' || c=='\t' || c==13 || c==10 ));
    // read token
	if(n==1){
      if(c=='\"'){
	    n=file->Read(&c,1);
	    while(n==1 && (c!='\"' && c!=13 && c!=10)){
	      a[i++]=c; n=file->Read(&c,1);
	    }
      } else {
		if((c=='[') || (c==']') || (c=='{') || (c=='}')){
		  a[i++]=c;
		} else {
	      do{ a[i++]=c; n=file->Read(&c,1); }while( n==1 && (c!=' ' && c!=',' && c!=':' && c!='\t' && c!=13 && c!=10));
		  if((c==13) || (c==10)) ln++;
		}
      }
    }
    a[i]=0;
  } while((a[0]=='#') && strcmp(a,"#VRML"));
  VFEOF = n==1? 0 : 1;
}


void ReadSep(CFile *file)
{
  char a[80]; int ins=1;
  while(ins && !VFEOF){
    GetVToken(a,file);
    if(!strcmp(a,"Coordinate")){
      GetVToken(a,file);
	  if(a[0]=='{'){
		TextWin.Output(1,"  - Token 'Coordinate'\n");
	  }
	}
    if(!strcmp(a,"TextureCoordinate")){
      GetVToken(a,file);
	  if(a[0]=='{'){
		TextWin.Output(1,"  - Token 'TextureCoordinate'\n");
	  }
	}
    if(!strcmp(a,"coordIndex")){
      GetVToken(a,file);
	  if(a[0]=='['){
		TextWin.Output(1,"  - Token 'coordIndex'\n");
	  }
	}
	if(a[0]=='{'){
	  ReadSep(file);
	} else {
	  if(a[0]=='}') ins=0;
	}
  }
}

void CVModelDoc::ImportVRML(char *filename, ViewPtr V)
{
  char      a[80],s[80];
  int       sm,fi,i,j,k,pn,fn,vn;
  CFile     file;
  DWORD     fpos;
  int       Sep  = 0;
  int       VO   = 0;
  int       MI   = 1;
  int       nn   = 0;
  PolyMesh *p    = NULL;
  vector   *vert = NULL;
  vector   *norm = NULL;
  MatPtr    mat  = NULL;
  MatPtr    MO   = NULL;
  MatPtr    mp;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  if(file.Open(filename, CFile::modeRead | CFile::shareDenyWrite)){
    GetVToken(a,&file);
    if(!strcmp(a,"#VRML")){
      GetVToken(a,&file);
      if(!strcmp(a,"V1.0")){
        GetVToken(a,&file);
        if(!strcmp(a,"ascii")){
          GetVToken(a,&file);
		  while(!VFEOF){
            if(!strcmp(a,"Separator") || !strcmp(a,"Group")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file); Sep++;
		    } else if(!strcmp(a,"ShapeHints")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				if(!strcmp(a,"vertexOrdering")){
				  GetVToken(a,&file);
				  if(!strcmp(a,"COUNTERCLOCKWISE")){
					VO=-1;
				  } else if(!strcmp(a,"CLOCKWISE")){
					VO=+1;
				  } else { TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
				}
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"Info")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"LightModel")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"Transform")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"MatrixTransform")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"MaterialBinding")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"NormalBinding")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				//TextWin.Output(1,"- Token '%s'\n",a);
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"Material")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
              mat=new Material; if(VMF.MatList){ for(mp=VMF.MatList; mp->next; mp=mp->next); mp->next=mat; } else VMF.MatList=mat;
              sprintf(mat->name,"Material%d",MI); MI++;
			  
			  GetVToken(a,&file);
			  while(strcmp(a,"}") && !VFEOF){
				if(!strcmp(a,"diffuseColor")){
			      GetVToken(a,&file);
				  if(!strcmp(a,"[")){
					MO=mat;
				    GetVToken(a,&file); mat->Reflection.R = (float)atof(a);
				    GetVToken(a,&file); mat->Reflection.G = (float)atof(a);
				    GetVToken(a,&file); mat->Reflection.B = (float)atof(a);
				    GetVToken(a,&file);
			        while(strcmp(a,"]")){
                      mat=new Material; if(VMF.MatList){ for(mp=VMF.MatList; mp->next; mp=mp->next); mp->next=mat; } else VMF.MatList=mat;
                      sprintf(mat->name,"Material%d",MI); MI++;

				                          mat->Reflection.R = (float)atof(a);
				      GetVToken(a,&file); mat->Reflection.G = (float)atof(a);
				      GetVToken(a,&file); mat->Reflection.B = (float)atof(a);
				      GetVToken(a,&file);
			        }
				  } else {
				                        mat->Reflection.R = (float)atof(a);
				    GetVToken(a,&file); mat->Reflection.G = (float)atof(a);
				    GetVToken(a,&file); mat->Reflection.B = (float)atof(a);
				  }
				}
				if(!strcmp(a,"specularColor")){
				  //mat->Ks=1.0f;
				  GetVToken(a,&file); //mat->Os.R = (float)atof(a);
				  GetVToken(a,&file); //mat->Os.G = (float)atof(a);
				  GetVToken(a,&file); //mat->Os.B = (float)atof(a);
				}
				if(!strcmp(a,"emissiveColor")){
				  GetVToken(a,&file); //mat->Ol.R = (float)atof(a);
				  GetVToken(a,&file); //mat->Ol.G = (float)atof(a);
				  GetVToken(a,&file); //mat->Ol.B = (float)atof(a);
				  //if((mat->Ol.R > 0.0f) || (mat->Ol.G > 0.0f) || (mat->Ol.B > 0.0f)) mat->Kl=1.0f;
				}
				if(!strcmp(a,"shininess")){
				  GetVToken(a,&file); // mat->n ???
				}
				GetVToken(a,&file);
			  }
			  GetVToken(a,&file);
		    } else if(!strcmp(a,"Coordinate3")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
		      if(!strcmp(a,"point")){
			    GetVToken(a,&file); if(strcmp(a,"[")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
                // count number of vertices for array size
                fpos=file.GetPosition();
			    GetVToken(a,&file); pn=0;
			    while(strcmp(a,"]") && !VFEOF){
				  pn++;
				  GetVToken(a,&file);
			    }
		        if(pn%3){ TextWin.Output(3,"- pointlist corrupt!\n"); file.Close(); return; }
		        pn/=3; TextWin.Output(1,"- %d vertices",pn);
                // read vertices again to fill array
                file.Seek(fpos, CFile::begin);
                if(vert) delete vert; vert=new(vector[pn]);
				for(i=0; i<pn; i++){
			      GetVToken(a,&file); (vert[i]).x=-(float)atof(a);
			      GetVToken(a,&file); (vert[i]).y= (float)atof(a);
			      GetVToken(a,&file); (vert[i]).z= (float)atof(a);
                  if(vert[i].x > maxx) maxx=vert[i].x;
                  if(vert[i].y > maxy) maxy=vert[i].y;
                  if(vert[i].z > maxz) maxz=vert[i].z;
                  if(vert[i].x < minx) minx=vert[i].x;
                  if(vert[i].y < miny) miny=vert[i].y;
                  if(vert[i].z < minz) minz=vert[i].z;
				}
			    GetVToken(a,&file); if(strcmp(a,"]")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			    GetVToken(a,&file); if(strcmp(a,"}")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			    GetVToken(a,&file);
		      } else { TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
		    } else if(!strcmp(a,"Normal")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
		      if(!strcmp(a,"vector")){
			    GetVToken(a,&file); if(strcmp(a,"[")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
                // count number of vertices for array size
                fpos=file.GetPosition();
			    GetVToken(a,&file); nn=0;
			    while(strcmp(a,"]") && !VFEOF){
				  nn++;
				  GetVToken(a,&file);
			    }
		        if(nn%3){ TextWin.Output(3,"- normal-list corrupt!\n"); file.Close(); return; }
		        nn/=3; TextWin.Output(1,", %d normals",nn);
                // read vertices again to fill array
                file.Seek(fpos, CFile::begin);
                if(norm) delete norm; norm=new(vector[nn]);
				for(i=0; i<nn; i++){
			      GetVToken(a,&file); (norm[i]).x=(float)atof(a);
			      GetVToken(a,&file); (norm[i]).y=(float)atof(a);
			      GetVToken(a,&file); (norm[i]).z=(float)atof(a);
				}
			    GetVToken(a,&file); if(strcmp(a,"]")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			    GetVToken(a,&file); if(strcmp(a,"}")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			    GetVToken(a,&file);
		      } else { TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
		    } else if(!strcmp(a,"IndexedFaceSet")){
			  GetVToken(a,&file); if(strcmp(a,"{")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
			  GetVToken(a,&file);
		      if(!strcmp(a,"coordIndex")){
			    GetVToken(a,&file); if(strcmp(a,"[")){ TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
                // count number of faces and vertices for array size
                fpos=file.GetPosition();
			    GetVToken(a,&file); fn=0; vn=0;
			    while(strcmp(a,"]") && !VFEOF){
			      while(strcmp(a,"-1") && !VFEOF){
				    vn++;
				    GetVToken(a,&file);
				  }
				  fn++;
				  GetVToken(a,&file);
			    }
				TextWin.Output(1,", %d %ses",fn,(3*fn == vn)? "triangl" : "fac");
                p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
                p->name=strdup("Object");
				p->nv=pn; p->pts=new(vector[p->nv]); memcpy(p->pts, vert, pn*sizeof(vector));
                // read vertices again to fill array
                file.Seek(fpos, CFile::begin);
				if(3*fn == vn){
                  p->nf = fn; p->pfv=new(ushort[3*(p->nf)]);
				  for(i=0; i<(p->nf); i++){
					if(VO>=0){
                      GetVToken(a,&file); (p->pfv[3*i+2])=atoi(a);
                      GetVToken(a,&file); (p->pfv[3*i+1])=atoi(a);
                      GetVToken(a,&file); (p->pfv[3*i  ])=atoi(a);
					} else {
                      GetVToken(a,&file); (p->pfv[3*i  ])=atoi(a);
                      GetVToken(a,&file); (p->pfv[3*i+1])=atoi(a);
                      GetVToken(a,&file); (p->pfv[3*i+2])=atoi(a);
					}
                    GetVToken(a,&file);
                  }
                } else { 
                  p->nf = fn; p->pnv=new(byte[fn]); p->pfv=new(ushort[vn]);
				  for(fi=0,i=0; i<(p->nf); i++){
                    GetVToken(a,&file); j=0;
                    while(strcmp(a,"-1")){
                      p->pfv[fi++]=atoi(a); j++;
					  GetVToken(a,&file);
					}
					p->pnv[i]=j;
				    if(VO>=0) for(k=0; k<(j>>1); k++) swap(p->pfv[fi-j+k], p->pfv[fi-1-k]);
				  }
				}
			    GetVToken(a,&file); if(strcmp(a,"]")){ TextWin.Output(3,"- syntax error: '%s'\n",a); Kill1(p); file.Close(); return; }
			    GetVToken(a,&file);
		        if(!strcmp(a,"materialIndex")){
			      GetVToken(a,&file);
				  if(!strcmp(a,"[")){
                    for(p->nm=0,mp=VMF.MatList; mp; mp=mp->next) p->nm++;
					p->mtl=(MatPtr *)new(MatPtr[p->nm]);
					for(i=0,mp=VMF.MatList; i<(p->nm); i++){ p->mtl[i]=mp; mp=mp->next; }
					p->mat=(byte *)new(byte[p->nf]);
					for(sm=1,i=0; i<(p->nf); i++){
                      GetVToken(a,&file); (p->mat[i])=atoi(a);
					  if(p->mat[i]!=p->mat[0]) sm=0;
			        }
                    GetVToken(a,&file); if(strcmp(a,"]")){ TextWin.Output(3,"- materialIndex corrupt!\n"); Kill1(p); file.Close(); return; }
                    if(sm){
					  p->Mat=p->mtl[p->mat[0]]; p->nm=0;
					  delete p->mat; p->mat=NULL;
					  delete p->mtl; p->mtl=NULL;
					}
				  }
			      GetVToken(a,&file);
				}
		        if(!strcmp(a,"normalIndex")){
			      GetVToken(a,&file); if(strcmp(a,"[")){ TextWin.Output(3,"- syntax error: '%s'\n",a); Kill1(p); file.Close(); return; }
				  if(3*fn == vn){
                    p->pfn=new(ushort[3*(p->nf)]);
				    for(i=0; i<(p->nf); i++){
					  if(VO>=0){
                        GetVToken(a,&file); (p->pfn[3*i+2])=atoi(a);
                        GetVToken(a,&file); (p->pfn[3*i+1])=atoi(a);
                        GetVToken(a,&file); (p->pfn[3*i  ])=atoi(a);
					  } else {
                        GetVToken(a,&file); (p->pfn[3*i  ])=atoi(a);
                        GetVToken(a,&file); (p->pfn[3*i+1])=atoi(a);
                        GetVToken(a,&file); (p->pfn[3*i+2])=atoi(a);
					  }
                      GetVToken(a,&file); if(strcmp(a,"-1")){ TextWin.Output(3,"- normalIndex corrupt!\n"); Kill1(p); file.Close(); return; }
                    }
                  } else { 
                    p->pfn=new(ushort[vn]);
				    for(fi=0,i=0; i<(p->nf); i++){
					  for(j=0; j<(p->pnv[i]); j++){
                        GetVToken(a,&file); p->pfn[fi++]=atoi(a);
					  }
					  GetVToken(a,&file); if(strcmp(a,"-1")){ TextWin.Output(3,"- normalIndex corrupt!\n"); Kill1(p); file.Close(); return; }
				      if(VO>=0) for(k=0; k<(j>>1); k++) swap(p->pfn[fi-j+k], p->pfn[fi-1-k]);
				    }
				  }
			      GetVToken(a,&file); if(strcmp(a,"]")){ TextWin.Output(3,"- syntax error: '%s'\n",a); Kill1(p); file.Close(); return; }
                  if(nn){
				    p->nn=nn; p->nrm=new(vector[p->nn]); memcpy(p->nrm, norm, nn*sizeof(vector));
				  } else {
				    delete p->pfn; p->pfn=NULL;
				  }
				  GetVToken(a,&file);
				}
				if(strcmp(a,"}")){ TextWin.Output(3,"- syntax error: '%s'\n",a); Kill1(p); file.Close(); return; }
                if(!mat){
                  VMF.MatList=new Material; strcpy(VMF.MatList->name,"default");
				  mat=VMF.MatList;
                }
                if(!(p->Mat)) p->Mat=mat;
				if(!VO) p->pmflags |= SURF_DOUBLESIDED;

                // default-transformation:
                p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);

                TextWin.Output(0," - building mesh");
                /*p->BuildMesh();*/ DelIsoVert(p, FALSE);
                TextWin.Output(0," - done.\n");
			    GetVToken(a,&file);
		      } else { TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
		    } else if(!strcmp(a,"}")){
			  Sep--; GetVToken(a,&file);
		    } else { TextWin.Output(3,"- syntax error: '%s'\n",a); file.Close(); return; }
		  }
          if(vert) delete vert;
          if(norm) delete norm;
		  // set up a view
		  vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
          float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
          V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
          V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
		  V->SetViewAngle(0.5F); WorldValid=FALSE;
          vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
          strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
        } else TextWin.Output(3,"- mode '%s' not supported!\n",a);
      } else if(!strcmp(a,"V2.0")){
        ln=0; GetVToken(a,&file);
        if(!strcmp(a,"utf8")){
          TextWin.Output(1,"- VRML V2.0 (utf8) ",a);
		  while(ln<1){
            GetVToken(a,&file);
            TextWin.Output(1,"%s ",a);
		  } TextWin.Output(1,"\n" ,a);
          GetVToken(a,&file);
		  while(!VFEOF){
            if(!strcmp(a,"DEF")){
              GetVToken(a,&file);
              GetVToken(a,&file);
			}
			TextWin.Output(1,"- Token '%s'\n",a);
		    GetVToken(s,&file);
			if(s[0]=='{'){
			  ReadSep(&file);
              GetVToken(a,&file);
			} else VFEOF=1;
		  }
        } else TextWin.Output(3,"- mode '%s' not supported!\n",a);
	  } else TextWin.Output(3,"- version %s not supported!\n",a);
    } else TextWin.Output(3,"- no VRML format!\n");
	file.Close();
  } else TextWin.Output(3,"- can't open file '%s'!\n",filename);
}

const float OB64 = 1.0f/64.0f;

int  groupcode;
char gvalue[256];

void GetGroup(FILE *file)
{
  char text[256]; char LF; int i=0;
  
  do fread(&(text[i++]),1,1,file); while((text[i-1]!=13) && (text[i-1]!=10));
  if(text[i-1]==13) fread(&LF,1,1,file); // read LF
  text[i-1]=0; groupcode=atoi(text);
  i=0;
  do fread(&(gvalue[i++]),1,1,file); while((gvalue[i-1]!=13) && (gvalue[i-1]!=10));
  if(gvalue[i-1]==13) fread(&LF,1,1,file); // read LF
  gvalue[i-1]=0;
}

LayPtr Layers;

LayPtr StLayer(char *lay)
{
  LayPtr L;
  for(L=Layers; L && stricmp(L->name, lay); L=L->next);
  if(!L){ L=new(LayList); L->next=Layers; Layers=L; strcpy(L->name, lay); }
  return L;
}

void CVModelDoc::ImportDXF(char *filename, ViewPtr V)
{
  FILE      *infile;
  PolyMesh  *p;
  int        pc;
  FacePtr    P,G,F,Faces=NULL; float AEL=0.0f; int Fc=0;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  Layers = NULL;

  infile = fopen(filename,"rb");
  if(infile){
    TextWin.Output(0,"reading file ...\n"); int secfound=0;
	while(!secfound){
	  GetGroup(infile);
	  if((groupcode == 0) && !strcmp(gvalue,"SECTION")){
	    GetGroup(infile);
        TextWin.Output(0,"- Section '%s' ...\n",gvalue);
	    if((groupcode == 2) && !strcmp(gvalue,"ENTITIES")){
	      secfound=1;
	    }
	  }
	}
    GetGroup(infile);
	while(strcmp(gvalue,"ENDSEC")){
	  pc=0;
	  if((groupcode == 0) && !strcmp(gvalue,"ARC")){
		vector pos,Az=vector(0.0f, 0.0f, 1.0f); float w1,w2,rad,len=0.0f;
	    do {
		  GetGroup(infile);
		  switch(groupcode){
            case  10: pos.x = (float)atof(gvalue); break;
            case  20: pos.y = (float)atof(gvalue); break;
            case  30: pos.z = (float)atof(gvalue); break;
            case  39: len   = (float)atof(gvalue); break;
            case  40: rad   = (float)atof(gvalue); break;
            case  50: w1    = (float)atof(gvalue)*Deg2Rad; break;
            case  51: w2    = (float)atof(gvalue)*Deg2Rad; break;
            case 210: Az.x  = (float)atof(gvalue); break;
            case 220: Az.y  = (float)atof(gvalue); break;
            case 230: Az.z  = (float)atof(gvalue); break;
		  }
	    } while(groupcode!=0);
		if(len){
          TextWin.Output(0,"- Arc r%f, l%f\n",rad,len);
          Cylinder *c = new Cylinder; c->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=c; VMDoc->VMF.Base=c;
          c->name = strdup("extruded_arc");
          if(!VMF.MatList){
	        VMF.MatList=new Material; VMF.MatList->name = strdup("default");
		  }
          c->Mat=VMF.MatList;
		  vector Ax = unit( (((fabs(Az.x) < OB64) && (fabs(Az.y) < OB64))? vector(0.0f, 1.0f, 0.0f) : vector(0.0f, 0.0f, 1.0f)) * Az);
          vector Ay = unit( Az * Ax); matrix TR;
		  TR.M11=Ax.x; TR.M12=Ax.y; TR.M13=Ax.z;
          TR.M21=Ay.x; TR.M22=Ay.y; TR.M23=Ay.z;
          TR.M31=Az.x; TR.M32=Az.y; TR.M33=Az.z;
          TR.M41=0.0f; TR.M42=0.0f; TR.M43=0.0f;
		  c->Scale(vector(rad,rad,(float)fabs(len)));
		  c->RotateZ(w1); c->angle=w2-w1;
          if(len<0.0f) c->Translate(vector(0.0f, 0.0f, len));
		  c->Translate(pos);
		  c->TM = c->TM*TR; c->Tb = c->TM.Inverse(); c->NM = c->TM.NTrans();
		}
	  } else if((groupcode == 0) && !strcmp(gvalue,"3DFACE")){
	    int face_complete=12;
        FacePtr F=new(FaceList); F->next=Faces; Faces=F; Fc++;
	    while(face_complete > 0){
		  GetGroup(infile);
		  switch(groupcode){
		    case  8: F->Lay  = StLayer(gvalue); break;
            case 10: F->p0.x = (float)atof(gvalue); face_complete--; break;
            case 20: F->p0.y = (float)atof(gvalue); face_complete--; break;
            case 30: F->p0.z = (float)atof(gvalue); face_complete--; break;
            case 11: F->p1.x = (float)atof(gvalue); face_complete--; break;
            case 21: F->p1.y = (float)atof(gvalue); face_complete--; break;
            case 31: F->p1.z = (float)atof(gvalue); face_complete--; break;
            case 12: F->p2.x = (float)atof(gvalue); face_complete--; break;
            case 22: F->p2.y = (float)atof(gvalue); face_complete--; break;
            case 32: F->p2.z = (float)atof(gvalue); face_complete--; break;
            case 13: F->p3.x = (float)atof(gvalue); face_complete--; break;
            case 23: F->p3.y = (float)atof(gvalue); face_complete--; break;
            case 33: F->p3.z = (float)atof(gvalue); face_complete--; break;
		  }
	    }
        if(F->p0.x > maxx) maxx=F->p0.x; if(F->p0.y > maxy) maxy=F->p0.y; if(F->p0.z > maxz) maxz=F->p0.z;
        if(F->p0.x < minx) minx=F->p0.x; if(F->p0.y < miny) miny=F->p0.y; if(F->p0.z < minz) minz=F->p0.z;
        if(F->p1.x > maxx) maxx=F->p1.x; if(F->p1.y > maxy) maxy=F->p1.y; if(F->p1.z > maxz) maxz=F->p1.z;
        if(F->p1.x < minx) minx=F->p1.x; if(F->p1.y < miny) miny=F->p1.y; if(F->p1.z < minz) minz=F->p1.z;
        if(F->p2.x > maxx) maxx=F->p2.x; if(F->p2.y > maxy) maxy=F->p2.y; if(F->p2.z > maxz) maxz=F->p2.z;
        if(F->p2.x < minx) minx=F->p2.x; if(F->p2.y < miny) miny=F->p2.y; if(F->p2.z < minz) minz=F->p2.z;
        if(F->p3.x > maxx) maxx=F->p3.x; if(F->p3.y > maxy) maxy=F->p3.y; if(F->p3.z > maxz) maxz=F->p3.z;
        if(F->p3.x < minx) minx=F->p3.x; if(F->p3.y < miny) miny=F->p3.y; if(F->p3.z < minz) minz=F->p3.z;
		AEL += 0.5f*(Betrag(F->p1 - F->p0)+Betrag(F->p2 - F->p1));
	  } else if((groupcode == 0) && !strcmp(gvalue,"POLYLINE")){
	    char *fnam = NULL; int flags=0,vcn,vcm; int j,i,ii,in,v1,v2,v3,v4; float len=0.0f;
		vector P,Az=vector(0.0f, 0.0f, 1.0f);
	    do {
		  GetGroup(infile);
		  switch(groupcode){
		    case   5: fnam  =      strdup(gvalue); break;
		    case   8:             StLayer(gvalue); break;
            case  39: len   = (float)atof(gvalue); break;
            case  70: flags =        atoi(gvalue); break;
            case  71: vcn   =        atoi(gvalue); break;
            case  72: vcm   =        atoi(gvalue); break;
            case 210: Az.x  = (float)atof(gvalue); break;
            case 220: Az.y  = (float)atof(gvalue); break;
            case 230: Az.z  = (float)atof(gvalue); break;
		  }
	    } while(groupcode!=0);

		if(flags & 64){
          TextWin.Output(0,"  - Polymesh: %d vertices, %d faces",vcn,vcm);
          p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p; pc=1;
          p->name = fnam? fnam : strdup("POLYMESH");
		  p->nv = vcn; p->pts=new(vector[p->nv]);

          for(i=0; i<vcn; i++){
	        do {
		      GetGroup(infile);
		      switch(groupcode){
                case 10: p->pts[i].x = (float)atof(gvalue); break;
                case 20: p->pts[i].y = (float)atof(gvalue); break;
                case 30: p->pts[i].z = (float)atof(gvalue); break;
		      }
	        } while(groupcode!=0);
            if(p->pts[i].x > maxx) maxx=p->pts[i].x; if(p->pts[i].y > maxy) maxy=p->pts[i].y; if(p->pts[i].z > maxz) maxz=p->pts[i].z;
            if(p->pts[i].x < minx) minx=p->pts[i].x; if(p->pts[i].y < miny) miny=p->pts[i].y; if(p->pts[i].z < minz) minz=p->pts[i].z;
		  }

          p->nf = vcm; p->pnv=new(byte[vcm]); p->pfv=new(ushort[4*p->nf]); int mnv=3;
          for(ii=0,i=0; i<vcm; i++){
		    v4=-1;
	        do {
		      GetGroup(infile);
		      switch(groupcode){
                case 71: v1 = abs(atoi(gvalue))-1; break;
                case 72: v2 = abs(atoi(gvalue))-1; break;
                case 73: v3 = abs(atoi(gvalue))-1; break;
                case 74: v4 = abs(atoi(gvalue))-1; break;
		      }
	        } while(groupcode!=0);
		    p->pfv[ii++]=v1; p->pfv[ii++]=v2; p->pfv[ii++]=v3;
		    if(v4>-1){
			  p->pfv[ii++]=v4;
			  p->pnv[i]=4;
			  mnv=4;
		    } else p->pnv[i]=3;
		  }
	      if(mnv==3){ delete p->pnv; p->pnv=NULL; }

		} else if(flags & 16){
          TextWin.Output(0,"  - quadmesh: %d x %d",vcn,vcm);
          p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p; pc=1;
          p->name = fnam? fnam : strdup("POLYMESH");
		  p->nv = vcn*vcm; p->pts=new(vector[p->nv]);

          for(i=0; i<(p->nv); i++){
	        do {
		      GetGroup(infile);
		      switch(groupcode){
                case 10: p->pts[i].x = (float)atof(gvalue); break;
                case 20: p->pts[i].y = (float)atof(gvalue); break;
                case 30: p->pts[i].z = (float)atof(gvalue); break;
		      }
	        } while(groupcode!=0);
            if(p->pts[i].x > maxx) maxx=p->pts[i].x; if(p->pts[i].y > maxy) maxy=p->pts[i].y; if(p->pts[i].z > maxz) maxz=p->pts[i].z;
            if(p->pts[i].x < minx) minx=p->pts[i].x; if(p->pts[i].y < miny) miny=p->pts[i].y; if(p->pts[i].z < minz) minz=p->pts[i].z;
		  }

          p->nf = ((flags & 32)? vcm : (vcm-1)) * ((flags & 1)? vcn : (vcn-1)); p->pnv=new(byte[p->nf]); p->pfv=new(ushort[4*(p->nf)]);
		  for(ii=0,in=0,j=0; j<(vcn-1); j++){
		    for(i=0; i<(vcm-1); i++){
		      p->pfv[ii++]=i+j*(vcm); p->pfv[ii++]=i+1+j*(vcm); p->pfv[ii++]=vcm+i+1+j*(vcm); p->pfv[ii++]=vcm+i+j*(vcm);
		      p->pnv[in++]=4;
			}
			if(flags & 32){
		      p->pfv[ii++]=i+j*(vcm); p->pfv[ii++]=j*(vcm); p->pfv[ii++]=vcm+j*(vcm); p->pfv[ii++]=vcm+i+j*(vcm);
		      p->pnv[in++]=4;
			}
		  }
		  if(flags & 1){
		    for(i=0; i<(vcm-1); i++){
		      p->pfv[ii++]=i+j*(vcm); p->pfv[ii++]=i+1+j*(vcm); p->pfv[ii++]=i+1; p->pfv[ii++]=i;
		      p->pnv[in++]=4;
			}
			if(flags & 32){
		      p->pfv[ii++]=i+j*(vcm); p->pfv[ii++]=j*(vcm); p->pfv[ii++]=0; p->pfv[ii++]=i;
		      p->pnv[in++]=4;
			}
		  }
		  // reorder vertex-pointers clockwise
		  for(ii=0,i=0; i<(p->nf); i++){
			for(j=0; j<2; j++){
			  swap(p->pfv[ii+j], p->pfv[ii+3-j]);
			}
			ii+=4;
		  }
		} else if(len){
          TextWin.Output(0,"  - extruded polyline: ");
          vector pt[20]; double arc[20]; int vc=0;
		  do{
			arc[vc]=0;
		    do {
		      GetGroup(infile);
		      switch(groupcode){
                case 10: pt [vc].x = (float)atof(gvalue); break;
                case 20: pt [vc].y = (float)atof(gvalue); break;
                case 30: pt [vc].z = (float)atof(gvalue); break;
				case 42: arc[vc]   =        atof(gvalue); break;
		      }
	        } while(groupcode!=0);
			vc++;
          } while(!strcmp(gvalue,"VERTEX"));

          TextWin.Output(0,"%d vertices\n",vc);
		  p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p; pc=1;
          p->name = fnam? fnam : strdup("POLYMESH");
		  p->nv = 2*vc; p->pts=new(vector[p->nv]);

		  vector Ax = unit( (((fabs(Az.x) < OB64) && (fabs(Az.y) < OB64))? vector(0.0f, 1.0f, 0.0f) : vector(0.0f, 0.0f, 1.0f)) * Az);
          vector Ay = unit( Az * Ax); matrix TR;
		  TR.M11=Ax.x; TR.M12=Ax.y; TR.M13=Ax.z;
          TR.M21=Ay.x; TR.M22=Ay.y; TR.M23=Ay.z;
          TR.M31=Az.x; TR.M32=Az.y; TR.M33=Az.z;
          TR.M41=0.0f; TR.M42=0.0f; TR.M43=0.0f;

          for(i=0; i<vc; i++){
			p->pts[2*i  ]=pt[i]*TR;
			p->pts[2*i+1]=p->pts[2*i]+len*Az;
            if(p->pts[2*i  ].x > maxx) maxx=p->pts[2*i  ].x; if(p->pts[2*i  ].y > maxy) maxy=p->pts[2*i  ].y; if(p->pts[2*i  ].z > maxz) maxz=p->pts[2*i  ].z;
            if(p->pts[2*i  ].x < minx) minx=p->pts[2*i  ].x; if(p->pts[2*i  ].y < miny) miny=p->pts[2*i  ].y; if(p->pts[2*i  ].z < minz) minz=p->pts[2*i  ].z;
            if(p->pts[2*i+1].x > maxx) maxx=p->pts[2*i+1].x; if(p->pts[2*i+1].y > maxy) maxy=p->pts[2*i+1].y; if(p->pts[2*i+1].z > maxz) maxz=p->pts[2*i+1].z;
            if(p->pts[2*i+1].x < minx) minx=p->pts[2*i+1].x; if(p->pts[2*i+1].y < miny) miny=p->pts[2*i+1].y; if(p->pts[2*i+1].z < minz) minz=p->pts[2*i+1].z;
		  }

          p->nf = vc-1; if(flags&1)p->nf++;
		  p->pnv=new(byte[p->nf]); p->pfv=new(ushort[4*(p->nf)]);
          for(i=0,j=0; i<(vc-1); i++){
			if(arc[i]){
              Cylinder *c = new Cylinder; c->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=c; VMDoc->VMF.Base=c;
              c->name = strdup("beveled edge");
              if(!VMF.MatList){
	            VMF.MatList=new Material; VMF.MatList->name = strdup("default");
			  }
              c->Mat=VMF.MatList;
		      vector pm=0.5f*   (pt[i]+pt[i+1]);
		      float   a=distance(pt[i],pt[i+1]);
		      double  h=arc[i]*0.5*a;
		      double  w=4.0*atan(2.0*h/a);
		      double  b=0.5*(a*tan(0.5*w));
		      vector dr=pt[i+1]-pt[i];
		      vector ce=pm+(float)b*unit(vector(-dr.y,dr.x,0.0f));
		      float rad=distance(pt[i],ce);
		      vector d0=pt[i]-ce;
		      float  w0=ang(d0.x,d0.y);
		      c->Scale(vector(rad,rad,(float)fabs(len)));
		      if(w>0.0){
		        c->RotateZ(w0         ); c->angle=(float)    (w);
			  } else {
		        c->RotateZ(w0+(float)w); c->angle=(float)fabs(w);
			  }
		      c->Translate(ce);
		      c->TM = c->TM*TR; c->Tb = c->TM.Inverse(); c->NM = c->TM.NTrans();
			  (p->nf)--;
			} else {
		      p->pnv[j]=4;
			  p->pfv[4*j]=2*i; p->pfv[4*j+1]=2*i+1; p->pfv[4*j+2]=2*(i+1)+1; p->pfv[4*j+3]=2*(i+1); j++;
			}
		  }
          if(flags&1){
			if(arc[i]){
              Cylinder *c = new Cylinder; c->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=c; VMDoc->VMF.Base=c;
              c->name = strdup("beveled edge");
              if(!VMF.MatList){
	            VMF.MatList=new Material; VMF.MatList->name = strdup("default");
			  }
              c->Mat=VMF.MatList;
		      vector pm=0.5f*   (pt[i]+pt[0]);
		      float   a=distance(pt[i],pt[0]);
		      double  h=arc[i]*0.5*a;
		      double  w=4.0*atan(2.0*h/a);
		      double  b=0.5*(a*tan(0.5*w));
		      vector dr=pt[0]-pt[i];
		      vector ce=pm+(float)b*unit(vector(-dr.y,dr.x,0.0f));
		      float rad=distance(pt[i],ce);
		      vector d0=pt[i]-ce;
		      float  w0=ang(d0.x,d0.y);
		      c->Scale(vector(rad,rad,(float)fabs(len)));
		      if(w>0.0){
		        c->RotateZ(w0         ); c->angle=(float)    (w);
			  } else {
		        c->RotateZ(w0+(float)w); c->angle=(float)fabs(w);
			  }
		      c->Translate(ce);
		      c->TM = c->TM*TR; c->Tb = c->TM.Inverse(); c->NM = c->TM.NTrans();
			  (p->nf)--;
			} else {
		      p->pnv[j]=4;
			  p->pfv[4*j]=2*i; p->pfv[4*j+1]=2*i+1; p->pfv[4*j+2]=1; p->pfv[4*j+3]=0; j++;
			}
		  }
		}
	  } else {
	    do GetGroup(infile); while(groupcode != 0);
	  }
	  if(pc){
        if(!VMF.MatList){
	      VMF.MatList=new Material; VMF.MatList->name = strdup("default");
		}
        p->Mat=VMF.MatList;

        TextWin.Output(0," - building mesh");
        //p->BuildMesh();
        TextWin.Output(0," - done.\n");
	  }
	  //do GetGroup(infile); while(groupcode != 0);
	  //while(groupcode != 0) GetGroup(infile);
	}
    TextWin.Output(0,"  - Layers:\n");
	for(LayPtr L=Layers; L; L=L->next){
      TextWin.Output(0,"    - '%s'\n",L->name);
    }
	if(Fc){
      TextWin.Output(0,"  - %d 3DFACEs (AEL: %f):\n",Fc,AEL/Fc);
	  while(Faces){
	    G=Faces; Faces=Faces->next; F=Faces; Fc=1;
	    while(F){
	      if(F->Lay == G->Lay){
	        if(F==Faces){
		      Faces = F->next;
		      F->next=G; G=F; F=Faces;
		    } else {
		      P->next=F->next;
		      F->next=G; G=F; F=P->next;
		    }
		    Fc++;
	      } else {
		    P=F; F=F->next;
	      }
	    }
        p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
        p->name = strdup(G->Lay->name);
        p->nf = Fc; p->pnv=new(byte[p->nf]); p->pfv=new(ushort[4*p->nf]); int mnv=3;
        TextWin.Output(0,"    - Object '%s' (%d faces)",p->name,Fc);
        VectorSet VB; int i,ii; float Ep = AEL/(1000.0f*Fc); F=G;
	    for(ii=0,i=0; i<(p->nf); i++){
	      p->pfv[ii++]=VB.Add(F->p0, Ep);
	      p->pfv[ii++]=VB.Add(F->p1, Ep);
	      p->pfv[ii++]=VB.Add(F->p2, Ep);
	      if(Betrag(F->p2 - F->p3) > Ep){
	        p->pfv[ii++]=VB.Add(F->p3, Ep);
	        p->pnv[i]=4;
		    mnv=4;
	      } else p->pnv[i]=3;
	      F=F->next;
	    }
	    if(mnv==3){ delete p->pnv; p->pnv=NULL; }

        // copy vertices from Vertex-Buffer and delete Vertex-Buffer
        p->nv = VB.Move2Array(&(p->pts));

	    if(!VMF.MatList){
	      VMF.MatList=new Material; VMF.MatList->name = strdup("default");
	    }
        p->Mat=VMF.MatList;

        TextWin.Output(0," - building mesh");
        //p->BuildMesh();
        TextWin.Output(0," - done.\n");
	  }
	}

    // set up a view
	vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
    float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
    V->Locate(tp+vector(1.0F*td,-2.5F*td,1.5F*td));
    V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
	V->SetViewAngle(0.5F); WorldValid=FALSE;
    vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 0.0f, 1.0f)-(vector(0.0f, 0.0f, 1.0f)%N)*N);
    strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
    fclose(infile);
  } else TextWin.Output(3,"- ERROR: couldn't open input-file!\n");
}

void freadline(FILE *file, char *buffer)
{
  char c=fgetc(file); int i=0;
  while((c!=10) && (c!=13) && !feof(file)){
    buffer[i++]=c; c=fgetc(file);
  }
  buffer[i]=0;
}

char  *LBP;

float getflt()
{
  char  token[20];
  int   i=0;
  while(*LBP==' ') LBP++;
  while((*LBP!=' ') && (*LBP!=0)){ token[i++]=*LBP; LBP++; }
  token[i]=0;
  return (float)atof(token);
}

int getint()
{
  char  token[20];
  int   i=0;
  while((*LBP==' ') || (*LBP=='/')) LBP++;
  while((*LBP!=' ') && (*LBP!='/') && (*LBP!=0)){ token[i++]=*LBP; LBP++; }
  token[i]=0;
  return atoi(token);
}


vector CalcNormal(vector p1, vector p2, vector p3)
{
  vector normal = (p2-p1) * (p3-p1);
  if(Betrag(normal) < (Betrag(p2-p1) * Betrag(p3-p1) * 0.001f)){
    normal = vector(0.0F, 1.0F, 0.0F);
  } else {
    normal = unit(normal);
  }
  return normal;
}


typedef struct pfce  *PfcPtr;
        struct pfce { PfcPtr nxt; MatPtr mat; char *nam; ushort *vrt; ushort *nrm; ushort *txc; byte npv; byte grp; };

typedef struct onam  *ONPtr;
        struct onam { ONPtr nxt; char *nam; };

void CVModelDoc::ImportOBJ(char *filename, ViewPtr V)
{
  FILE      *infile;
  char       LBuffer[1024];
  int        va[64];
  int        na[64];
  int        ta[64];
  PfcPtr     FRoot=NULL;
  PfcPtr     FPntr=NULL;
  PfcPtr     FObj,FL;
  ONPtr      names=NULL;
  ONPtr      np,cn;
  MatPtr     mta[128],mp,cm=NULL;
  vector    *vert=NULL;
  vector    *norm=NULL;
  vector    *texc=NULL;
  int        c,i,ii,ap,nv,nn,nt,vp,pc,fc,vc,vm,mc,nvn,ntc;
  int        sgr,sm,sg=0;
  PolyMesh  *p;

  float   maxx=-70000.0F;
  float   maxy=-70000.0F;
  float   maxz=-70000.0F;
  float   minx=+70000.0F;
  float   miny=+70000.0F;
  float   minz=+70000.0F;

  if(infile = fopen(filename,"rb")){
    TextWin.Output(1,"  counting vertices ... "); nv=0; nn=0; nt=0;
    while(!feof(infile)){
      freadline(infile, LBuffer);
	  for(i=0; (LBuffer[i]==9) || (LBuffer[i]==32); i++);
      if(LBuffer[i]=='v'){
		if(LBuffer[i+1]==' ') nv++;
		if(LBuffer[i+1]=='t') nt++;
		if(LBuffer[i+1]=='n') nn++;
	  }
    }
    vert = new(vector[nv]); vp=0;
    TextWin.Output(1,"\n  reading %d vertices ... ",nv);
    rewind(infile);
    while(!feof(infile)){
      freadline(infile,LBuffer);
	  for(i=0; (LBuffer[i]==9) || (LBuffer[i]==32); i++); LBP=LBuffer+i+2;
      if((LBuffer[i]=='v') && (LBuffer[i+1]==' ')){
        vert[vp].x = 1000.0f*getflt();
        vert[vp].y = 1000.0f*getflt();
        vert[vp].z = 1000.0f*getflt();
        if(vert[vp].x > maxx) maxx=vert[vp].x;
        if(vert[vp].y > maxy) maxy=vert[vp].y;
        if(vert[vp].z > maxz) maxz=vert[vp].z;
        if(vert[vp].x < minx) minx=vert[vp].x;
        if(vert[vp].y < miny) miny=vert[vp].y;
        if(vert[vp].z < minz) minz=vert[vp].z;
        vp++;
      }
    }
	if(nn){
      norm = new(vector[nn]); vp=0;
      TextWin.Output(1,"\n  reading %d normals ... ",nn);
      rewind(infile);
      while(!feof(infile)){
        freadline(infile,LBuffer);
	    for(i=0; (LBuffer[i]==9) || (LBuffer[i]==32); i++); LBP=LBuffer+i+2;
        if((LBuffer[i]=='v') && (LBuffer[i+1]=='n')){
          norm[vp].x = getflt();
          norm[vp].y = getflt();
          norm[vp].z = getflt();
          vp++;
        }
      }
	}
	if(nt){
      texc = new(vector[nt]); vp=0;
      TextWin.Output(1,"\n  reading %d texture coords ... ",nt);
      rewind(infile);
      while(!feof(infile)){
        freadline(infile,LBuffer);
	    for(i=0; (LBuffer[i]==9) || (LBuffer[i]==32); i++); LBP=LBuffer+i+2;
        if((LBuffer[i]=='v') && (LBuffer[i+1]=='t')){
          texc[vp].x = getflt();
          texc[vp].y = getflt();
          texc[vp].z = getflt();
          vp++;
        }
      }
	}
    TextWin.Output(1,"\n  reading faces ... ");
    rewind(infile); fc=0;
    names=new(struct onam); names->nam=strdup("Object"); names->nxt=NULL; cn=names;
    while(!feof(infile)){
      freadline(infile,LBuffer);
	  for(i=0; (LBuffer[i]==9) || (LBuffer[i]==32); i++); LBP=LBuffer+i+2;
      if((LBuffer[i]=='g') && (LBuffer[i+1]==' ') && (LBuffer[i+2]!=0)){
        np=names;
        while(np && strcmp(np->nam, &(LBuffer[i+2]))) np=np->nxt;
        if(!np){
          np=new(struct onam);
          np->nxt=names; names=np;
          np->nam=strdup(&(LBuffer[i+2]));
        }
        cn=np;
      } else if(!strnicmp(LBuffer,"usemtl ",7)){
        mp=VMF.MatList;
        while(mp && strncmp(mp->name, &(LBuffer[7]), 18)) mp=mp->next;
        if(!mp){
          mp=new Material; VMF.MatList=mp; mp->name=(char *)malloc(24); strncpy(mp->name, &(LBuffer[7]), 18);
        }
        cm=mp;
      } else if((LBuffer[i]=='s') && (LBuffer[i+1]==' ') && (LBuffer[i+2]!=0)){
        sg=getint();
      } else if((LBuffer[i]=='f') && ( (LBuffer[i+1]==' ') || ((LBuffer[i+1]=='o') && (LBuffer[i+2]==' ')) )){
        vc=-1; fc++;
        do {
                        va[++vc]=getint();                // vertex number
          if(*LBP=='/') ta[  vc]=getint(); else ta[vc]=0; // texture coordinates
          if(*LBP=='/') na[  vc]=getint(); else na[vc]=0; // phong normals
        } while(va[vc]);
        if(FPntr){
          FPntr->nxt  = new(struct pfce); FPntr=FPntr->nxt;
        } else {
          FRoot=FPntr = new(struct pfce);
        }
        FPntr->nxt = NULL;
        FPntr->mat = cm;
        FPntr->nam = cn->nam;
        FPntr->npv = vc;
        FPntr->grp = sg;
        FPntr->vrt = new(ushort[vc]); for(    i=0; i<vc; i++)     (FPntr->vrt)[i]=va[i]-1;
        FPntr->nrm = new(ushort[vc]); for(c=0,i=0; i<vc; i++) c+=((FPntr->nrm)[i]=na[i] ); if(!c){ delete FPntr->nrm; FPntr->nrm=NULL; }
        FPntr->txc = new(ushort[vc]); for(c=0,i=0; i<vc; i++) c+=((FPntr->txc)[i]=ta[i] ); if(!c){ delete FPntr->txc; FPntr->txc=NULL; }
      }
    }
    TextWin.Output(1,"- %d faces found.\n",fc);

    /*p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
	p->nv=nv; p->pts=vert;
	p->nf=fc; p->pnv = new(byte[p->nf]); p->pfv = new(ushort[FPntr->npv]);

    // default-transformation:
    p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);*/

    while(FRoot){
      FObj=FRoot; FRoot=FRoot->nxt; FPntr=FRoot; FObj->nxt=NULL; FL=NULL;
      TextWin.Output(1,"  - converting object '%s' ",FObj->nam);
      while(FPntr){
        if(FPntr->nam == FObj->nam){
          if(FL){
            FL->nxt=FPntr->nxt;
            FPntr->nxt=FObj;
            FObj=FPntr;
            FPntr=FL->nxt;
          } else {
            FRoot=FPntr->nxt;
            FPntr->nxt=FObj;
            FObj=FPntr;
            FPntr=FRoot;
          }
        } else {
          FL=FPntr; FPntr=FPntr->nxt;
        }
      }

      // count polygons, vertices, materials, ...
      FPntr=FObj; pc=0; vc=0; vm=0; nvn=0; ntc=0; sgr=FPntr->grp; sm=0; mc=0;
      while(FPntr){
		pc++; vc+=FPntr->npv;
		if((FPntr->npv)>vm  ) vm=vc;
		if( FPntr->nrm      ) nvn++;
		if( FPntr->txc      ) ntc++;
		if((FPntr->grp)!=sgr) sm=1;
		if((FPntr->mat)     ) { for(i=0; (i<mc) && (mta[i]!=(FPntr->mat)); i++); if(i>=mc) mta[mc++]=(FPntr->mat); }
		FPntr=FPntr->nxt;
	  } if(nvn) sm=0;
      TextWin.Output(1,"(%d faces)\n",pc);

      p = new PolyMesh; p->succ=VMF.Base; if(VMF.Base) VMF.Base->pred=p; VMF.Base=p;
      p->pmflags = sm? SURF_GROUPED : (((sgr!=0) && !nvn)? SURF_ROUNDED : 0);
      p->nf = pc;
	  if(mc>1){ p->mtl = new(MatPtr[mc]); p->nm=mc; memcpy(p->mtl, mta, mc*sizeof(MatPtr));
	            p->mat = new(  byte[pc]); }
	  if(vm>3)  p->pnv = new(  byte[pc]);
	  if(sm  )  p->grp = new(  byte[pc]);
	            p->pfv = new(ushort[vc]);
      if(nvn)   p->pfn = new(ushort[vc]);
      if(ntc)   p->pft = new(ushort[vc]);
      VectorSet VB,NB,TB; TextWin.Output(1,"    collecting vertices ...\n");
      FPntr=FObj; ap=0; ii=0;
      while(FPntr){
        if(vm>3)                                         p->pnv[ii]=FPntr->npv;
        if(sm  )                                         p->grp[ii]=FPntr->grp;
		if(mc>1){ for(i=0; (mta[i]!=(FPntr->mat)); i++); p->mat[ii]=i;        }
		ii++;
        for(i=0; i<(FPntr->npv); i++){
          p->pfv[ap]=VB.Add(vert[(FPntr->vrt)[(FPntr->npv)-1-i]], 0.0001f);
          if(nvn){
		    if(FPntr->nrm && ((FPntr->nrm)[(FPntr->npv)-1-i]))
              p->pfn[ap]=NB.Add(-norm[(FPntr->nrm)[(FPntr->npv)-1-i]-1], 0.0001f);
		    else
              p->pfn[ap]=NB.Add(CalcNormal(vert[(FPntr->vrt)[(FPntr->npv)-1]], vert[(FPntr->vrt)[(FPntr->npv)-2]], vert[(FPntr->vrt)[(FPntr->npv)-3]]), 0.0001f);
		  }
          if(ntc){
		    if(FPntr->txc && ((FPntr->txc)[(FPntr->npv)-1-i]))
              p->pft[ap]=TB.Add(texc[(FPntr->txc)[(FPntr->npv)-1-i]-1], 0.0001f);
		    else
              p->pft[ap]=TB.Add(vector(0.0f, 0.0f, 0.0f), 0.0001f);
		  }
		  ap++;
        }
        FPntr=FPntr->nxt;
      }
              p->nv = VB.Move2Array(&(p->pts));
      if(nvn) p->nn = NB.Move2Array(&(p->nrm));
      if(ntc) p->nt = TB.Move2Array(&(p->txc));
      p->name = strdup(FObj->nam);

      if(mp){
		p->Mat=mp;
	  } else {
	    MatPtr mat = VMF.MatList; while(mat && strcmp(mat->name,"default")) mat=mat->next;
        if(!mat){
          mat=new Material; mat->next=VMF.MatList; VMF.MatList=mat;
          mat->name = strdup("default");
        }
        p->Mat=mat;
	  }
      // default-transformation:
      p->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);

      TextWin.Output(1,"    building mesh");
      //p->BuildMesh();
      TextWin.Output(1,", orienting normals");
      int edg=0,crs=0;
      //for(i=0; i<(p->mesh.nl); i++){
      //if((p->mesh.line[i].n2   ) == NO_NORMAL) edg++;
      //if((p->mesh.line[i].flags) &  CROSSLINE) crs++;
      //}
      if(crs){
        p->pmflags |= SURF_DOUBLESIDED;
        //p->BuildMesh();
      } else {
        if(edg) p->pmflags |= SURF_DOUBLESIDED;
        //p->BuildMesh(); //UnifyNormals(p);
      }
      TextWin.Output(1," - done.\n");

      while(FObj){
		if(FObj->vrt) delete FObj->vrt;
		if(FObj->nrm) delete FObj->nrm;
		if(FObj->txc) delete FObj->txc;
		FPntr=FObj; FObj=FObj->nxt; delete FPntr;
	  }
    }

    while(names){ np=names; names=names->nxt; delete np; }
    delete vert; if(norm) delete norm; if(texc) delete texc;
    fclose(infile);
    vector tp = 0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
    float  td = Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
    V->Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
    V->LookAt(tp); V->Up=vector(0.0f, 0.0f, 1.0f);
	V->SetViewAngle(0.5F); WorldValid=FALSE;
    vector N = unit(V->Target - V->Location); V->Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
    strcpy(VMF.Object,filename); strcpy(VMF.Author,"?"); strcpy(VMF.Date,"?");
  } else TextWin.Output(3,"- can't open file '%s'!\n",filename);
}


