// Virtual Modeller V0.5GL  for MS Visual C++ V4.2   12.12.1997  Michael Granz

#include "stdafx.h"
#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "TextWin.h"

#include "gl\gl.h"
#include "gl\glu.h"

#define MAXPVERT 64

BOOL      WorldValid;   // quader-representation of the hole world

void GetExtrema(vector V, vector *EMin, vector *EMax)
{
  if(V.x > EMax->x) EMax->x = V.x;
  if(V.y > EMax->y) EMax->y = V.y;
  if(V.z > EMax->z) EMax->z = V.z;
  if(V.x < EMin->x) EMin->x = V.x;
  if(V.y < EMin->y) EMin->y = V.y;
  if(V.z < EMin->z) EMin->z = V.z;
}

extern void	SetTexture(TexMapPtr map, int TexMode);

void BuildWorldQuader(PrimPtr p)
{
  int    i;
  vector WMin = vector( 9999.9f,  9999.9f,  9999.9f);
  vector WMax = vector(-9999.9f, -9999.9f, -9999.9f);
  while(p){
    if((p->idnt==POLYMESH) || (p->idnt==CUBE) || (p->idnt==FUNCTION)){
      switch(p->idnt){
        case CUBE:
          GetExtrema(vector(-0.5f, -0.5f, -0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector(-0.5f, -0.5f,  0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector( 0.5f, -0.5f,  0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector( 0.5f, -0.5f, -0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector(-0.5f,  0.5f, -0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector(-0.5f,  0.5f,  0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector( 0.5f,  0.5f,  0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
          GetExtrema(vector( 0.5f,  0.5f, -0.5f) * ((Cube *)p)->TM, &WMin, &WMax);
		  break;
        case FUNCTION:
          GetExtrema(((Function *)p)->Min * ((Function *)p)->TM, &WMin, &WMax);
          GetExtrema(((Function *)p)->Max * ((Function *)p)->TM, &WMin, &WMax);
		  break;
        case POLYMESH:
          for(i=0; i<((PolyMesh *)p)->nv; i++){
            GetExtrema(((PolyMesh *)p)->pts[i] * ((PolyMesh *)p)->TM, &WMin, &WMax);
          }
		  break;
      }
    }
    p=p->succ;
  }
  VMDoc->WMP = 0.5*(WMin+WMax); VMDoc->WRD = Betrag(WMax-(VMDoc->WMP));
  if(VMDoc->VMF.Sky){
    VMDoc->VMF.Sky->TM.InitTransformation(10.0f*(VMDoc->WRD), 10.0f*(VMDoc->WRD), 10.0f*(VMDoc->WRD), 0.0f, 0.0f, 0.0f, VMDoc->WMP.x, VMDoc->WMP.y, VMDoc->WMP.z);
    VMDoc->VMF.Sky->Tb = VMDoc->VMF.Sky->TM.Inverse(); // calculate backtransformation matrix
    VMDoc->VMF.Sky->NM = VMDoc->VMF.Sky->TM.NTrans();  // calculate transformation matrix for normals
	VMDoc->VMF.Sky->dflags &= ~1;
  }
  WorldValid=TRUE;
}


void DelIsoVert(PrimPtr p, BOOL stat)
{
  PolyMesh  *t;
  vector    *pts;
  ushort    *vuct;
  int        vc=0;
  int        i,j,k;
  if(p->idnt==POLYMESH){
    t=(PolyMesh *)p;
    vuct=new(ushort[t->nv]);
    // reset usage-counters
    for(i=0; i<(t->nv); i++) vuct[i]=0;
    // count vertex-usage
    for(k=0,i=0; i<(t->nf); i++){
	  for(j=0; j<(t->pnv? t->pnv[i] : 3); j++){
        vuct[t->pfv[k++]]++;
	  }
    }
    // count unused vertices
    for(i=0; i<(t->nv); i++) if(vuct[i]==0) vc++;
    if(vc){
	  TextWin.Output(1,"  %d isolated %s found",vc,(vc==1? "vertex" : "vertices"));
      pts = new(vector[(t->nv)-vc]);
      // copy used vertices and store new index
      for(j=0,i=0; i<(t->nv); i++){
        if(vuct[i]) pts[j++]=t->pts[i];
        vuct[i]=j-1;
      }
      // replace vertex-list
      delete t->pts; t->pts=pts; t->nv -= vc;
      // change vertex-indices in face-list
      for(k=0,i=0; i<(t->nf); i++){
   	    for(j=0; j<(t->pnv? t->pnv[i] : 3); j++){
          t->pfv[k] = vuct[t->pfv[k]]; k++;
	    }
      }
      TextWin.Output(1," and removed.\n");
    } else if(stat) TextWin.Output(1,"  no isolated vertices found.\n");
    delete vuct; //if(vc) t->BuildMesh();
  }
}


void UnifyNormals(PrimPtr p)
{
  if(p->idnt==POLYMESH){
    PolyMesh *tp = (PolyMesh *)p;
    if(!(tp->nn)){
      int ii,jj,vi,ki,vj,kj;
      int i,j,k,c,g,m;
      int ai,aj,ni[MAXPVERT],nj[MAXPVERT];
      if(!(tp->flg)){ tp->flg = new(byte[tp->nf]); for(k=0; k<tp->nf; k++) tp->flg[k] = 0; }
      for(i=0; i<(tp->nf); i++) tp->flg[i] &= (~JUSTIFIED & ~FINISHED);
      tp->flg[0] |= JUSTIFIED;
      do{ c=0;
        for(i=0,vi=0; i<(tp->nf); i++,vi+=ki){
		  ki = tp->pnv? tp->pnv[i] : 3;
          if( (!(tp->flg[i] & FINISHED)) && (tp->flg[i] & JUSTIFIED) ){
            for(j=0,vj=0; j<(tp->nf); j++,vj+=kj){
			  kj = tp->pnv? tp->pnv[j] : 3;
              if(!(tp->flg[j] & JUSTIFIED)){ g=0;
                for(ii=0; ii<ki; ii++){
				  for(jj=0; jj<kj; jj++){
					if(tp->pfv[vi+ii] == tp->pfv[vj+jj]){
					  ni[g]=ii; nj[g]=jj; g++;
					}
				  }
				}
                if(g==2){
                  ai = ( (ni[1]==ni[0]+1) || ((ni[0]==ki-1) && (ni[1]==0)) );
                  aj = ( (nj[1]==nj[0]+1) || ((nj[0]==kj-1) && (nj[1]==0)) );
                  if( !( ai ^ aj ) ){
					// flip face
				    for(jj=0; jj<(kj>>1); jj++) swap(tp->pfv[vj+jj], tp->pfv[vj+kj-1-jj]);
                  }
                  tp->flg[j] |= JUSTIFIED;
                }
              }
            }
            tp->flg[i] |= FINISHED; c++;
          }
        }
      } while(c>0);
      // find 'highest' vertex
      for(m=0,i=1; i<(tp->nv); i++) if(tp->pts[i].y > tp->pts[m].y) m=i;
      // calculate surface normal at 'highest' vertex
      vector N = vector(0.0F, 0.0F, 0.0F);
      for(i=0,vi=0; i<(tp->nf); i++,vi+=ki){
        ki = tp->pnv? tp->pnv[i] : 3;
		for(ii=0; ii<ki; ii++){
          if(tp->pfv[vi+ii]==m){
            N=N+unit( (tp->pts[tp->pfv[vi+1]]-tp->pts[tp->pfv[vi]]) * (tp->pts[tp->pfv[vi+2]]-tp->pts[tp->pfv[vi]]) );
		  }
        }
      }
      // if normal points downwards then flip all faces
      if(N.y > 0.0F){
        for(i=0,vi=0; i<(tp->nf); i++,vi+=ki){
		  ki = tp->pnv? tp->pnv[i] : 3;
		  for(ii=0; ii<(ki>>1); ii++) swap(tp->pfv[vi+ii], tp->pfv[vi+ki-1-ii]);
		}
      }
      // rebuild mesh with 'unified' and 'oriented' surface-normals
      tp->dflags &= ~1;
	  if(tp->mesh.lin) delete tp->mesh.lin; tp->mesh.lin=NULL;
	  if(tp->mesh.snm) delete tp->mesh.snm; tp->mesh.snm=NULL;
    }
  }
}

void OptimizeMesh(PolyMesh *p)
{
  float      SMD0,SMD1,SMD,SZE,MinX,MinY,MinZ,MaxX,MaxY,MaxZ;
  float      d,amd=0.0f; int amc=0;
  int        GRP=0;
  ushort    *nfv;
  byte      *nnv;
  byte      *ngr;
  byte      *nmt;
  int        ov,ap,fc,vc,mv,SM,dc=0,mc=0;
  int        ki,kj,fi,fj,mi,mj,ii,jj,i2,j2,i,j,k;
  
  // calculate mesh dimensions
  MinX=MaxX=p->pts[0].x; MinY=MaxY=p->pts[0].y; MinZ=MaxZ=p->pts[0].z;
  for(i=0; i<(p->nv); i++){
	if(p->pts[i].x < MinX) MinX=p->pts[i].x; if(p->pts[i].x > MaxX) MaxX=p->pts[i].x;
	if(p->pts[i].y < MinY) MinY=p->pts[i].y; if(p->pts[i].y > MaxY) MaxY=p->pts[i].y;
	if(p->pts[i].z < MinZ) MinZ=p->pts[i].z; if(p->pts[i].z > MaxZ) MaxZ=p->pts[i].z;
  }
  SZE=SMD0=SMD1=Betrag(vector(MaxX,MaxY,MaxZ)-vector(MinX,MinY,MinZ));
  for(ii=0,i=0; i<(p->nf); i++){
	k=(p->pnv? p->pnv[i] : 3);
    for(j=1; j<k; j++){
	  d=distance(p->pts[p->pfv[ii+j]], p->pts[p->pfv[ii+j-1]]);
	  if(d<SMD0) SMD0=d; if((d<SMD1) && (d>(SZE/10000.0f))) SMD1=d; amd+=d; amc++;
	}
    d=distance(p->pts[p->pfv[ii]], p->pts[p->pfv[ii+k-1]]);
	if(d<SMD0) SMD0=d; if((d<SMD1) && (d>(SZE/10000.0f))) SMD1=d; amd+=d; amc++;
  }
  amd/=(float)amc;
  TextWin.Output(1,"- osize = %f, smd = %f(%f), avg = %f\n", SZE, SMD1, SMD0, amd);
  
  // delete isolated vertices
  DelIsoVert((PrimPtr) p, TRUE);

  // weld coincident vertices
  VectorSet VB; ov=p->nv; SMD = SMD1/2.0f;
  if(p->grp) GRP=1;
  for(ap=0,i=0; i<(p->nf); i++){
    // search/insert vertices in buffer and receive new index
    for(j=0; j<(p->pnv? p->pnv[i] : 3); j++){
      p->pfv[ap] = VB.Add(p->pts[p->pfv[ap]],SMD);
	  ap++;
    }
  }
  delete(p->pts);
  // copy vertices from Vertex-Buffer and delete Vertex-Buffer
  p->nv = VB.Move2Array(&(p->pts));
  TextWin.Output(1,"- %d vertices welded.\n", ov-(p->nv));

  // weld coincident texture coordinates
  if(ov=p->nt){
    for(ap=0,i=0; i<(p->nf); i++){
      // search/insert vertices in buffer and receive new index
      for(j=0; j<(p->pnv? p->pnv[i] : 3); j++){
        p->pft[ap] = VB.Add(p->txc[p->pft[ap]], 0.0001f);
	    ap++;
	  }
	}
    delete(p->txc);
    // copy vertices from Vertex-Buffer and delete Vertex-Buffer
    p->nt = VB.Move2Array(&(p->txc));
    TextWin.Output(1,"- %d texture coordinates welded.\n", ov - (p->nt));
  }

  // remove double faces
  if(p->flg){
    for(i=0; i<(p->nf); i++) p->flg[i] &= ~REMOVED;
  } else {
	p->flg = new(byte[p->nf]); for(k=0; k<p->nf; k++) p->flg[k] = 0;
  }
  for(fi=0,i=0; i<(p->nf); i++){
    ki = p->pnv? p->pnv[i] : 3;
	if(!(p->flg[i] & REMOVED)){
      for(fj=fi+ki,j=i+1; j<(p->nf); j++){
        kj = p->pnv? p->pnv[j] : 3;
	    if(!(p->flg[j] & REMOVED)){
	      for(mi=1,ii=0; ii<ki; ii++){
	        for(mj=0,jj=0; jj<kj; jj++){
		      if(p->pfv[fj+jj]==p->pfv[fi+ii]) mj=1;
		    }
	        if(mj!=1) mi=0;
	      }
	      if(mi){ if(ki<kj) p->flg[i]|=REMOVED; else p->flg[j]|=REMOVED; dc++; }
		}
	    fj+=kj;
	  }
	}
	fi+=ki;
  }
  
  if(dc){
	TextWin.Output(1,"- %d double faces found", dc);
    fc=0; vc=0; mv=0; SM=1;
    for(i=0; i<(p->nf); i++) if(!((p->flg[i]) & REMOVED)){
      fc++; vc+=(p->pnv? p->pnv[i] : 3);
    }
    if(fc > 0){
      nfv=new(ushort[vc]);
      if(p->pnv) nnv=new(byte[fc]);
	  if(p->grp) ngr=new(byte[fc]);
	  if(p->mat) nmt=new(byte[fc]);
      for(i=0,i2=0,j=0,j2=0; i<(p->nf); i++){
	    if(!((p->flg[i]) & REMOVED)) {
          for(k=0; k<(p->pnv? p->pnv[i] : 3); k++){
		    nfv[j2+k] = p->pfv[i2+k];
		  }
          if(p->pnv)   nnv[j]=k;
          if(p->grp)   ngr[j]=p->grp[i];
          if(p->mat) { nmt[j]=p->mat[i]; if(nmt[j]!=nmt[0]) SM=0; }
		  j2+=k; j++; if(k>mv) mv=k;
        }
	    i2+=(p->pnv? p->pnv[i] : 3);
	  }
      delete(p->pfv); p->pfv=nfv;
	  if(p->pnv){
	    delete(p->pnv);
	    if(mv>3) p->pnv=nnv; else { p->pnv=NULL; delete(nnv); }
	  }
	  if(p->grp){ delete(p->grp); p->grp=ngr; }
	  if(p->mat){
	    delete(p->mat);
	    if(SM){
	      p->Mat = p->mtl[nmt[0]]; p->nm=0;
		  delete(   nmt); p->mat=NULL;
		  delete(p->mtl); p->mtl=NULL;
	    } else p->mat=nmt;
      }
      p->nf = fc;
      TextWin.Output(1," and removed.\n");
    }
  } else {
	TextWin.Output(1,"- no double faces found.\n");
  }

  // reduce double materials in multimaterial meshes
  if(p->nm > 1){
    byte *newnum = new byte[p->nm]; int nnp=0;
	for(i=0; i < p->nm; i++){
	  for(j=0; (j<i) && (p->mtl[i] != p->mtl[j]); j++);
	  if(j<i){
		newnum[i]=j;
	  } else {
		p->mtl[nnp] = p->mtl[i];
		newnum[i]=nnp++;
	  }
	}
	TextWin.Output(1,"- %d double materials replaced\n", p->nm-nnp);
    if(nnp < p->nm){
      for(i=0; i<(p->nf); i++) p->mat[i] = newnum[p->mat[i]];
	  p->nm = nnp;
	}
	delete newnum;
  }

  p->dflags &= ~1;
  if (p->mesh.lin)        { delete p->mesh.lin; p->mesh.lin=NULL; }
  if (p->mesh.snm)        { delete p->mesh.snm; p->mesh.snm=NULL; }
  if((p->nrm) && !(p->nn)){ delete p->nrm;      p->nrm=NULL;      }
}

void SplitConnected(PolyMesh *r)
{
  MatPtr     mats[50];
  PolyMesh  *t;
  int        i,fi,j,fj,k,fk,m,n,i2,j2,sc,rc,fc,vc,mv,mp,mc=0;
  int        Conn,GRP=0;

              // clear flags
              if(r->flg){
			    for(k=0; k < r->nf; k++) r->flg[k] &= ~(MATCH | REMOVED);
			  } else {
			    r->flg = new(byte[r->nf]);
			    for(k=0; k < r->nf; k++) r->flg[k] = 0;
			  }
			  for(rc=0; rc<(r->nf);){
			    // find connected faces
			    for(i=0,fi=0; (r->flg[i] & REMOVED); fi += (r->pnv? r->pnv[i] : 3), i++);
			    r->flg[i] |= MATCH; fc=1;
				do{
			      for(sc=0,j=i+1,fj=fi+(r->pnv? r->pnv[i] : 3); j<(r->nf); fj += (r->pnv? r->pnv[j] : 3), j++){
				    if(!(r->flg[i] & REMOVED)){
				      for(k=0,fk=0; !(r->flg[j] & MATCH) && (k<(r->nf)); fk += (r->pnv? r->pnv[k] : 3), k++){
				        if(r->flg[k] & MATCH){
                          Conn=0;
					      for(m=0; !Conn && (m<(r->pnv? r->pnv[j] : 3)); m++)
                            for(n=0; !Conn && (n<(r->pnv? r->pnv[k] : 3)); n++)
						      if(r->pfv[fj+m] == r->pfv[fk+n]) Conn=1;
					      if(Conn){ r->flg[j] |= MATCH; fc++; sc++; }
						}
					  }
				    }
				  }
			    } while(sc>0);
			    TextWin.Output(1,"- %d of %d faces seperated.\n", fc, (r->nf)-rc);
                VectorSet VB; fc=0; vc=0;
                if(r->grp) GRP=1;
			    if(!(r->mat)){ for(mp=0; (mp<mc) && (r->Mat != mats[mp]); mp++); if(mp>=mc) mats[mc++]=r->Mat; }
                for(i=0,i2=0; i<(r->nf); i++){
                  if((r->flg[i] & MATCH) && !(r->flg[i] & REMOVED)){
                    // search/insert vertices in buffer and receive new index
                    for(k=0; k<(r->pnv? r->pnv[i] : 3); k++){
	    		      r->pfv[i2+k] = VB.Add(r->pts[r->pfv[i2+k]], 0.0001f);
	                }
			        // store material
				    if(r->mat){ for(mp=0; (mp<mc) && (r->mtl[r->mat[i]] != mats[mp]); mp++); if(mp>=mc) mats[mc++]=r->mtl[r->mat[i]]; }
                    vc+=k; fc++;
                  }
	              i2 += (r->pnv? r->pnv[i] : 3);
                }
                if(fc){
                  t = new PolyMesh; t->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=t; VMDoc->VMF.Base=t; VMDoc->WorldValid=FALSE;
                  t->TM      = r->TM;
                  t->pmflags = r->pmflags;
                  t->Mat     = r->Mat;
                  t->name    = new char[strlen(r->name)+1]; strcpy(t->name,r->name);
                  t->nf      = fc; t->pfv = new(ushort[vc]); t->pnv = new(byte[fc]);
                  t->nv      = VB.Move2Array(&(t->pts)); j=0; j2=0; mv=0;
		          if(GRP) { t->grp = new(byte[fc]); t->pmflags &= ~SURF_ROUNDED; t->pmflags |= SURF_GROUPED; }
			      if(mc>1){
			        t->mtl = new(MatPtr[mc]); for(mp=0; mp<mc; mp++) t->mtl[mp]=mats[mp];
			        t->mat = new(byte[fc]); t->nm = mc;
			      } else {
                    t->Mat = mats[0];
			      }
                  // copy faces to new Mesh
                  for(i=0,i2=0; i<(r->nf); i++){
                    if((r->flg[i] & MATCH) && !(r->flg[i] & REMOVED)){
                      for(k=0; k<(r->pnv? r->pnv[i] : 3); k++){
                        t->pfv[j2+k] = r->pfv[i2+k];
                      }
                      t->pnv[j]=k; if(k>mv) mv=k;
				      if(GRP) t->grp[j] = (r->grp? r->grp[i] : (r->pmflags & SURF_ROUNDED)? 255 : 0);
				      if(mc>1){
					    if(r->mat){
			              for(mp=0; r->mtl[r->mat[i]] != mats[mp]; mp++);
					    } else {
			              for(mp=0;        r->Mat     != mats[mp]; mp++);
					    }
					    t->mat[j]=mp;
					  }
                      r->flg[i] |= REMOVED; rc++; j2+=k; j++;
                    }
				    i2+=(r->pnv? r->pnv[i] : 3);
                  }
                  if(mv==3) { delete(t->pnv); t->pnv=NULL; }
                  //t->BuildMesh();
			    }
			  }
}

void SplitMaterials(PolyMesh *r)
{
  PolyMesh  *t;
  int        i,j,k,v1,v2,fc,nf,fv,nv,mv;

  for(i=0; i<r->nm; i++){
	nf=0; nv=0; mv=0;
	for(j=0; j<r->nf; j++){
      if(r->mat[j]==i){
        fv=((r->pnv)? r->pnv[j] : 3);
        nv+=fv; if(fv>mv) mv=fv;
		nf++;
	  }
	}
	if(nf){
      TextWin.Output(1,"- Material '%s': %d faces with %d vertices.\n", r->mtl[i]->name, nf ,nv);
      t = new PolyMesh; t->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=t; VMDoc->VMF.Base=t; VMDoc->WorldValid=FALSE;
      t->TM      = r->TM;
      t->pmflags = r->pmflags;
      t->Mat     = r->mtl[i];
      t->name    = strdup(r->mtl[i]->name);
      t->nf      = nf; t->pfv = new(ushort[nv]); if(mv>3) t->pnv = new(byte[nf]);
	  t->pts     = new vector[r->nv]; memcpy(t->pts, r->pts, r->nv*sizeof(vector)); t->nv=r->nv;
	  if(r->nt){
	    t->txc   = new vector[r->nt]; memcpy(t->txc, r->txc, r->nt*sizeof(vector)); t->nt=r->nt; t->pft = new(ushort[nv]);
	  }
	  for(j=0,fc=0,v1=0,v2=0; j<r->nf; j++){
        if(r->mat[j]==i){
		  if(mv>3) t->pnv[fc++]=(fv=r->pnv[j]); else fv=3;
		  for(k=0; k<fv; k++){
			          t->pfv[v2]=r->pfv[v1];
			if(r->nt) t->pft[v2]=r->pft[v1];
			v2++; v1++;
		  }
		} else {
          v1+=((mv>3)? r->pnv[j] : 3);
		}
	  }
      // delete isolated vertices
      DelIsoVert(t, TRUE);
	}
  }
}