// Virtual Modeler V0.65 for MS Visual C++ V6.0   04/13/2005  Michael Granz

#include "stdafx.h"

#include <string.h>

#include "gl\gl.h"
#include "gl\glu.h"

#include "general.h"
#include "Primitives.h"
#include "VModel.h"
#include "MainFrm.h"
#include "VModelFile.h"
#include "VModelDoc.h"
#include "VModelView.h"

#include "TextWin.h"
#include "GraphWin.h"

#include "RaVi.h"
#include "RaVi_Tree.h"

#ifdef _DEBUG
  #define new DEBUG_NEW
  #undef THIS_FILE
  static char THIS_FILE[] = __FILE__;
#endif

#define NZ 39

extern void BuildWorldQuader(PrimPtr p);
extern void UnifyNormals(PrimPtr p);
extern void OptimizeMesh(PolyMesh *p);
extern void DelIsoVert(PrimPtr p, BOOL stat);
extern void CreateGL (PrimPtr p);
extern void CreateGL1(PrimPtr p, int hl);
extern void DrawObj(PrimPtr p);
extern PrimPtr SelectObjects(char *Prompt);
extern PrimPtr SelectFaces(char *Prompt);
extern vector GetPoint(char *Prompt, PrimPtr Selection);
extern vector GetWPt(char *Prompt);

extern void OpenPhoto();
extern void OpenBackground();
extern void MatchCam1();
extern void MatchCam2();
extern void MatchVhd();
extern void GetTex3D(PrimPtr p);

extern void SplitConnected(PolyMesh *);
extern void SplitMaterials(PolyMesh *);

extern float wlcoff(float cR, float cG, float cB, float v);

extern vector  ParsePoint(char *a);

extern bool    pointcomplete;
extern vector *RV_CamPtr;

int    RV_Caustics_Samples=10000;
int    RV_Caustics_Quality=   30;
short  Demo;
QLst   QLT[16];

const float OB64 = 1.0f/64.0f;

float wavlen     = 555.0f;

OPENFILENAME  OFName;
ushort        OGLP          = 1;
ushort        OGLT          = 1;
char          filename[320] = "";
Primitive    *RootPtr       = NULL;
int           SPx,SPy, EPx,EPy;

Material      ENVIRONMENT;

inline float afract(float x){ double dummy; return (float)fabs(modf(x, &dummy)); }


void PushView(ViewPtr vp)
{
  ViewPtr V = new View; *V = *vp;
  V->next = VMDoc->ViewStack; VMDoc->ViewStack = V;
}


void PDel(PrimPtr p)
{
  if(p->idnt != RAY){
    if(glIsList(((Surface *)p)->OGLL)) glDeleteLists(((Surface *)p)->OGLL,1);
    if(((Surface *)p)->name     ) delete ((Surface *)p)->name;
  }
  switch(p->idnt){
    case POLYMESH: {
      if(((PolyMesh *)p)->pfv) delete ((PolyMesh *)p)->pfv;
      if(((PolyMesh *)p)->pfn) delete ((PolyMesh *)p)->pfn;
      if(((PolyMesh *)p)->pft) delete ((PolyMesh *)p)->pft;
      //if(((PolyMesh *)p)->hgt) delete ((PolyMesh *)p)->hgt;
      if(((PolyMesh *)p)->pts) delete ((PolyMesh *)p)->pts;
      if(((PolyMesh *)p)->nrm) delete ((PolyMesh *)p)->nrm;
      if(((PolyMesh *)p)->txc) delete ((PolyMesh *)p)->txc;
      if(((PolyMesh *)p)->mtl) delete ((PolyMesh *)p)->mtl;
      if(((PolyMesh *)p)->pnv) delete ((PolyMesh *)p)->pnv;
      if(((PolyMesh *)p)->flg) delete ((PolyMesh *)p)->flg;
      if(((PolyMesh *)p)->grp) delete ((PolyMesh *)p)->grp;
      if(((PolyMesh *)p)->mat) delete ((PolyMesh *)p)->mat;

      if(((PolyMesh *)p)->mesh.snm) delete ((PolyMesh *)p)->mesh.snm;
      if(((PolyMesh *)p)->mesh.lin) delete ((PolyMesh *)p)->mesh.lin;
      break;
    }
	case CYLINDER: {
      if(((Cylinder *)p)->txc) delete ((Cylinder *)p)->txc;
      break;
	}
  }
  delete (byte *)p;
}

PrimPtr Kill1(PrimPtr p)
{
  PrimPtr r = p->succ;
  if(p->pred) p->pred->succ = p->succ;
  if(p->succ) p->succ->pred = p->pred;
  PDel(p); VMDoc->WorldValid=FALSE;
  return r;
}

void Primitive::Kill()
{
  PrimPtr p2,p1 = this;
  while(p1){
    p2=p1->succ;
    if((p1->idnt) == ROOT) p1->succ=NULL; else PDel(p1);
    p1=p2;
  }
  VMDoc->WorldValid=FALSE;
}

extern vector LPos;

extern BSPplane *VM2RaVi(PrimPtr p); long int sSMem, sSObj;
	  
void InitRaVi()
{
  TextWin.Output(1,"- converting primitives ...");
  RV_SceneTree = VM2RaVi(VMDoc->VMF.Base);
  if(!RV_Lights){
    LitPtr l = new RV_Light;
	l->v     = 1.0f;
	l->next  = RV_Lights; RV_Lights=l;
	l->Prim  = NULL;
    l->p     = LPos;
  }
  RV_Init();
  TextWin.Output(1,"  - BSP-tree extended to %d levels.\n", MaxSplitLevel);
}

void SetCaption()
{
  char cap[250]="";
  if(filename) strcpy(cap,filename);
  if(GraphWin.GScreen.V.name && *(GraphWin.GScreen.V.name)){
    if(*cap) strcat(cap, " - ");
	strcat(cap, GraphWin.GScreen.V.name);
  }
  GraphWin.SetWindowText(cap);
}

void upd_usr_views()
{
  CMenu* pTopMenu = AfxGetMainWnd()->GetMenu();
  CMenu* pUserViewMenu = pTopMenu->GetSubMenu(2)->GetSubMenu(3);
  ASSERT(pUserViewMenu != NULL);
  ViewPtr V; int i;
  // delete all user view items
  for(i=pUserViewMenu->GetMenuItemCount()-1; i>6; i--) pUserViewMenu->DeleteMenu(i, MF_BYPOSITION);
  // then, add an item for each available user view
  for(i=0,V=VMDoc->VMF.ViewList; V; V=V->next,i++){
	pUserViewMenu->AppendMenu(MF_STRING, 1230+i, V->name);
  }
}

void transview(ViewPtr V2)
{
  View V1= GraphWin.GScreen.V;
  if((!(V2->h) && !(V1.h)) || (V2->h && V1.h)){
    DWORD ts=GetTickCount(), tc=ts, te = ts+750; float i;
    while(tc<te){
	  i=(float)(tc-ts)/(float)(te-ts);
      GraphWin.GScreen.V.Location=blend(i, V2->Location, V1.Location);
      GraphWin.GScreen.V.Target  =blend(i, V2->Target  , V1.Target  );
      GraphWin.GScreen.V.Zoom    =blend(i, V2->Zoom    , V1.Zoom    );
      GraphWin.GScreen.V.Up      =blend(i, V2->Up      , V1.Up      );
      GraphWin.GScreen.V.h       =blend(i, V2->h       , V1.h       );
	  GraphWin.InvalidateRect(NULL,FALSE); WaitTask(NULL); tc = GetTickCount();
    }
  }
  GraphWin.GScreen.V = *V2;
}

extern byte       *BMPbitsB;
extern BITMAPINFO  BMPinfoB;

float CalcDistVL2D(float x1, float y1, float x2, float y2, float xs, float ys, float *xl, float *yl)
{
  float dx=x2-x1; float sdx=sqr(dx);
  float dy=y2-y1; float sdy=sqr(dy);
  if(fabs(dx)<fabs(dy)){
    *yl=(dx*dy*(xs-x1)+y1*sdx+ys*sdy)/(sdy+sdx);
    //if((*yl<min(y1,y2)) || (*yl>max(y1,y2))) return min( sqr(xs-x1)+sqr(ys-y1) , sqr(xs-x2)+sqr(ys-y2) );
    *xl=x1+((*yl-y1)*dx)/dy;
  } else {
	if(dx==0.0f){ *xl=x1; *yl=ys; return sqr(xs-x1)+sqr(ys-y1); }
    *xl=(dx*dy*(ys-y1)+x1*sdy+xs*sdx)/(sdy+sdx);
    //if((*xl<min(x1,x2)) || (*xl>max(x1,x2))) return min( sqr(xs-x1)+sqr(ys-y1) , sqr(xs-x2)+sqr(ys-y2) );
    *yl=y1+((*xl-x1)*dy)/dx;
  }
  /*float dx=x2-x1; float dy=y2-y1; float m1=dy/dx; float m2=-1.0f/m1;
  float b1=y1-m1*x1;
  float b2=ys-m2*xs;
  *xl=(b1-b2)/(m2-m1);
  *yl=m2*(*xl)+b2;*/
  return sqr(xs-*xl)+sqr(ys-*yl);
}

void DrawVLink(CDC *hdc, vector pg, vector pp)
{
  GLdouble   x0,y0,z0;
  GLdouble   mm[16];
  GLdouble   mp[16];
  GLint      mv[16];
  glGetDoublev(GL_MODELVIEW_MATRIX,mm);
  glGetDoublev(GL_PROJECTION_MATRIX,mp);
  glGetIntegerv(GL_VIEWPORT,mv);
  gluProject(pg.x, pg.y, pg.z, mm, mp, mv, &x0, &y0, &z0);
  GraphWin.GetClientRect(&(GraphWin.m_oldRect));

  if(BMPbitsB){
	float fx=(float)GraphWin.m_oldRect.right /(float)BMPinfoB.bmiHeader.biWidth;
	float fy=(float)GraphWin.m_oldRect.bottom/(float)BMPinfoB.bmiHeader.biHeight;
	if(fy>fx){
	  pp.x = pp.x*fx; pp.y = pp.y*fx + (GraphWin.m_oldRect.bottom - fx*BMPinfoB.bmiHeader.biHeight)*0.5f;
	} else {
	  pp.x = pp.x*fy + (GraphWin.m_oldRect.right - fy*BMPinfoB.bmiHeader.biWidth)*0.5f; pp.y = pp.y*fy;
	}
  }

  hdc->Ellipse((int)(x0  +0.5f)-5, (int)(GraphWin.m_oldRect.bottom-y0-0.5f)-5, (int)(x0  +0.5f)+5, (int)(GraphWin.m_oldRect.bottom-y0-0.5f)+5);
  hdc->Ellipse((int)(pp.x+0.5f)-5, (int)(pp.y                        +0.5f)-5, (int)(pp.x+0.5f)+5, (int)(pp.y                        +0.5f)+5);
  hdc->MoveTo ((int)(x0  +0.5f)  , (int)(GraphWin.m_oldRect.bottom-y0-0.5f));
  hdc->LineTo ((int)(pp.x+0.5f)  , (int)(pp.y                        +0.5f));
}

void DrawELink(CDC *hdc, vector p1, vector p2, vector pp)
{
  float xl,yl;
  float x1,y1;
  float x2,y2;
  GLdouble   x0,y0,z0;
  GLdouble   mm[16];
  GLdouble   mp[16];
  GLint      mv[16];
  glGetDoublev(GL_MODELVIEW_MATRIX,mm);
  glGetDoublev(GL_PROJECTION_MATRIX,mp);
  glGetIntegerv(GL_VIEWPORT,mv);
  gluProject(p1.x, p1.y, p1.z, mm, mp, mv, &x0, &y0, &z0); x1=(float)x0; y1=(float)y0;
  gluProject(p2.x, p2.y, p2.z, mm, mp, mv, &x0, &y0, &z0); x2=(float)x0; y2=(float)y0;
  
  GraphWin.GetClientRect(&(GraphWin.m_oldRect));

  if(BMPbitsB){
	float fx=(float)GraphWin.m_oldRect.right /(float)BMPinfoB.bmiHeader.biWidth;
	float fy=(float)GraphWin.m_oldRect.bottom/(float)BMPinfoB.bmiHeader.biHeight;
	if(fy>fx){
	  pp.x = pp.x*fx; pp.y = pp.y*fx + (GraphWin.m_oldRect.bottom - fx*BMPinfoB.bmiHeader.biHeight)*0.5f;
	} else {
	  pp.x = pp.x*fy + (GraphWin.m_oldRect.right - fy*BMPinfoB.bmiHeader.biWidth)*0.5f; pp.y = pp.y*fy;
	}
  }

  CalcDistVL2D(x1,y1, x2,y2, pp.x, GraphWin.m_oldRect.bottom-pp.y, &xl,&yl);
  
  hdc->Ellipse((int)(xl  +0.5f)-5, (int)(GraphWin.m_oldRect.bottom-yl-0.5f)-5, (int)(xl  +0.5f)+5, (int)(GraphWin.m_oldRect.bottom-yl-0.5f)+5);
  hdc->Ellipse((int)(pp.x+0.5f)-5, (int)(pp.y                        +0.5f)-5, (int)(pp.x+0.5f)+5, (int)(pp.y                        +0.5f)+5);
  hdc->MoveTo ((int)(xl  +0.5f)  , (int)(GraphWin.m_oldRect.bottom-yl-0.5f));
  hdc->LineTo ((int)(pp.x+0.5f)  , (int)(pp.y                        +0.5f));
}

void MapFaces(PrimPtr S)
{
  PrimPtr    p;
  PolyMesh  *t;
  int        fi,i,j,k,n,cp=0,lop=1;
  vector     ca[32],ct[32];

  do{
    p=S; vector vt=GetPoint("- select vertex: ", p);
    if(!(*(TextWin.inputstring))) lop=0;
	if(lop){
	  TextWin.Input("- enter mapping coordinates: ");
	  if(!strnicmp(TextWin.inputstring,"CHAIN",5)){
	    do{
		  ca[cp++]=vt;
          vt=GetPoint("- select next vertex in chain: ", p);
		} while(*(TextWin.inputstring));
		float cl=0.0f, lg=0.0f;
        for(i=1; i<cp; i++) lg += distance(ca[i],ca[i-1]);
		TextWin.Output(1,"- chain length: %f\n",lg);
		TextWin.Input("- start mapping coordinates: ");
		ct[0]=ParsePoint(TextWin.inputstring);
		TextWin.Input("- end mapping coordinates: ");
		ct[cp-1]=ParsePoint(TextWin.inputstring);
        for(i=1; i<(cp-1); i++){
		  cl += distance(ca[i],ca[i-1]);
		  ct[i] = blend(cl/lg, ct[cp-1], ct[0]);
		}
	  } else {
	    ca[0]=vt; ct[0]=ParsePoint(TextWin.inputstring); cp=1;
	  }
      while(p){
        switch(p->idnt){
          case POLYMESH: {
            t=(PolyMesh *)p;
            if(!t->nt){
              t->txc=new(vector[cp]); for(i=0; i<cp; i++) t->txc[i]=ct[i]; t->nt=cp;
              for(fi=0,i=0; i<(t->nf); i++){ k=(t->pnv)? t->pnv[i] : 3; fi+=k; }
              t->pft=new(ushort[fi]);
              for(fi=0,i=0; i<(t->nf); i++){
	            k=(t->pnv)? t->pnv[i] : 3;
			      for(j=0; j<k; j++){
			        t->pft[fi+j]=0;
				    for(n=0; n<cp; n++){
				      if(distance(ca[n], t->pts[t->pfv[fi+j]])<0.0001f) t->pft[fi+j] = n;
					}
				  }
			    fi+=k;
			  }
			} else {
		      vector *otc = t->txc;
              t->txc=new(vector[t->nt+cp]);
		      for(i=0; i<(t->nt); i++) t->txc[i]=otc[i]; delete otc;
			  for(i=0; i<cp; i++) t->txc[t->nt+i]=ct[i];
              for(fi=0,i=0; i<(t->nf); i++){
			    k=(t->pnv)? t->pnv[i] : 3;
                if(t->flg[i] & SELECTED){
			      for(j=0; j<k; j++){
				    for(n=0; n<cp; n++){
				      if(distance(ca[n], t->pts[t->pfv[fi+j]])<0.0001f) t->pft[fi+j] = t->nt + n;
					}
				  }
				}
			    fi+=k;
			  }
			  t->nt += cp;
			}
		    break;
		  }
          case CYLINDER: {
            Cylinder *t=(Cylinder *)p;
			if(t->angle){
			  if(!(t->txc)){ t->txc=new(vector[4]); }
			  for(n=0; n<cp; n++){
			    if(distance(ca[n], vector(   1.0f                      ,                 0.0f         , 0.0f)*t->TM) < 0.0001f) t->txc[0] = ct[n];
			    if(distance(ca[n], vector(t->r_top                     ,                 0.0f         , 1.0f)*t->TM) < 0.0001f) t->txc[2] = ct[n];
			    if(distance(ca[n], vector(t->r_top*(float)cos(t->angle), t->r_top*(float)sin(t->angle), 1.0f)*t->TM) < 0.0001f) t->txc[3] = ct[n];
			    if(distance(ca[n], vector(         (float)cos(t->angle),          (float)sin(t->angle), 0.0f)*t->TM) < 0.0001f) t->txc[1] = ct[n];
			  }
			}
		    break;
		  }
          case RING: {
            Ring *t=(Ring *)p;
			if(t->angle){
			  if(!(t->txc)){ t->txc=new(vector[4]); }
			  for(n=0; n<cp; n++){
			    if(distance(ca[n], vector(   1.0f                    ,    0.0f                    , 0.0f)*t->TM) < 0.0001f) t->txc[2] = ct[n];
			    if(distance(ca[n], vector(t->r_i                     ,    0.0f                    , 0.0f)*t->TM) < 0.0001f) t->txc[0] = ct[n];
			    if(distance(ca[n], vector(t->r_i*(float)cos(t->angle), t->r_i*(float)sin(t->angle), 0.0f)*t->TM) < 0.0001f) t->txc[1] = ct[n];
			    if(distance(ca[n], vector(       (float)cos(t->angle),        (float)sin(t->angle), 0.0f)*t->TM) < 0.0001f) t->txc[3] = ct[n];
			  }
			}
		    break;
		  }
          case SPHERE: {
            Sphere *t=(Sphere *)p;
			if((t->vangle>0) && (t->vangle<6.2831)){
			  if(!(t->txc)){ t->txc=new(vector[4]); }
			  for(n=0; n<cp; n++){
			    if(distance(ca[n], vector((float)                cos(t->ustart) , (float)sin(t->ustart),         0.0f                          )*t->TM) < 0.0001f) t->txc[3] = ct[n];
			    if(distance(ca[n], vector((float)(cos(t->vangle)*cos(t->ustart)), (float)sin(t->ustart), (float)(sin(t->vangle)*cos(t->ustart)))*t->TM) < 0.0001f) t->txc[1] = ct[n];
			    if(distance(ca[n], vector((float)                cos(t->ustop ) , (float)sin(t->ustop ),         0.0f                          )*t->TM) < 0.0001f) t->txc[2] = ct[n];
			    if(distance(ca[n], vector((float)(cos(t->vangle)*cos(t->ustop )), (float)sin(t->ustop ), (float)(sin(t->vangle)*cos(t->ustop )))*t->TM) < 0.0001f) t->txc[0] = ct[n];
			  }
			}
		    break;
		  }
          case TORUS: {
            Torus *t=(Torus *)p;
			if((t->vangle>0) && (t->vangle<6.2831)){
			  if(!(t->txc)){ t->txc=new(vector[4]); }
			  for(n=0; n<cp; n++){
			    if(distance(ca[n], vector((t->R1 + t->R2 * cosf(t->ustart))                  ,                  0.0f                              , t->R2 * sinf(t->ustart))*t->TM) < 0.0001f) t->txc[0] = ct[n];
			    if(distance(ca[n], vector((t->R1 + t->R2 * cosf(t->ustart)) * cosf(t->vangle), (t->R1 + t->R2 * cosf(t->ustart)) * sinf(t->vangle), t->R2 * sinf(t->ustart))*t->TM) < 0.0001f) t->txc[1] = ct[n];
			    if(distance(ca[n], vector((t->R1 + t->R2 * cosf(t->ustop ))                  ,                  0.0f                              , t->R2 * sinf(t->ustop ))*t->TM) < 0.0001f) t->txc[2] = ct[n];
			    if(distance(ca[n], vector((t->R1 + t->R2 * cosf(t->ustop )) * cosf(t->vangle), (t->R1 + t->R2 * cosf(t->ustop )) * sinf(t->vangle), t->R2 * sinf(t->ustop ))*t->TM) < 0.0001f) t->txc[3] = ct[n];
			  }
			}
		    break;
		  }
		}
        ((SurfPtr)p)->dflags &= ~1;
		if(((SurfPtr)p)->mesh.pts) delete ((SurfPtr)p)->mesh.pts; ((SurfPtr)p)->mesh.pts=NULL;

        p=p->succ;
      }
	}
  } while(lop);
}


char *SkipS(char *ip) { while(*ip==' ') ip++; return ip; }

BOOL GetFloat(float *f, char *s) { *f=(float)atof(s); return TRUE; }


/////////////////////////////////////////////////////////////////////////////
// CVModelApp

BEGIN_MESSAGE_MAP(CVModelApp, CWinApp)
	//{{AFX_MSG_MAP(CVModelApp)
	ON_COMMAND(ID_APP_ABOUT, OnAppAbout)
	ON_COMMAND(ID_VIEW_DOLLY, OnViewDolly)
	ON_COMMAND(ID_VIEW_ROTATE, OnViewRotate)
	ON_COMMAND(ID_VIEW_LOOKAROUND, OnViewLookaround)
	ON_COMMAND(ID_VIEW_WALK, OnViewWalk)
	ON_COMMAND(ID_DISPLAY_SHADED, OnDisplayShaded)
	ON_UPDATE_COMMAND_UI(ID_DISPLAY_SHADED, OnUpdateDisplayShaded)
	ON_COMMAND(ID_MODIFY_OBJECT_ERASE, OnModifyObjectErase)
	ON_COMMAND(ID_VIEW_ZOOM_ANGLE, OnViewZoomAngle)
	ON_COMMAND(ID_VIEW_RESTORE_TOP, OnViewRestoreTop)
	ON_COMMAND(ID_VIEW_RESTORE_FRONT, OnViewRestoreFront)
	ON_COMMAND(ID_VIEW_RESTORE_RIGHT, OnViewRestoreRight)
	ON_COMMAND(ID_VIEW_ZOOM_DYNAMIC, OnViewZoomDynamic)
	ON_COMMAND(ID_VIEW_PAN, OnViewPan)
	ON_COMMAND(ID_DISPLAY_PERSPECTIVE, OnDisplayPerspective)
	ON_UPDATE_COMMAND_UI(ID_DISPLAY_PERSPECTIVE, OnUpdateDisplayPerspective)
	ON_COMMAND(ID_DISPLAY_SHOW_INVERSE, OnDisplayShowInverse)
	ON_COMMAND(ID_DISPLAY_SHOW_NOT, OnDisplayShowNot)
	ON_COMMAND(ID_DISPLAY_SHOW_ONLY, OnDisplayShowOnly)
	ON_COMMAND(ID_DISPLAY_SHOW_ALL, OnDisplayShowAll)
	ON_COMMAND(ID_MODIFY_OBJECT_DOUBLESIDED, OnModifyObjectDoublesided)
	ON_COMMAND(ID_MODIFY_OBJECT_SINGLESIDED, OnModifyObjectSinglesided)
	ON_COMMAND(ID_MODIFY_OBJECT_FACETTED, OnModifyObjectFacetted)
	ON_COMMAND(ID_MODIFY_OBJECT_SMOOTH, OnModifyObjectSmooth)
	ON_COMMAND(ID_MODIFY_OBJECT_MERGE, OnModifyObjectMerge)
	ON_COMMAND(ID_MODIFY_FACE_SEP, OnModifyFaceSep)
	ON_COMMAND(ID_VIEW_PREVIOUS, OnViewPrevious)
	ON_UPDATE_COMMAND_UI(ID_VIEW_PREVIOUS, OnUpdateViewPrevious)
	ON_COMMAND(ID_MODIFY_FACE_DELETE, OnModifyFaceDelete)
	ON_COMMAND(ID_MODIFY_FACE_SMOOTHING_GROUP1, OnModifyFaceSmoothingGroup1)
	ON_COMMAND(ID_MODIFY_FACE_SMOOTHING_GROUP2, OnModifyFaceSmoothingGroup2)
	ON_COMMAND(ID_MODIFY_FACE_SMOOTHING_GROUP3, OnModifyFaceSmoothingGroup3)
	ON_COMMAND(ID_MODIFY_FACE_SMOOTHING_GROUP4, OnModifyFaceSmoothingGroup4)
	ON_COMMAND(ID_MODIFY_FACE_SMOOTHING_NONE, OnModifyFaceSmoothingNone)
	ON_COMMAND(ID_FILE_IMPORT_3DSTUDIO, OnImport3DS)
	ON_COMMAND(ID_MODIFY_OBJECT_NORMALS_FLIP, OnModObjNormFlip)
	ON_COMMAND(ID_FILE_EXPORT_POVRAY, OnExportPOV)
	ON_COMMAND(ID_MODIFY_OBJECT_NORMALS_UNIFY, OnNormalsUnify)
	ON_COMMAND(ID_MODIFY_OBJECT_OPTIMIZE, OnOptimizeMesh)
	ON_COMMAND(ID_FILE_IMPORT_VRML, OnImportWRL)
	ON_COMMAND(ID_MODIFY_FACE_FLIP, OnFaceFlip)
	ON_COMMAND(ID_SPLIT_CONNECTED, OnSplitConnected)
	ON_COMMAND(ID_FILE_IMPORT_LIGHTWAVE, OnImportLWO)
	ON_COMMAND(ID_FILE_EXPORT_AUTOCADDXF, OnExportDXF)
	ON_COMMAND(ID_FILE_IMPORT_AUTOCADDXF, OnImportDXF)
	ON_COMMAND(ID_FILE_IMPORT_WAVEFRONT, OnImportObj)
	ON_COMMAND(ID_VIEW_AUTO, OnViewAuto)
	ON_COMMAND(ID_DISPLAY_TEXTURES, OnDisplayTextures)
	ON_UPDATE_COMMAND_UI(ID_DISPLAY_TEXTURES, OnUpdateDisplayTextures)
	ON_COMMAND(ID_RAYTRACE, OnRaytrace)
	ON_COMMAND(ID_PHOTO_GETTEX3D, OnPhGetTex3D)
	ON_COMMAND(ID_FILE_INFO, OnFileInfo)
	ON_COMMAND(ID_MODIFY_OBJECT_MOVE, OnModifyObjectMove)
	ON_COMMAND(ID_MEASURE_COORDINATES, OnMeasureCoords)
	ON_COMMAND(ID_PHOTO_OPEN_EXTERNAL, OnPhOpenExt)
	ON_COMMAND(ID_PHOTO_OPEN_BACKDROP, OnPhOpenBack)
	ON_COMMAND(ID_PHOTO_LINK, OnPhLink)
	ON_COMMAND(ID_VIEW_SAVE, OnViewSave)
	ON_COMMAND(ID_FACE_MAP, OnFaceMap)
	ON_UPDATE_COMMAND_UI(ID_DISPLAY_AA, OnUpdateDisplayAa)
	ON_COMMAND(ID_DISPLAY_AA, OnDisplayAa)
	ON_COMMAND(ID_MODIFY_OBJECT_UNMAP, OnModifyObjectUnmap)
	ON_COMMAND(ID_FILE_EXPORT_RIB, OnExportRib)
	ON_COMMAND(ID_FILE_EXPORT_VRML, OnExportVRML)
	ON_COMMAND(ID_FILE_EXPORT_VL, OnExportVL)
	ON_COMMAND(ID_MODIFY_OBJECT_ROTATE_ZAXIS, OnRotateZ)
	ON_COMMAND(ID_MODIFY_OBJECT_ROTATE_YAXIS, OnRotateY)
	ON_COMMAND(ID_MODIFY_OBJECT_ROTATE_XAXIS, OnRotateX)
	ON_COMMAND(ID_CREATE_POLYGON, OnCreatePolygon)
	ON_COMMAND(ID_CREATE_CYLINDER, OnCreateCylinder)
	ON_COMMAND(ID_VIEW_RESTORE_LEFT, OnViewRestoreLeft)
	ON_COMMAND(ID_PHOTO_MATCH1, OnPhMatch1)
	ON_COMMAND(ID_PHOTO_MATCH_VHD, OnPhMatchVhd)
	ON_COMMAND(ID_PHOTO_MATCH2, OnPhotoMatch2)
	ON_COMMAND(ID_PHOTO_LINK2, OnPhLinkEdge)
	ON_COMMAND(ID_FILE_IMPORT_DEM, OnImportDEM)
	ON_COMMAND(ID_SPLIT_BYMATERIALS, OnSplitMaterials)
	ON_COMMAND(ID_FILE_IMPORT_VIDEOSCAPE, OnImportGEO)
	//}}AFX_MSG_MAP
	// Standard file based document commands
	ON_COMMAND(ID_FILE_NEW, CWinApp::OnFileNew)
	ON_COMMAND(ID_FILE_OPEN, CWinApp::OnFileOpen)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// The one and only CVModelApp object

CVModelApp  VMApp;
HICON       AppIcon = VMApp.LoadStandardIcon("IDI_APPLICATION");


CVModelApp::CVModelApp()
{
  // TODO: add construction code here,
  // Place all significant initialization in InitInstance
}

extern int go; extern int crs; extern DWORD gs;

BOOL CVModelApp::InitInstance()
{
  // Standard initialization
  // If you are not using these features and wish to reduce the size
  // of your final executable, you should remove from the following
  // the specific initialization routines you do not need.

  char       *ip;

  #ifdef _AFXDLL
    Enable3dControls();		// using MFC in a shared DLL
  #else
    Enable3dControlsStatic();	// linking to MFC statically
  #endif

  // Load standard INI file options (including MRU)
  LoadStdProfileSettings();

  // Register the application's document templates. Document templates
  // serve as the connection between documents, frame windows and views.

  CSingleDocTemplate* pDocTemplate;
  pDocTemplate = new CSingleDocTemplate(
                       IDR_MAINFRAME,
                       RUNTIME_CLASS(CVModelDoc),
                       RUNTIME_CLASS(CMainFrame),
                       RUNTIME_CLASS(CVModelView)
                     );
  AddDocTemplate(pDocTemplate);

  EnableShellOpen();  // Enable DDE Execute open

  RegisterShellFileTypes(TRUE);

  // Parse command line for standard shell commands, DDE, file open
  CCommandLineInfo cmdInfo;
  ParseCommandLine(cmdInfo);

  // Dispatch commands specified on the command line
  if (!ProcessShellCommand(cmdInfo)) return FALSE;

  // Enable drag/drop open
  m_pMainWnd->DragAcceptFiles();

  OFName.lStructSize = sizeof(OPENFILENAME);
  OFName.hwndOwner = NULL;
  OFName.lpstrCustomFilter = NULL;
  OFName.nFilterIndex = 1;
  OFName.lpstrFile = filename;
  OFName.nMaxFile = sizeof(filename);
  OFName.lpstrFileTitle = NULL;
  OFName.lpstrInitialDir = NULL;
  OFName.lpstrTitle = NULL;
  OFName.Flags = 0;
  OFName.nFileOffset = 0;
  OFName.nFileExtension = 0;
  OFName.lpfnHook = NULL;
  
  cframe=0;

  Delay(500); // wait while initializing windows

  TextWin.Output(1,"***  MG's Virtual Modeler V0.65  ***\n");
  TextWin.Output(3,"     Michael Granz   04/13/2005\n");

  do{
    TextWin.Input(">");
    if(!strnicmp(TextWin.inputstring,"MEM",3)){
	  TextWin.Output(1,"-free memory space: %lu bytes.\n",GetFreeSpace(0));
    } else if(!strnicmp(TextWin.inputstring,"INFO",4)){
      long int oc=0, vc=0, fc=0;
	  for(PrimPtr p = VMDoc->VMF.Base; p; p=p->succ){ p->CountFV(vc,fc); oc++; }
	  TextWin.Output(1,"- %d vertices on %d faces in %d objects.\n",vc,fc,oc);
    } else if(!strnicmp(TextWin.inputstring,"NEW",3)){
      VMDoc->KillDoc(); RV_SceneTree=NULL;
	  VMDoc->InitDoc(); upd_usr_views();
      GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"LOAD",4)){
      ip=TextWin.inputstring; ip=SkipS(ip+4);
      if(!(*ip)){
        OFName.lpstrFilter = "Virtual Models  (*.VMD)\000*.vmd\000All Files  (*.*)\000*.*\000\000";
        OFName.lpstrDefExt = "VMD";
        if(GetOpenFileName(&OFName)){
          TextWin.Complete(filename);
  	      VMDoc->KillDoc(); RV_SceneTree=NULL;
	      VMDoc->InitDoc();
		  VMDoc->Load(filename, &(GraphWin.GScreen.V));
		  AddToRecentFileList(filename); SetCaption();
          CreateGL(VMDoc->VMF.Base);
		  BuildWorldQuader(VMDoc->VMF.Base);
		  GraphWin.InvalidateRect(NULL,FALSE); upd_usr_views();
	    } else TextWin.Complete("<nothing>");
	  } else {
	    VMDoc->KillDoc(); RV_SceneTree=NULL;
	    VMDoc->InitDoc();
		VMDoc->Load(ip, &(GraphWin.GScreen.V));
 	    AddToRecentFileList(ip); strcpy(filename,ip); SetCaption();
        CreateGL(VMDoc->VMF.Base);
	    BuildWorldQuader(VMDoc->VMF.Base);
	    GraphWin.InvalidateRect(NULL,FALSE); upd_usr_views();
      }
    } else if(!strnicmp(TextWin.inputstring,"IMPORT",6)){
      ip=TextWin.inputstring; ip=SkipS(ip+6);
      if(!strnicmp(ip,"3DS",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "3D-Studio Mesh (*.3DS)\000*.3ds\0003D-Studio Project (*.PRJ)\000*.prj\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "3DS";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->Import3DS(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->Import3DS(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"4DS",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "Mafia Object (*.4DS)\000*.4ds\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "4DS";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->Import4DS(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->Import4DS(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"GEO",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "Videoscape Scene (*.GEO)\000*.geo\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "GEO";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportGEO(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportGEO(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"LWO",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "Lighwave Object (*.LWO)\000*.lwo\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "LWO";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportLWO(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportLWO(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"DXF",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "AutoCAD DXF (*.DXF)\000*.dxf\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "DXF";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportDXF(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportDXF(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"DEM",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "*";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportDEM(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportDEM(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"OBJ",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "Wavefront Object (*.OBJ)\000*.obj\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "OBJ";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportOBJ(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportOBJ(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"VRML",4)){
        ip=SkipS(ip+4);
        if(!(*ip)){
          OFName.lpstrFilter = "VRML  (*.WRL)\000*.wrl\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "WRL";
          if(GetOpenFileName(&OFName)){
            TextWin.Complete(filename);
  	        VMDoc->KillDoc();
	        VMDoc->InitDoc();
		    VMDoc->ImportVRML(filename, &(GraphWin.GScreen.V));
		    *filename=0; SetCaption();
            CreateGL(VMDoc->VMF.Base);
		    BuildWorldQuader(VMDoc->VMF.Base);
		    GraphWin.InvalidateRect(NULL,FALSE);
	      } else TextWin.Complete("<nothing>");
	    } else {
	      VMDoc->KillDoc();
	      VMDoc->InitDoc();
		  VMDoc->ImportVRML(ip, &(GraphWin.GScreen.V));
 	      *filename=0; SetCaption();
          CreateGL(VMDoc->VMF.Base);
	      BuildWorldQuader(VMDoc->VMF.Base);
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  }
    } else if(!strnicmp(TextWin.inputstring,"EXPORT",6)){
      ip=TextWin.inputstring; ip=SkipS(ip+6);
      if(!strnicmp(ip,"POV",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "POV-Ray 3.0  (*.POV)\000*.pov\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "POV";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->ExportPOV(filename, &(GraphWin.GScreen.V));
	      } else TextWin.Complete("<nothing>");
	    } else {
		  VMDoc->ExportPOV(ip, &(GraphWin.GScreen.V));
        }
	  } else if(!strnicmp(ip,"DXF",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "AutoCAD DXF  (*.DXF)\000*.dxf\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "DXF";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->ExportDXF(filename, &(GraphWin.GScreen.V));
	      } else TextWin.Complete("<nothing>");
	    } else {
		  VMDoc->ExportDXF(ip, &(GraphWin.GScreen.V));
        }
	  } else if(!strnicmp(ip,"RIB",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "Renderman RIB  (*.RIB)\000*.rib\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "RIB";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->ExportRIB(filename, &(GraphWin.GScreen.V));
	      } else TextWin.Complete("<nothing>");
	    } else {
		  VMDoc->ExportRIB(ip, &(GraphWin.GScreen.V));
        }
	  } else if(!strnicmp(ip,"VRML",4)){
        ip=SkipS(ip+4);
        if(!(*ip)){
          OFName.lpstrFilter = "VRML WRL  (*.WRL)\000*.wrl\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "wrl";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->ExportVRML(filename, &(GraphWin.GScreen.V));
	      } else TextWin.Complete("<nothing>");
	    } else {
		  VMDoc->ExportVRML(ip, &(GraphWin.GScreen.V));
        }
	  } else if(!strnicmp(ip,"VIB",3)){
        ip=SkipS(ip+3);
        if(!(*ip)){
          OFName.lpstrFilter = "VirtuaLight (*.VIB)\000*.vib\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "vib";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->ExportVIB(filename, &(GraphWin.GScreen.V));
	      } else TextWin.Complete("<nothing>");
	    } else {
		  VMDoc->ExportVIB(ip, &(GraphWin.GScreen.V));
        }
	  }
    } else if(!strnicmp(TextWin.inputstring,"SAVEAS",6)){
      ip=TextWin.inputstring; ip=SkipS(ip+6);
      if(!(*ip)){
        OFName.lpstrFilter = "Virtual Models  (*.VMD)\000*.vmd\000All Files  (*.*)\000*.*\000\000";
        OFName.lpstrDefExt = "VMD";
        if(GetSaveFileName(&OFName)){
          TextWin.Complete(filename);
		  VMDoc->Save(filename, &(GraphWin.GScreen.V));
		  AddToRecentFileList(filename); SetCaption();
	    } else TextWin.Complete("<nothing>");
	  } else {
		VMDoc->Save(ip, &(GraphWin.GScreen.V));
 	    AddToRecentFileList(ip); strcpy(filename,ip); SetCaption();
      }
    } else if(!strnicmp(TextWin.inputstring,"SAVE",4)){
      ip=TextWin.inputstring; ip=SkipS(ip+4);
      if(!(*ip)){
		if(*filename){
		  VMDoc->Save(filename, &(GraphWin.GScreen.V));
		} else {
          OFName.lpstrFilter = "Virtual Models  (*.VMD)\000*.vmd\000All Files  (*.*)\000*.*\000\000";
          OFName.lpstrDefExt = "VMD";
          if(GetSaveFileName(&OFName)){
            TextWin.Complete(filename);
		    VMDoc->Save(filename, &(GraphWin.GScreen.V));
		    AddToRecentFileList(filename); SetCaption();
	      } else TextWin.Complete("<nothing>");
		}
	  } else {
		VMDoc->Save(ip, &(GraphWin.GScreen.V));
 	    AddToRecentFileList(ip); strcpy(filename,ip); SetCaption();
      }
    } else if(!strnicmp(TextWin.inputstring,"SIZE?",5)){
	  RECT cldim;
      GraphWin.GetClientRect(&cldim);
      TextWin.Output(1,"-client rectangle: %d x %d pixel.\n", cldim.right+1, cldim.bottom+1);
    } else if(!strnicmp(TextWin.inputstring,"wrgb",4)){
      CDC      *hdc;
      hdc = GraphWin.GetDC();
	  for(float w=300.0f; w<800.0f; w++){
		float r = wlcoff(1.0f, 0.0f, 0.0f, w);
		float g = wlcoff(0.0f, 1.0f, 0.0f, w);
		float b = wlcoff(0.0f, 0.0f, 1.0f, w);
		float m = wlcoff(0.5f, 0.5f, 0.5f, w);
		hdc->SetPixel((int)w-300, 150-(int)(100.0f*m)+3,0x00808080);
		hdc->SetPixel((int)w-300, 150-(int)(100.0f*b)+2,0x00FF0000);
		hdc->SetPixel((int)w-300, 150-(int)(100.0f*g)+1,0x0000FF00);
		hdc->SetPixel((int)w-300, 150-(int)(100.0f*r)  ,0x000000FF);
		for(int y=0; y<16; y++){
		  hdc->SetPixel((int)w-300, 200-y,((int)(255.0f*b)<<16) | ((int)(255.0f*g)<<8) | (int)(255.0f*r) );
		}
	  }
      GraphWin.ReleaseDC(hdc);
    } else if(!strnicmp(TextWin.inputstring,"DOLLY",5)){
      PushView(&(GraphWin.GScreen.V));
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_DOLLY;
	  while(GraphWin.InputMode == IM_DRAG) Delay(200);
    } else if(!strnicmp(TextWin.inputstring,"PAN",3)){
      PushView(&(GraphWin.GScreen.V));
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_PAN;
	  while(GraphWin.InputMode == IM_DRAG) Delay(200);
    } else if(!strnicmp(TextWin.inputstring,"ZOOM",4)){
      PushView(&(GraphWin.GScreen.V));
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_ZOOM;
	  while(GraphWin.InputMode == IM_DRAG) Delay(200);
    } else if(!strnicmp(TextWin.inputstring,"VROTATE",7)){
      PushView(&(GraphWin.GScreen.V));
 	  GraphWin.GScreen.V.name = NULL; SetCaption();
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_VROTATE;
	  while(GraphWin.InputMode == IM_DRAG) Delay(200);
    } else if(!strnicmp(TextWin.inputstring,"AUTOROTATE",10)){
      PushView(&(GraphWin.GScreen.V));
 	  GraphWin.GScreen.V.name = NULL; SetCaption();
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_VAUTO;
	  float d=(float)sqrt(sqr(GraphWin.GScreen.V.Location.x-GraphWin.GScreen.V.Target.x) + sqr(GraphWin.GScreen.V.Location.y-GraphWin.GScreen.V.Target.y));
      float w=(float)(((fabs(GraphWin.GScreen.V.Location.x-GraphWin.GScreen.V.Target.x)>0.001)? atan((GraphWin.GScreen.V.Location.y-GraphWin.GScreen.V.Target.y)/(GraphWin.GScreen.V.Location.x-GraphWin.GScreen.V.Target.x)) : Pi_2) + (((GraphWin.GScreen.V.Location.x-GraphWin.GScreen.V.Target.x)<0.0)? Pi : 0.0)); if(w<0.0f) w+=TwoPi;
      DWORD to=0,tc=0,ts = GetTickCount();
	  float pm[16]; int fc=0;
	  while(GraphWin.InputMode == IM_DRAG){
        GraphWin.GScreen.V.Location.x = GraphWin.GScreen.V.Target.x+d*(float)cos(w+0.001f*tc);
        GraphWin.GScreen.V.Location.y = GraphWin.GScreen.V.Target.y+d*(float)sin(w+0.001f*tc);
		// calculate new orthogonal up-vector
		glMatrixMode(GL_PROJECTION); glPushMatrix(); glLoadIdentity();
        gluLookAt(GraphWin.GScreen.V.Location.x, GraphWin.GScreen.V.Location.y, GraphWin.GScreen.V.Location.z,
                  GraphWin.GScreen.V.Target.x,   GraphWin.GScreen.V.Target.y,   GraphWin.GScreen.V.Target.z,
  	              0.0,                           0.0,                           0.1 );
        glGetFloatv(GL_PROJECTION_MATRIX, pm);
        GraphWin.GScreen.V.Up.x = pm[1]; GraphWin.GScreen.V.Up.y = pm[5]; GraphWin.GScreen.V.Up.z = pm[9];
        glPopMatrix(); glMatrixMode(GL_MODELVIEW);
		GraphWin.InvalidateRect(NULL,FALSE); WaitTask(NULL);
		tc=GetTickCount()-ts; fc++;
		if((tc>0) && (tc>(to+1000))){
		  TextWin.Output(2,"- rotating...  (%4.1f fps)",(float)fc/(0.001f*(float)tc)); TextWin.px=0; TextWin.cx=0;
		  to=tc;
		}
	  }
	  TextWin.Output(2,"\n");
    } else if(!strnicmp(TextWin.inputstring,"VTURN",5)){
      PushView(&(GraphWin.GScreen.V));
 	  GraphWin.GScreen.V.name = NULL; SetCaption();
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_VTURN;
	  while(GraphWin.InputMode == IM_DRAG) Delay(200);
    } else if(!strnicmp(TextWin.inputstring,"VWALK",5)){
      PushView(&(GraphWin.GScreen.V));
 	  GraphWin.GScreen.V.name = NULL; SetCaption();
	  GraphWin.InputMode = IM_DRAG; GraphWin.DragMode = DM_VWALK; crs=1; GraphWin.SetCapture();
	  while(GraphWin.InputMode == IM_DRAG){
		if(gs>0){
		  DWORD t = GetTickCount() - gs; vector N;
		  if(go>0){
			N = unit(GraphWin.GScreen.V.Target - GraphWin.GScreen.V.Location);
		  }
		  if(go<0){
			N = -unit(GraphWin.GScreen.V.Target - GraphWin.GScreen.V.Location);
		  }
		  N=unit(vector(N.x, N.y, 0.0f));
          GraphWin.GScreen.V.Location = GraphWin.GScreen.V.Location + N*0.001f*(float)t;
          GraphWin.GScreen.V.Target   = GraphWin.GScreen.V.Target   + N*0.001f*(float)t;
          GraphWin.InvalidateRect(NULL,FALSE);
		  gs = abs(go)>1? GetTickCount() : 0;
		}
		WaitTask(NULL);
	  }
    } else if(!strnicmp(TextWin.inputstring,"VSHADE",6)){
	  GraphWin.DisplayMode ^= VM_SHADED;
	  GraphWin.DisplayMode |= VM_CHANGED;
	  GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"VTEXTURES",9)){
	  GraphWin.DisplayMode ^= VM_TEXTURED;
	  GraphWin.DisplayMode |= VM_REBUILD;
	  GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"VANTIALIAS",10)){
	  GraphWin.DisplayMode ^= VM_ANTIALIAS;
	  GraphWin.DisplayMode |= VM_REBUILD;
	  GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"VANGLE",6)){
      float a;
      ip=TextWin.inputstring; ip=SkipS(ip+6);
      if(!strnicmp(ip,"Previous",8)){
        GraphWin.GScreen.V.h = GraphWin.GScreen.V.old_h;
		GraphWin.InvalidateRect(NULL,FALSE);
	  } else if(GetFloat(&a,ip)){
        PushView(&(GraphWin.GScreen.V));
		if(!a){
		  GraphWin.GScreen.V.Zoom = VMDoc->WRD;
          if(GraphWin.GScreen.V.h) GraphWin.GScreen.V.old_h = GraphWin.GScreen.V.h;
		}
        GraphWin.GScreen.V.SetViewAngle(a*Deg2Rad);
		GraphWin.InvalidateRect(NULL,FALSE);
      } else {
        TextWin.Output(2,"- invalid viewing-angle '%s'!\n",ip);
      }
    } else if(!strnicmp(TextWin.inputstring,"VPREVIOUS",9)){
      ViewPtr V = VMDoc->ViewStack;
	  transview(VMDoc->ViewStack);
      VMDoc->ViewStack = V->next; delete(V);
 	  SetCaption(); GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"VSAVE",5)){
      ip=TextWin.inputstring; ip=SkipS(ip+5);
      GraphWin.GScreen.V.name = new char[strlen(ip)+1]; strcpy(GraphWin.GScreen.V.name, ip);
      ViewPtr V;
	  for(V=VMDoc->VMF.ViewList; V && strcmp(ip, V->name); V=V->next);
	  if(!V){
		V = new View; *V=GraphWin.GScreen.V; V->next=VMDoc->VMF.ViewList; VMDoc->VMF.ViewList=V;
	  } else *V=GraphWin.GScreen.V;
	  SetCaption(); upd_usr_views();
    } else if(!strnicmp(TextWin.inputstring,"VRESTORE",8)){
      ip=TextWin.inputstring; ip=SkipS(ip+8);
      PushView(&(GraphWin.GScreen.V));
      ViewPtr V; View V2; V2.Target = VMDoc->WMP; V2.Zoom = VMDoc->WRD; V2.h = 0.0f;
      for(V=VMDoc->VMF.ViewList; V && strcmp(ip,V->name); V=V->next);
	  if(V){
		V2=*V;
	  } else {
        if(GraphWin.GScreen.V.h) V2.old_h = GraphWin.GScreen.V.h;
        if(!strnicmp(ip,"Top",3)){
		  V2.Location = V2.Target + vector(0.0f, 0.0f,  4.0f*VMDoc->WRD);
		  V2.Up       = vector(0.0f, -1.0f, 0.0f);
		  V2.name     = "Top-View";
        } else if(!strnicmp(ip,"Front",5)){
		  V2.Location = V2.Target + vector(0.0f,  4.0f*VMDoc->WRD, 0.0f);
		  V2.Up       = vector(0.0f, 0.0f, 1.0f);
		  V2.name     = "Front-View";
        } else if(!strnicmp(ip,"Left",4)){
		  V2.Location = V2.Target + vector( 4.0f*VMDoc->WRD, 0.0f, 0.0f);
		  V2.Up       = vector(0.0f, 0.0f, 1.0f);
		  V2.name     = "Left-View";
        } else if(!strnicmp(ip,"Right",5)){
		  V2.Location = V2.Target + vector(-4.0f*VMDoc->WRD, 0.0f, 0.0f);
		  V2.Up       = vector(0.0f, 0.0f, 1.0f);
		  V2.name     = "Right-View";
        } else {
          TextWin.Output(2,"- view '%s' not found.\n",ip);
        }
	  }
  	  transview(&V2); SetCaption();
	  GraphWin.InvalidateRect(NULL,FALSE);
    } else if(!strnicmp(TextWin.inputstring,"SHOW",4)){
      ip=TextWin.inputstring; ip=SkipS(ip+4);
      if(!strnicmp(ip,"INVERSE",7)){
        PrimPtr Temp  = VMDoc->VMF.Base;
        VMDoc->VMF.Base   = VMDoc->VMF.Hidden;
        VMDoc->VMF.Hidden = Temp;
        GraphWin.InvalidateRect(NULL,TRUE);
      } else if(!strnicmp(ip,"ALL",3)){
        if(VMDoc->VMF.Hidden){
		  if(VMDoc->VMF.Base){
            PrimPtr P=VMDoc->VMF.Base; while(P->succ) P=P->succ;
            P->succ = VMDoc->VMF.Hidden;
            VMDoc->VMF.Hidden->pred = P;
		  } else {
			VMDoc->VMF.Base = VMDoc->VMF.Hidden;
		  }
          VMDoc->VMF.Hidden = NULL;
          GraphWin.InvalidateRect(NULL,TRUE);
        }
      } else if(!strnicmp(ip,"ONLY",4)){
        PrimPtr Selection=SelectObjects("- select objects to show alone:\n");
        PrimPtr P=VMDoc->VMF.Hidden;
		if(P){
		  while(P->succ) P=P->succ;
		  P->succ = VMDoc->VMF.Base;
		  if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=P;
		} else {
		  VMDoc->VMF.Hidden=VMDoc->VMF.Base;
		}
		VMDoc->VMF.Base=Selection;
        GraphWin.InvalidateRect(NULL,TRUE);
      } else {
        PrimPtr Selection=SelectObjects("- select objects to hide:\n");
        if(Selection){
          PrimPtr P=VMDoc->VMF.Hidden;
		  if(P){
			while(P->succ) P=P->succ;
            P->succ = Selection;
            Selection->pred = P;
		  } else {
			VMDoc->VMF.Hidden=Selection;
		  }
        }
        GraphWin.InvalidateRect(NULL,TRUE);
      }
    } else if(!strnicmp(TextWin.inputstring,"POLYGON",7)){
	  vector p[32]; int i,n=0;
	  do{
		char prompt[80]; sprintf(prompt,"- vertex %d ([RETURN] to %s): ",n+1,(n>2? "close" : "abort"));
        p[n]=GetPoint(prompt, NULL); n++;
	  } while(pointcomplete); n--;
	  if(n>=3){
        PolyMesh * pm = new PolyMesh; pm->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=pm; VMDoc->VMF.Base=pm;
        pm->name = strdup("Polygon"); pm->flags=SURF_DOUBLESIDED;
        pm->nf = 1; pm->pfv=new(ushort[n*pm->nf]);
        if(n>3) { pm->pnv=new(byte[pm->nf]); pm->pnv[0]=n; }
        pm->nv=n; pm->pts=new vector[pm->nv]; for(i=0; i<n; i++) { pm->pts[i]=p[i]; pm->pfv[i]=i; }
	    if(!(VMDoc->VMF.MatList)){
	      VMDoc->VMF.MatList=new Material; VMDoc->VMF.MatList->name = strdup("default");
		}
        pm->Mat=VMDoc->VMF.MatList;
        if(pm->succ){
	      glDrawBuffer(GL_FRONT); DrawObj(pm); glFinish();
		} else {
	      BuildWorldQuader(VMDoc->VMF.Base);
          GraphWin.InvalidateRect(NULL,TRUE);
		}
	  }
    } else if(!strnicmp(TextWin.inputstring,"CYLINDER",8)){
	  //vector pb=GetPoint("- basepoint: ", NULL);
      Cylinder *c = new Cylinder; c->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=c; VMDoc->VMF.Base=c;
      c->name = strdup("Cylinder");
	  if(!(VMDoc->VMF.MatList)){
	    VMDoc->VMF.MatList=new Material; VMDoc->VMF.MatList->name = strdup("default");
	  }
      c->Mat=VMDoc->VMF.MatList;
      if(c->succ){
	    glDrawBuffer(GL_FRONT); DrawObj(c); glFinish();
	  } else {
	    BuildWorldQuader(VMDoc->VMF.Base);
        GraphWin.InvalidateRect(NULL,TRUE);
	  }
    } else if(!strnicmp(TextWin.inputstring,"OPENPHOTO",9)){
	  OpenPhoto();
    } else if(!strnicmp(TextWin.inputstring,"OPENBACKGROUND",14)){
	  OpenBackground();
    } else if(!strnicmp(TextWin.inputstring,"MATCHCAM Direct",15)){
      PushView(&(GraphWin.GScreen.V));
	  MatchCam2();
 	  GraphWin.GScreen.V.name = "matched"; SetCaption();
    } else if(!strnicmp(TextWin.inputstring,"MATCHCAM Cascaded",17)){
      PushView(&(GraphWin.GScreen.V));
	  MatchCam1();
 	  GraphWin.GScreen.V.name = "matched"; SetCaption();
    } else if(!strnicmp(TextWin.inputstring,"MATCHCAM VhD",12)){
      PushView(&(GraphWin.GScreen.V));
	  MatchVhd();
 	  GraphWin.GScreen.V.name = "matched"; SetCaption();
    } else if(!strnicmp(TextWin.inputstring,"LINKVERTEX",10)){
      vector p1=GetPoint("- select geometry vertex: ", NULL);
      vector p2=GetWPt  ("- select image point: ");
      CDC *hdc;
      hdc = GraphWin.GetDC();
	  DrawVLink(hdc,p1,p2);
      GraphWin.ReleaseDC(hdc);
	  VLPtr vl = new(VLNode); vl->pg=p1; vl->pp=p2; vl->edg=0; vl->next=VMDoc->VLList; VMDoc->VLList=vl;
    } else if(!strnicmp(TextWin.inputstring,"LINKEDGE",8)){
      vector p1=GetPoint("- select first  edge vertex: ", NULL);
      vector p2=GetPoint("- select second edge vertex: ", NULL);
      vector p3=GetWPt  ("- select image point: ");
      CDC *hdc;
      hdc = GraphWin.GetDC();
	  DrawELink(hdc,p1,p2,p3);
      GraphWin.ReleaseDC(hdc);
	  VLPtr vl = new(VLNode); vl->pg=p1; vl->pe=p2; vl->pp=p3; vl->edg=1; vl->next=VMDoc->VLList; VMDoc->VLList=vl;
    } else if(!strnicmp(TextWin.inputstring,"GETTEX3D",8)){
      //PrimPtr Selection=SelectFaces("- select face(s) to get texture:\n");
      //Primitive *p;
      //if(p=Selection){
      //  while(p){
      //    if(p->idnt==POLYMESH){
      //      GetTex3D((PolyMesh *)p);
      //    }
      //    p=p->succ;
      //  }
      //  p=VMDoc->VMF.Base; while(p->succ) p=p->succ;
      //  p->succ = Selection;
      //  p->succ->pred = p;
	  //  //GraphWin.InvalidateRect(NULL,FALSE);
      //}
      GetTex3D(VMDoc->VMF.Base);
    } else if(!strnicmp(TextWin.inputstring,"ERASE",5)){
	  PrimPtr    Selection=SelectObjects("- select object(s) to erase:\n");
      if(Selection){
        Selection->Kill();
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"COORDS",6)){
      vector p1=GetPoint("- select point: ", NULL);
    } else if(!strnicmp(TextWin.inputstring,"MOVE",4)){
	  PrimPtr  p,Selection=SelectObjects("- select object(s) to move:\n");
      if(p=Selection){
        vector p1=GetPoint("- from point: ", p);
        vector p2=GetPoint("-   to point: ", p);
        while(p){
          (((Surface *)p)->TM.M41) += (p2-p1).x;
          (((Surface *)p)->TM.M42) += (p2-p1).y;
          (((Surface *)p)->TM.M43) += (p2-p1).z;
           ((Surface *)p)->Tb = ((Surface *)p)->TM.Inverse(); // calculate backtransformation matrix
           ((Surface *)p)->NM = ((Surface *)p)->TM.NTrans();  // calculate transformation matrix for normals
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"ROTATE",6)){
      ip=TextWin.inputstring; ip=SkipS(ip+6);
      float a; char ax=*ip; ip=SkipS(ip+1);
	  if(GetFloat(&a,ip)){
	    PrimPtr  p,Selection=SelectObjects("- select object(s) to rotate:\n");
        if(p=Selection){
          while(p){
			switch(ax){
			  case 'X': ((Surface *)p)->RotateX(a*Deg2Rad); break;
			  case 'Y': ((Surface *)p)->RotateY(a*Deg2Rad); break;
			  case 'Z': ((Surface *)p)->RotateZ(a*Deg2Rad); break;
			}
            ((Surface *)p)->NM = ((Surface *)p)->TM.NTrans();  // calculate transformation matrix for normals
            ((Surface *)p)->dflags &= ~1;
			if(((Surface *)p)->mesh.snm) delete ((Surface *)p)->mesh.snm; ((Surface *)p)->mesh.snm=NULL;
            p=p->succ;
          }
          if(p=VMDoc->VMF.Base){
		    while(p->succ) p=p->succ;
            p->succ = Selection;
            p->succ->pred = p;
		  } else VMDoc->VMF.Base=Selection;
	      GraphWin.InvalidateRect(NULL,FALSE);
		}
      } else {
        TextWin.Output(2,"- invalid angle '%s'!\n",ip);
      }
    } else if(!strnicmp(TextWin.inputstring,"MERGE",5)){
      PrimPtr    p;
      PolyMesh  *t;
	  int        GRP=0,TM=1,VC=0,FC=0,NC=0,TC=0,FV=0;
	  int        ofv,ofn,oft,ii,i,j,k;
	  PrimPtr    Selection=SelectObjects("- select meshes to merge together:\n");
      if(Selection){
        for(p=Selection; p; p=p->succ){
          if(p->idnt==POLYMESH){
            VC+=((PolyMesh *)p)->nv;
            NC+=((PolyMesh *)p)->nn;
            TC+=((PolyMesh *)p)->nt;
            FC+=((PolyMesh *)p)->nf;
			if(((PolyMesh *)p)->grp) GRP=1;
			if(((PolyMesh *)p)->pnv){
			  for(i=0; i<((PolyMesh *)p)->nf; i++) FV+=((PolyMesh *)p)->pnv[i];
			  TM=0;
			} else {
			  FV+=3*((PolyMesh *)p)->nf;
			}
          }
        }
        t=new PolyMesh; t->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=t; VMDoc->VMF.Base=t;
        t->TM      = ((PolyMesh *)Selection)->TM;
        t->Tb      = ((PolyMesh *)Selection)->Tb;
        t->NM      = ((PolyMesh *)Selection)->NM;
        t->pmflags = ((PolyMesh *)Selection)->pmflags;
        t->Mat     = ((PolyMesh *)Selection)->Mat;
        t->name    = ((PolyMesh *)Selection)->name; ((PolyMesh *)Selection)->name = NULL;  
		t->nf      = FC; t->pfv = new ushort[FV]; int fp=0, vpp=0;
		t->nv      = VC; t->pts = new vector[VC]; int vp=0;
		t->nn      = NC; if(t->nn){ t->nrm = new vector[t->nn]; t->pfn = new ushort[FV]; } int np=0, npp=0;
		t->nt      = TC; if(t->nt){ t->txc = new vector[t->nt]; t->pft = new ushort[FV]; } int tp=0, tpp=0;
        if(!TM) t->pnv = new byte[FC];
        if(GRP) t->grp = new byte[FC];

        for(p=Selection; p; p=p->succ){
          if(p->idnt==POLYMESH){
			ofv=vp; ofn=np; oft=tp;
			for(     i=0; i<((PolyMesh *)p)->nv; i++) t->pts[vp++]=((PolyMesh *)p)->pts[i];
			for(     i=0; i<((PolyMesh *)p)->nn; i++) t->nrm[np++]=((PolyMesh *)p)->nrm[i];
			for(     i=0; i<((PolyMesh *)p)->nt; i++) t->txc[tp++]=((PolyMesh *)p)->txc[i];
			for(ii=0,i=0; i<((PolyMesh *)p)->nf; i++){
			  k=((PolyMesh *)p)->pnv? ((PolyMesh *)p)->pnv[i] : 3;
			  for(j=0; j<k; j++){
				          t->pfv[vpp++]=                       ((PolyMesh *)p)->pfv[ii+j]+ofv    ;
				if(t->nn) t->pfn[npp++]=                       ((PolyMesh *)p)->pfn[ii+j]+ofn    ;
				if(t->nt) t->pft[tpp++]= ((PolyMesh *)p)->pft? ((PolyMesh *)p)->pft[ii+j]+oft : 0;
			  }
			  if(GRP) t->grp[fp] = ((PolyMesh *)p)->grp? ((PolyMesh *)p)->grp[i] : 0;
			  if(!TM) t->pnv[fp] = k;
			  fp++; ii+=k;
			}
          }
        }
        Selection->Kill();
  	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    /*} else if(!strnicmp(TextWin.inputstring,"MAKEMESH",8)){
      PolyMesh  *t;
	  int        i,j;
      
	  t=new PolyMesh; t->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=t; VMDoc->VMF.Base=t;
      
	  if(!VMDoc->VMF.MatList){
	    VMDoc->VMF.MatList=new MNode; VMDoc->VMF.MatList->next=NULL; VMDoc->VMF.MatList->map=NULL;
        strcpy(VMDoc->VMF.MatList->name,"default");
        // default-material:
        VMDoc->VMF.MatList->Od.R = 0.9F; VMDoc->VMF.MatList->Od.G = 0.8F; VMDoc->VMF.MatList->Od.B = 0.7F;
        VMDoc->VMF.MatList->Os.R = 1.0F; VMDoc->VMF.MatList->Os.G = 1.0F; VMDoc->VMF.MatList->Os.B = 1.0F;
        VMDoc->VMF.MatList->Ot.R = 1.0F; VMDoc->VMF.MatList->Ot.G = 1.0F; VMDoc->VMF.MatList->Ot.B = 1.0F;
        VMDoc->VMF.MatList->Ol.R = 1.0F; VMDoc->VMF.MatList->Ol.G = 1.0F; VMDoc->VMF.MatList->Ol.B = 1.0F;
        VMDoc->VMF.MatList->Kd   = 0.6F;
        VMDoc->VMF.MatList->Ks   = 0.3F; VMDoc->VMF.MatList->n    = 5.0F;
        VMDoc->VMF.MatList->Kt   = 0.0F; VMDoc->VMF.MatList->i    = 1.0F;
        VMDoc->VMF.MatList->Kl   = 0.0F;
        VMDoc->VMF.MatList->Kr   = 0.0F;
        VMDoc->VMF.MatList->R    = 0.0F;
      }	
      t->Mat=VMDoc->VMF.MatList;

      t->TM.InitTransformation(1.0F, 1.0F, 1.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F, 0.0F);
      t->pmflags = 0;
      t->name    = strdup("Mesh");  
	  t->nf      = (NZ-1)*(NZ-1); t->pfv = new ushort[4*t->nf]; int fp=0;
	  t->nv      = NZ*NZ; t->pts = new vector[t->nv]; int vp=0;
	  t->nn      = 0; t->nrm = NULL; t->pfn = NULL;
	  t->nt      = 0; t->txc = NULL; t->pft = NULL;
      t->pnv     = new byte[t->nf];
      t->grp     = NULL;

	  double ac[NZ][NZ]; double mx=0.0;

	  for(i=0; i<NZ; i++){
		for(j=0; j<NZ; j++){
		  ac[i][j]=0.0;
		}
	  }
      double a,b,x,y;
      for(i=0; i<1000000; i++){
        //a = 1.5707963*rnd();
        //a = acos(rnd());
        a = acos(pow(rnd(),1.0/5.0));
        b = 6.2831853*rnd();
        x = a*sin(b);
		y = a*cos(b);
	    ac[(int)(NZ/2+NZ/4*x)][(int)(NZ/2+NZ/4*y)]++;
		if(ac[(int)(NZ/2+NZ/4*x)][(int)(NZ/2+NZ/4*y)] > mx) mx=ac[(int)(NZ/2+NZ/4*x)][(int)(NZ/2+NZ/4*y)];
	  }

	  for(i=0; i<NZ; i++){
		for(j=0; j<NZ; j++){
          t->pts[i*NZ+j] = vector((float)(i-NZ/2), 10.0*ac[i][j]/mx, (float)(j-NZ/2));
		  if((i>0)&&(j>0)){
		    t->pnv[fp++]=4;
			t->pfv[vp++]=(i-1)*NZ+(j-1);
			t->pfv[vp++]=(i-1)*NZ+ j   ;
			t->pfv[vp++]= i   *NZ+ j   ;
			t->pfv[vp++]= i   *NZ+(j-1);
		  }
		}
	  }
      vector tp = vector(0.0f, 0.0f, 0.0f); //0.5 * (vector(minx,miny,minz)+vector(maxx,maxy,maxz));
      float  td = 15.0f; //Betrag(vector(maxx,maxy,maxz)-vector(minx,miny,minz));
      GraphWin.GScreen.V.Locate(tp+vector(1.0F*td,1.5F*td,2.5F*td));
      GraphWin.GScreen.V.LookAt(tp);
      GraphWin.GScreen.V.Tilt(0.0F); GraphWin.GScreen.V.SetViewAngle(0.5F); VMDoc->WorldValid=FALSE;
      vector N = unit(GraphWin.GScreen.V.Target - GraphWin.GScreen.V.Location); GraphWin.GScreen.V.Up = unit(vector(0.0f, 1.0f, 0.0f)-(vector(0.0f, 1.0f, 0.0f)%N)*N);
      GraphWin.InvalidateRect(NULL,FALSE);*/
    } else if(!strnicmp(TextWin.inputstring,"SINGLESIDED",11)){
      PrimPtr p, Selection=SelectObjects("- select objects to be singlesided:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            (((PolyMesh *)p)->pmflags) &= (~SURF_DOUBLESIDED);
            ((PolyMesh *)p)->dflags &= ~1;
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
  	    GraphWin.InvalidateRect(NULL,FALSE);
      }
	} else if(!strnicmp(TextWin.inputstring,"DOUBLESIDED",11)){
      PrimPtr p, Selection=SelectObjects("- select objects to be doublesided:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            (((PolyMesh *)p)->pmflags) |= SURF_DOUBLESIDED;
            glDrawBuffer(GL_FRONT); glDepthFunc(GL_LEQUAL);
		    CreateGL1(p, 0); glCallList(((PolyMesh *)p)->OGLL); glFinish();
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
      }
    } else if(!strnicmp(TextWin.inputstring,"FACETTED",8)){
      PrimPtr p,Selection=SelectObjects("- select objects to be facetted:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            (((PolyMesh *)p)->pmflags) &= (~(SURF_ROUNDED | SURF_GROUPED));
			if(((PolyMesh *)p)->grp){ delete(((PolyMesh *)p)->grp); ((PolyMesh *)p)->grp=NULL; }
			if(((PolyMesh *)p)->pfn){ delete(((PolyMesh *)p)->pfn); ((PolyMesh *)p)->pfn=NULL; }
			if(((PolyMesh *)p)->nrm){ delete(((PolyMesh *)p)->nrm); ((PolyMesh *)p)->nrm=NULL; }
            glDrawBuffer(GL_FRONT); glDepthFunc(GL_LEQUAL);
		    CreateGL1(p, 0); glCallList(((PolyMesh *)p)->OGLL); glFinish();
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
      }
	} else if(!strnicmp(TextWin.inputstring,"SMOOTH",6)){
      PrimPtr p,Selection=SelectObjects("- select objects to be smoothed:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            (((PolyMesh *)p)->pmflags) &= (~SURF_GROUPED);
            (((PolyMesh *)p)->pmflags) |=   SURF_ROUNDED ;
			if(((PolyMesh *)p)->grp){ delete(((PolyMesh *)p)->grp); ((PolyMesh *)p)->grp=NULL; }
			if(((PolyMesh *)p)->pfn){ delete(((PolyMesh *)p)->pfn); ((PolyMesh *)p)->pfn=NULL; }
			if(((PolyMesh *)p)->nrm){ delete(((PolyMesh *)p)->nrm); ((PolyMesh *)p)->nrm=NULL; ((PolyMesh *)p)->nn=0; }
            glDrawBuffer(GL_FRONT); glDepthFunc(GL_LEQUAL);
		    CreateGL1(p, 0); glCallList(((PolyMesh *)p)->OGLL); glFinish();
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
      }
	} else if(!strnicmp(TextWin.inputstring,"OPTIMIZE",8)){
      PrimPtr p,Selection=SelectObjects("- select mesh-objects to optimize:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
			OptimizeMesh((PolyMesh *)p);
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
  	    GraphWin.InvalidateRect(NULL,FALSE);
      }
	} else if(!strnicmp(TextWin.inputstring,"UNMAP",5)){
      PrimPtr p,Selection=SelectObjects("- select objects to remove mapping:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
			if(((PolyMesh *)p)->txc){ delete ((PolyMesh *)p)->txc; ((PolyMesh *)p)->txc = NULL; }
			if(((PolyMesh *)p)->pft){ delete ((PolyMesh *)p)->pft; ((PolyMesh *)p)->pft = NULL; }
			   ((PolyMesh *)p)->nt = 0; 
          }
          p=p->succ;
        }
        if(p=VMDoc->VMF.Base){
		  while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
  	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"NORMALS",7)){
      ip=TextWin.inputstring; ip=SkipS(ip+7);
      if(!strnicmp(ip,"FLIP",4)){
		int        fi,i,j,k;
        PrimPtr p,Selection=SelectObjects("- select object(s) to flip normals:\n");
        if(p=Selection){
          while(p){
            if(p->idnt==POLYMESH){
              PolyMesh *tp = (PolyMesh *)p;
              for(fi=0,i=0; i<(tp->nf); i++){
			    k=(tp->pnv)? tp->pnv[i] : 3;
				for(j=0; j<(k>>1); j++){
				              swap(tp->pfv[fi+j], tp->pfv[fi+k-1-j]);
				  if(tp->pfn) swap(tp->pfn[fi+j], tp->pfn[fi+k-1-j]);
				  if(tp->pft) swap(tp->pft[fi+j], tp->pft[fi+k-1-j]);
				}
				fi+=k;
			  }
              tp->dflags &= ~1;
			  if(tp->mesh.lin) delete tp->mesh.lin; tp->mesh.lin=NULL;
			  if(tp->mesh.snm) delete tp->mesh.snm; tp->mesh.snm=NULL;
			  if(tp->nrm) delete tp->nrm; tp->nrm=NULL;
            }
            p=p->succ;
          }
          if(p=VMDoc->VMF.Base){
		    while(p->succ) p=p->succ;
            p->succ = Selection;
            p->succ->pred = p;
		  } else VMDoc->VMF.Base=Selection;
  	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  } else if(!strnicmp(ip,"UNIFY",5)){
        PrimPtr p,Selection=SelectObjects("- select object(s) to unify normals:\n");
        if(p=Selection){
          while(p){
			UnifyNormals(p);
            p=p->succ;
          }
          if(p=VMDoc->VMF.Base){
  		    while(p->succ) p=p->succ;
            p->succ = Selection;
            p->succ->pred = p;
		  } else VMDoc->VMF.Base=Selection;
  	      GraphWin.InvalidateRect(NULL,FALSE);
        }
	  }
    } else if(!strnicmp(TextWin.inputstring,"SEPERATE",8) || !strnicmp(TextWin.inputstring,"DELFACES",8)){
	  MatPtr     mats[50];
	  PrimPtr    p;
	  PolyMesh  *r;
	  PolyMesh  *t;
	  ushort    *nfv;
	  byte      *nnv;
	  byte      *ngr;
	  byte      *nmt;
	  int        i,j,k,i2,j2,fc,vc,mv,SM,mp,mc=0;
	  int        GRP=0,Sep = !strnicmp(TextWin.inputstring,"SEPERATE",8);
	  PrimPtr    Selection = SelectFaces(Sep? "- select faces to seperate:\n" : "- select faces to delete:\n");
	  if(p=Selection){
        if(Sep){
          VectorSet VB; fc=0; vc=0;
          while(p){
            if(p->idnt==POLYMESH){
              r=(PolyMesh *)p; if(r->grp) GRP=1;
			  if(!(r->mat)){ for(mp=0; (mp<mc) && (r->Mat != mats[mp]); mp++); if(mp>=mc) mats[mc++]=r->Mat; }
              for(i=0,i2=0; i<(r->nf); i++){
                if(r->flg[i] & SELECTED){
                  // search/insert vertices in buffer and receive new index
                  for(k=0; k<(r->pnv? r->pnv[i] : 3); k++){
	    		    r->pfv[i2+k] = VB.Add(r->pts[r->pfv[i2+k]], 0.0001f);
	              }
			      // store material
				  if(r->mat){ for(mp=0; (mp<mc) && (r->mtl[r->mat[i]] != mats[mp]); mp++); if(mp>=mc) mats[mc++]=r->mtl[r->mat[i]]; }
                  vc+=k; fc++;
                }
	            i2 += (r->pnv? r->pnv[i] : 3);
              }
            }
            p=p->succ;
          }
          if(fc){
            r = (PolyMesh *)Selection; VMDoc->WorldValid=FALSE;
            t = new PolyMesh; t->succ=VMDoc->VMF.Base; if(VMDoc->VMF.Base) VMDoc->VMF.Base->pred=t; VMDoc->VMF.Base=t;
            t->TM      = r->TM;
            t->pmflags = r->pmflags;
            t->Mat     = r->Mat;
            t->name    = new char[strlen(r->name)+1]; strcpy(t->name,r->name);
            t->nf      = fc; t->pfv = new(ushort[vc]); t->pnv = new(byte[fc]);
            t->nv      = VB.Move2Array(&(t->pts)); j=0; j2=0; mv=0;
		    if(GRP) { t->grp = new(byte[fc]); t->pmflags &= ~SURF_ROUNDED; t->pmflags |= SURF_GROUPED; }
			if(mc>1){
			  t->mtl = new(MatPtr[mc]); for(mp=0; mp<mc; mp++) t->mtl[mp]=mats[mp];
			  t->mat = new(byte[fc]); t->nm = mc;
			} else {
              t->Mat = mats[0];
			}
            // copy faces to new Mesh
            while(r){
              if(r->idnt==POLYMESH){
                for(i=0,i2=0; i<(r->nf); i++){
                  if(r->flg[i] & SELECTED){
                    for(k=0; k<(r->pnv? r->pnv[i] : 3); k++){
                      t->pfv[j2+k] = r->pfv[i2+k];
                    }
                    t->pnv[j]=k; if(k>mv) mv=k;
				    if(GRP) t->grp[j] = (r->grp? r->grp[i] : (r->pmflags & SURF_ROUNDED)? 255 : 0);
					if(mc>1){
					  if(r->mat){
			            for(mp=0; r->mtl[r->mat[i]] != mats[mp]; mp++);
					  } else {
			            for(mp=0;        r->Mat     != mats[mp]; mp++);
					  }
					  t->mat[j]=mp;
					}
                    j2+=k; j++;
                  }
                  i2+=(r->pnv? r->pnv[i] : 3);
                }
              }
              r=(PolyMesh *)r->succ;
            }
            if(mv==3) { delete(t->pnv); t->pnv=NULL; }
          }
        }
        // remove selected faces from meshes
        p=Selection;
        while(p){
          fc=1;
          if(p->idnt==POLYMESH){
            t=(PolyMesh *)p; fc=0; vc=0; mv=0; SM=1;
            for(i=0; i<(t->nf); i++) if(!((t->flg[i]) & SELECTED)){
			  fc++; vc+=(t->pnv? t->pnv[i] : 3);
			}
            if(fc > 0){
              nfv=new(ushort[vc]);
			  if(t->pnv) nnv=new(byte[fc]);
			  if(t->grp) ngr=new(byte[fc]);
			  if(t->mat) nmt=new(byte[fc]);
              for(i=0,i2=0,j=0,j2=0; i<(t->nf); i++){
				if(!((t->flg[i]) & SELECTED)) {
                  for(k=0; k<(t->pnv? t->pnv[i] : 3); k++){
				    nfv[j2+k] = t->pfv[i2+k];
				  }
                  if(t->pnv)   nnv[j]=k;
                  if(t->grp)   ngr[j]=t->grp[i];
                  if(t->mat) { nmt[j]=t->mat[i]; if(nmt[j]!=nmt[0]) SM=0; }
				  j2+=k; j++; if(k>mv) mv=k;
                }
				i2+=(t->pnv? t->pnv[i] : 3);
			  }
              delete(t->pfv); t->pfv=nfv;
			  if(t->pnv){
				delete(t->pnv);
				if(mv>3) t->pnv=nnv; else { t->pnv=NULL; delete(nnv); }
			  }
			  if(t->grp){ delete(t->grp); t->grp=ngr; }
			  if(t->mat){
				delete(t->mat);
				if(SM){
				  t->Mat = t->mtl[nmt[0]]; t->nm=0;
				  delete(   nmt); t->mat=NULL;
				  delete(t->mtl); t->mtl=NULL;
			    } else t->mat=nmt;
			  }
              t->nf = fc; DelIsoVert(p,FALSE);
              t->dflags &= ~1;
			  if(t->mesh.lin) delete t->mesh.lin; t->mesh.lin=NULL;
			  if(t->mesh.snm) delete t->mesh.snm; t->mesh.snm=NULL;
			  if((t->nrm) && !(t->nn)) { delete t->nrm; t->nrm=NULL; }
            }
          }
          if(fc) p=p->succ; else p=Kill1(p);
        }
        if(Selection){
		  if(p=VMDoc->VMF.Base){
			while(p->succ) p=p->succ;
            p->succ = Selection;
            Selection->pred = p;
		  } else VMDoc->VMF.Base=Selection;
        }
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"SPLIT",5)){
      ip=TextWin.inputstring; ip=SkipS(ip+5);
      if(!strnicmp(ip,"CONNECTED",9)){
	    PrimPtr p,Selection=SelectObjects("- select objects to split:\n");
	    if(p=Selection){
          while(p){
            if(p->idnt==POLYMESH) SplitConnected((PolyMesh *)p);
            p=p->succ;
          }
          Selection->Kill();
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
      } else if(!strnicmp(ip,"MATERIALS",9)){
	    PrimPtr p,Selection=SelectObjects("- select objects to split:\n");
	    if(p=Selection){
          while(p){
            if(p->idnt==POLYMESH) SplitMaterials((PolyMesh *)p);
            p=p->succ;
          }
          Selection->Kill();
	      GraphWin.InvalidateRect(NULL,FALSE);
        }
      }
    } else if(!strnicmp(TextWin.inputstring,"SMGRP",5)){
      PrimPtr    p;
      PolyMesh  *t;
      int        i,fc,grpnr;
	  ip=TextWin.inputstring; ip=SkipS(ip+5); grpnr=(int)(*ip)-(int)'0';
      PrimPtr    Selection=SelectFaces("- select face(s) for smoothing group:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            t=(PolyMesh *)p; fc=0;
			if(!(t->grp)) t->grp = new(byte[t->nf]);
            for(i=0; i<(t->nf); i++) {
              if(t->flg[i] & SELECTED) {
                t->grp[i]=grpnr; fc++;
              } else {
                if(!(t->pmflags & SURF_GROUPED)){
                  t->grp[i] = (t->pmflags & SURF_ROUNDED)? 255 : 0;
                }
              }
            }
            if(fc>0){
              t->pmflags |=  SURF_GROUPED;
              t->dflags &= ~1;
			  if(t->nrm) delete t->nrm; t->nn=0; t->nrm=NULL;
            } else {
              t->pmflags &= ~SURF_GROUPED;
			  delete(t->grp);
			}
          }
          p=p->succ;
        }
        if(VMDoc->VMF.Base){
		  p=VMDoc->VMF.Base; while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"FACEFLIP",8)){
      PrimPtr    p;
      PolyMesh  *t;
      int        fi,i,j,k;
      PrimPtr    Selection=SelectFaces("- select face(s) to flip:\n");
      if(p=Selection){
        while(p){
          if(p->idnt==POLYMESH){
            t=(PolyMesh *)p;
            for(fi=0,i=0; i<(t->nf); i++){
			  k=(t->pnv)? t->pnv[i] : 3;
              if(t->flg[i] & SELECTED){
			    for(j=0; j<(k>>1); j++) swap(t->pfv[fi+j], t->pfv[fi+k-1-j]);
			  }
			  fi+=k;
			}
			if(t->mesh.lin) delete t->mesh.lin; t->mesh.lin=NULL;
			if(t->mesh.snm) delete t->mesh.snm; t->mesh.snm=NULL;
          }
          p=p->succ;
        }
        if(VMDoc->VMF.Base){
		  p=VMDoc->VMF.Base; while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"FACEMAP",7)){
      PrimPtr p, Selection = SelectFaces("- select face(s) to apply mapping:\n");
      if(Selection){
		MapFaces(Selection);
        if(VMDoc->VMF.Base){
		  p=VMDoc->VMF.Base; while(p->succ) p=p->succ;
          p->succ = Selection;
          p->succ->pred = p;
		} else VMDoc->VMF.Base=Selection;
	    GraphWin.InvalidateRect(NULL,FALSE);
      }
    } else if(!strnicmp(TextWin.inputstring,"RAYTRACE",8)){
      RECT        win;
	  vector      Ray;
      BITMAPINFO  BMPinfo;
      byte       *BMPbits;

	  if(!RV_SceneTree) InitRaVi();

      GraphWin.GetClientRect(&win);

      BMPinfo.bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
      BMPinfo.bmiHeader.biWidth         = win.right;
      BMPinfo.bmiHeader.biHeight        = win.bottom;
      BMPinfo.bmiHeader.biPlanes        = 1;
      BMPinfo.bmiHeader.biBitCount      = 32;
      BMPinfo.bmiHeader.biCompression   = BI_RGB;
      BMPinfo.bmiHeader.biSizeImage     = win.right*win.bottom*4;
      BMPinfo.bmiHeader.biXPelsPerMeter = 0;
      BMPinfo.bmiHeader.biYPelsPerMeter = 0;
      BMPinfo.bmiHeader.biClrUsed       = 0;
      BMPinfo.bmiHeader.biClrImportant  = 0;

      BMPbits = new byte[BMPinfo.bmiHeader.biSizeImage];

	  ray_thresh=0.001f;
	 *RV_CamPtr = GraphWin.GScreen.V.Location;
	  RV_Target = GraphWin.GScreen.V.Target;
	  RV_Up     = GraphWin.GScreen.V.Up;
	  RV_FoV    = 2.0f*atanf(GraphWin.GScreen.V.h/2.0f);
	  //Light     = LPos;

      DWORD ts = GetTickCount();

	  RV_Quality=1; RV_AntiAlias=1; RV_Filter=0; RV_Shader = &Shade;
      RV_RenderImage(BMPbits, win.right, win.bottom, 0); cframe++;

	  DWORD tc=GetTickCount()-ts;
	  TextWin.Output(2,"- rendering time: %5.2f seconds\n",0.001f*(float)tc);

      CDC *hdc; hdc = GraphWin.GetDC();
      SetDIBitsToDevice(hdc->m_hDC, 0, 0, win.right, win.bottom, 0, 0, 0, win.bottom, BMPbits, &(BMPinfo), DIB_RGB_COLORS);
	  GraphWin.ReleaseDC(hdc); if(BMPbits) delete BMPbits;
    } else if(strnicmp(TextWin.inputstring,"QUIT",4)){
      TextWin.Output(3,"-unknown command '%s'!\n",TextWin.inputstring);
    }
  }while(strnicmp(TextWin.inputstring,"quit",4));
  TextWin.Output(1,"Bye!\n");
  VMDoc->KillDoc();
  PostQuitMessage(0);

  return TRUE;
}


class CAboutDlg : public CDialog
{
  public: CAboutDlg();

  // Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

  // Implementation
  protected:
	//{{AFX_MSG(CAboutDlg)
		// No message handlers
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
		// No message handlers
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

// App command to run the dialog
void CVModelApp::OnAppAbout()
{
  CAboutDlg aboutDlg;
  aboutDlg.DoModal();
}


CDocument* CVModelApp::OpenDocumentFile(LPCTSTR lpszFileName) 
{
  TextWin.KeyIn("LOAD %s\n", lpszFileName);
  return VMDoc;
}


//***********************************************************************
// Function: OnCmdMsg()
//
// Purpose:
//    OnCmdMsg() is called by the framework to route and dispatch
//    command messages and to handle the update of command
//    user-interface objects.
//
//    Here we extend the standard command routing to intercept command
//    messages with variable command IDs. Otherwise,
//    we route the command to the standard OnCmdMsg processing.
//
// Parameters:
//    nID           -- contains the command ID
//    nCode         -- identifies the command notification code
//    pExtra        -- used according to the value of nCode
//    pHandlerInfo  -- if not NULL, filled in with pTarget and pmf
//                     members of CMDHANDLERINFO structure, instead
//                     of dispatching the command. Typically NULL.
//
// Returns:
//    nonzero if message handled, otherwise 0.
//
// Comments:
//    See the documentation for CCmdTarget::OnCmdMsg() for further
//    information.  Command routing is also discussed in tech note #21.
//
//***********************************************************************
BOOL CVModelApp::OnCmdMsg(UINT nID, int nCode, void* pExtra, AFX_CMDHANDLERINFO* pHandlerInfo) 
{
  // If pHandlerInfo is NULL, then handle the message
  if(pHandlerInfo == NULL){
    // Filter the commands sent to a text color menu option
	if((nID>=1230) && (nID<=1240)){
      if(nCode == CN_COMMAND){
        // Handle WM_COMMAND message
        ViewPtr V=VMDoc->VMF.ViewList;
		for(unsigned int i=1230; i<nID; i++) V=V->next;
        TextWin.KeyIn("VRESTORE %s\n",V->name);
      }	else if(nCode == CN_UPDATE_COMMAND_UI){
        // Update UI element state
	    ((CCmdUI*)pExtra)->Enable(TRUE);
      }
      return TRUE;
	}
  }
  // If we didn't process the command, call the base class
  // version of OnCmdMsg so the message-map can handle the message
  return CWinApp::OnCmdMsg(nID, nCode, pExtra, pHandlerInfo);
}



/////////////////////////////////////////////////////////////////////////////
// CVModelApp commands

void CVModelApp::OnImport3DS() { TextWin.KeyIn("IMPORT 3DS\n" ); }
void CVModelApp::OnImportWRL() { TextWin.KeyIn("IMPORT VRML\n"); }
void CVModelApp::OnImportLWO() { TextWin.KeyIn("IMPORT LWO\n" ); } 
void CVModelApp::OnImportDXF() { TextWin.KeyIn("IMPORT DXF\n" ); } 
void CVModelApp::OnImportObj() { TextWin.KeyIn("IMPORT OBJ\n" ); }
void CVModelApp::OnImportDEM() { TextWin.KeyIn("IMPORT DEM\n" ); } 
void CVModelApp::OnImportGEO() { TextWin.KeyIn("IMPORT GEO\n" ); } 

void CVModelApp::OnExportPOV () { TextWin.KeyIn("EXPORT POV\n" ); }
void CVModelApp::OnExportDXF () { TextWin.KeyIn("EXPORT DXF\n" ); }
void CVModelApp::OnExportRib () { TextWin.KeyIn("EXPORT RIB\n" ); } 
void CVModelApp::OnExportVRML() { TextWin.KeyIn("EXPORT VRML\n"); } 
void CVModelApp::OnExportVL  () { TextWin.KeyIn("EXPORT VIB\n" ); }  

void CVModelApp::OnViewDolly()       { TextWin.KeyIn("DOLLY\n"    ); }
void CVModelApp::OnViewRotate()      { TextWin.KeyIn("VROTATE\n"  ); }
void CVModelApp::OnViewLookaround()  { TextWin.KeyIn("VTURN\n"    ); }
void CVModelApp::OnViewWalk()        { TextWin.KeyIn("VWALK\n"    ); }
void CVModelApp::OnViewZoomAngle()   { TextWin.KeyIn("VANGLE "    ); }
void CVModelApp::OnViewZoomDynamic() { TextWin.KeyIn("ZOOM\n"     ); }
void CVModelApp::OnViewPan()         { TextWin.KeyIn("PAN\n"      ); }
void CVModelApp::OnViewPrevious()    { TextWin.KeyIn("VPREVIOUS\n"); }

void CVModelApp::OnViewRestoreTop()   { TextWin.KeyIn("VRESTORE Top\n"   ); }
void CVModelApp::OnViewRestoreFront() { TextWin.KeyIn("VRESTORE Front\n" ); }
void CVModelApp::OnViewRestoreRight() { TextWin.KeyIn("VRESTORE Right\n" ); }
void CVModelApp::OnViewRestoreLeft()  { TextWin.KeyIn("VRESTORE Left\n"  ); } 
void CVModelApp::OnViewSave()         { TextWin.KeyIn("VSAVE "           ); } 

void CVModelApp::OnViewAuto() { TextWin.KeyIn("AUTOROTATE\n" ); }
void CVModelApp::OnFileInfo() { TextWin.KeyIn("INFO\n" ); }

void CVModelApp::OnDisplayShaded()    { TextWin.KeyIn("VSHADE\n"    ); }
void CVModelApp::OnDisplayTextures()  { TextWin.KeyIn("VTEXTURES\n" ); }
void CVModelApp::OnDisplayAa()        { TextWin.KeyIn("VANTIALIAS\n"); }

void CVModelApp::OnDisplayShowInverse() { TextWin.KeyIn("SHOW INVERSE\n" ); }
void CVModelApp::OnDisplayShowNot()     { TextWin.KeyIn("SHOW NOT\n"     ); }
void CVModelApp::OnDisplayShowOnly()    { TextWin.KeyIn("SHOW ONLY\n"    ); }
void CVModelApp::OnDisplayShowAll()     { TextWin.KeyIn("SHOW ALL\n"     ); }

void CVModelApp::OnModifyObjectDoublesided() { TextWin.KeyIn("DOUBLESIDED\n" ); }
void CVModelApp::OnModifyObjectSinglesided() { TextWin.KeyIn("SINGLESIDED\n" ); }
void CVModelApp::OnModifyObjectFacetted()    { TextWin.KeyIn("FACETTED\n"    ); }
void CVModelApp::OnModifyObjectSmooth()      { TextWin.KeyIn("SMOOTH\n"      ); }

void CVModelApp::OnDisplayPerspective()
{
  if(GraphWin.GScreen.V.h){
	GraphWin.GScreen.V.old_h = GraphWin.GScreen.V.h;
	TextWin.KeyIn("VANGLE 0\n" );
  } else {
	TextWin.KeyIn("VANGLE PREVIOUS\n" );
  }
}
void CVModelApp::OnRaytrace() { TextWin.KeyIn("RAYTRACE\n"); }

void CVModelApp::OnModifyObjectUnmap() { TextWin.KeyIn("UNMAP\n" ); } 
void CVModelApp::OnModifyObjectErase() { TextWin.KeyIn("ERASE\n" ); }
void CVModelApp::OnModifyObjectMerge() { TextWin.KeyIn("MERGE\n" ); }
void CVModelApp::OnModifyObjectMove()  { TextWin.KeyIn("MOVE\n"  ); }

void CVModelApp::OnRotateZ() { TextWin.KeyIn("ROTATE Z "); } 
void CVModelApp::OnRotateY() { TextWin.KeyIn("ROTATE Y "); }
void CVModelApp::OnRotateX() { TextWin.KeyIn("ROTATE X "); }

void CVModelApp::OnModObjNormFlip()    { TextWin.KeyIn("NORMALS FLIP\n"  ); }
void CVModelApp::OnNormalsUnify()      { TextWin.KeyIn("NORMALS UNIFY\n" ); }
void CVModelApp::OnOptimizeMesh()      { TextWin.KeyIn("OPTIMIZE\n"      ); }

void CVModelApp::OnSplitConnected()    { TextWin.KeyIn("SPLIT CONNECTED\n"); }
void CVModelApp::OnSplitMaterials()    { TextWin.KeyIn("SPLIT MATERIALS\n"); }

void CVModelApp::OnModifyFaceSep()     { TextWin.KeyIn("SEPERATE\n" ); }
void CVModelApp::OnModifyFaceDelete()  { TextWin.KeyIn("DELFACES\n" ); }
void CVModelApp::OnFaceFlip()          { TextWin.KeyIn("FACEFLIP\n" ); }
void CVModelApp::OnFaceMap()           { TextWin.KeyIn("FACEMAP\n"  ); }

void CVModelApp::OnCreatePolygon()     { TextWin.KeyIn("POLYGON\n"  ); }
void CVModelApp::OnCreateCylinder()    { TextWin.KeyIn("CYLINDER\n" ); } 

void CVModelApp::OnMeasureCoords()     { TextWin.KeyIn("COORDS\n"   ); }

void CVModelApp::OnModifyFaceSmoothingNone()   { TextWin.KeyIn("SMGRP 0\n" ); }
void CVModelApp::OnModifyFaceSmoothingGroup1() { TextWin.KeyIn("SMGRP 1\n" ); }
void CVModelApp::OnModifyFaceSmoothingGroup2() { TextWin.KeyIn("SMGRP 2\n" ); }
void CVModelApp::OnModifyFaceSmoothingGroup3() { TextWin.KeyIn("SMGRP 3\n" ); }
void CVModelApp::OnModifyFaceSmoothingGroup4() { TextWin.KeyIn("SMGRP 4\n" ); }

void CVModelApp::OnPhGetTex3D() { TextWin.KeyIn("GETTEX3D\n"      ); }
void CVModelApp::OnPhOpenBack() { TextWin.KeyIn("OPENBACKGROUND\n"); }  
void CVModelApp::OnPhOpenExt()  { TextWin.KeyIn("OPENPHOTO\n"     ); } 
void CVModelApp::OnPhLink()     { TextWin.KeyIn("LINKVERTEX\n"    ); }
void CVModelApp::OnPhLinkEdge() { TextWin.KeyIn("LINKEDGE\n"      ); } 
void CVModelApp::OnPhMatch1()   { TextWin.KeyIn("MATCHCAM Cascaded\n" ); }
void CVModelApp::OnPhotoMatch2(){ TextWin.KeyIn("MATCHCAM Direct\n"   ); }
void CVModelApp::OnPhMatchVhd() { TextWin.KeyIn("MATCHCAM VhD\n"      ); }

void CVModelApp::OnUpdateDisplayShaded(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck((GraphWin.DisplayMode & VM_SHADED)? 1 : 0);
}

void CVModelApp::OnUpdateDisplayTextures(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck((GraphWin.DisplayMode & VM_TEXTURED)? 1 : 0);
}

void CVModelApp::OnUpdateDisplayAa(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck((GraphWin.DisplayMode & VM_ANTIALIAS)? 1 : 0);
}

void CVModelApp::OnUpdateDisplayPerspective(CCmdUI* pCmdUI) 
{
  pCmdUI->SetCheck((GraphWin.GScreen.V.h)? 1 : 0);
}

void CVModelApp::OnUpdateViewPrevious(CCmdUI* pCmdUI) 
{
  pCmdUI->Enable(VMDoc->ViewStack? 1 : 0);
}

